/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.connectionCache;

import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

public class ConnectionCachePurger<INFO, CONN> implements Runnable, ConnectionCache.DiscardTester
{
	protected ConnectionCache<INFO, CONN> cache;
	protected int minSpareConnections;

	protected long mLastFailedNewSpare = 0L;
	protected long mWaitTimeAfterFailedOpen = 60 * 60 * 1000L;
	//
	public boolean debugFlag = false;
	public boolean progressLog = true;

	public ConnectionCachePurger(ConnectionCache<INFO, CONN> cache, ScheduledExecutorService executorService,
	                             int minSpareConnections)
	{
		this.cache = cache;
		this.minSpareConnections = minSpareConnections;
		long sleepTime = 15 * 1000L;
		executorService.scheduleWithFixedDelay(this, sleepTime, sleepTime, TimeUnit.MILLISECONDS);
	}

	public void log(String s)
	{
		System.err.println("%% " + cache.getCacheName() + ": " + s);
	}

	public synchronized void run()
	{
		if (debugFlag)
			log("run: start");
		try {
			doCheck();
		}
		catch (Exception ex) {
			log("Hmm, exception at top-level of run():\n" + ex.toString());
		}
		if (debugFlag)
			log("run: sleeps");
	}

	protected synchronized void doCheck() throws Exception
	{
		int numOldConns = cache.getNumCachedConnections();
		if (debugFlag)
			log("doCheck.1: " + cache.getNumCachedConnections());
		int purged = purgeOldConnections();
		if (debugFlag)
			log("doCheck.2: " + cache.getNumCachedConnections());
		int created = checkSpareConnections();
		if (debugFlag)
			log("doCheck.3: " + cache.getNumCachedConnections());
		if (progressLog && (purged != 0 || created != 0)) {
			log("had " + numOldConns + ", purged " + purged + ", created " + created);
		}
	}

	protected int purgeOldConnections() throws Exception
	{
		return cache.discardOldConnections(this, 30 * 60 * 1000L);
	}

	@Override
	public boolean isWanted(long cacheAge, int numIdle)
	{
		if (numIdle > 10)
			return cacheAge < 30 * 1000L;
		if (numIdle > 5)
			return cacheAge < 4 * 60 * 1000L;
		return cacheAge < 15 * 60 * 1000L;
	}

	public int checkSpareConnections() throws Exception
	{
		if (System.currentTimeMillis() - mLastFailedNewSpare < mWaitTimeAfterFailedOpen)
			return 0;
		int numCached = cache.getNumCachedConnections();
		if (numCached >= minSpareConnections)
			return 0;
		if (debugFlag) {
			log("has " + numCached + " spare conns, want to have " + minSpareConnections);
		}
		int numCreated = 0;
		for (int i = numCached; i < minSpareConnections; i++) {
			numCached = Math.max(numCached, cache.getNumCachedConnections());
			if (numCached >= minSpareConnections)
				break; // extra check
			try {
				if (debugFlag) {
					log("creates spare connection (had " + numCached + ")");
				}
				cache.createNewSpareConnection();
				numCreated++;
				if (debugFlag) {
					log("after new spare: has " + cache.getNumCachedConnections());
				}
			}
			catch (Exception ex) {
				if (debugFlag) {
					log("Got exception while creating spare connection:\n" + ex.toString());
				}
				mLastFailedNewSpare = System.currentTimeMillis();
				break;
			}
		}
		if (debugFlag)
			log("done creating spare conns");
		return numCreated;
	}
}
