/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.runtime.lib;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public abstract class TemplateDataLoader
{
	public static void setFileBaseDir(@NotNull File baseDir)
	{
		loader = new UsingFilesystem(baseDir);
	}

	public static void setUseClassResources()
	{
		loader = new UsingClassResource();
	}

	@NotNull
	public static TemplateData load(Class<? extends TemplateBase> cls) throws Exception
	{
		assert loader != null;
		return loader.loadData(cls);
	}

	@Nullable
	private static TemplateDataLoader loader;


	@NotNull
	public abstract TemplateData loadData(Class<? extends TemplateBase> cls) throws Exception;

	@NotNull
	public static String getRelativePath(Class<? extends TemplateBase> cls)
	{
		return cls.getName().replace('.', '/') + ".tmpldata";
	}

	public static class UsingClassResource extends TemplateDataLoader
	{
		protected static Map<Class<?>, TemplateData> cache = new HashMap<>();

		@NotNull
		public synchronized TemplateData loadData(Class<? extends TemplateBase> cls) throws Exception
		{
			TemplateData td = cache.get(cls);
			if (td != null)
				return td;
			String path = "/" + getRelativePath(cls);
			InputStream is = getClass().getResourceAsStream(path);
			if (is == null)
				throw new TemplateException("TemplateData not found for " + path);
			if (!(is instanceof BufferedInputStream))
				is = new BufferedInputStream(is, 4096);
			try {
				td = new TemplateData(new DataInputStream(is));
			}
			finally {
				is.close();
			}
			cache.put(cls, td);
			return td;
		}
	}

	public static class UsingFilesystem extends TemplateDataLoader
	{
		public final File baseDirF;

		public UsingFilesystem(@NotNull File baseDirF)
		{
			this.baseDirF = baseDirF;
		}

		protected static class MetaData
		{
			@NotNull final TemplateData data;
			final long moddate;
			final File f;

			MetaData(@NotNull TemplateData data, long moddate, File f)
			{
				this.data = data;
				this.moddate = moddate;
				this.f = f;
			}
		}

		@NotNull
		protected static final Map<Class<?>, MetaData> cache = new HashMap<>();

		@NotNull
		public synchronized TemplateData loadData(Class<? extends TemplateBase> cls) throws Exception
		{
			MetaData meta = cache.get(cls);
			if (meta != null && meta.moddate == meta.f.lastModified())
				return meta.data;
			File f = new File(baseDirF, getRelativePath(cls));
			if (!f.exists())
				throw new TemplateException("TemplateData file not found for " + cls.getName());
			long moddate = f.lastModified();
			try (FileInputStream fis = new FileInputStream(f)) {
				TemplateData td = new TemplateData(new BufferedInputStream(fis, 8196));
				meta = new MetaData(td, moddate, f);
				cache.put(cls, meta);
				return td;
			}
		}
	}
}
