/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.buildsystem;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.jabhts.annotation.global.TCompile;
import org.jabhts.annotation.global.TIgnore;
import org.jabhts.translator.compiler.utils.AnnotationHelper;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.LogLevel;

public class TCompileSet
{
	public final CompilerLog log;

	private final Map<String, Class<?>> map = new HashMap<>();
	private final Set<String> warnedUnknown = new HashSet<>();
	private final List<CompilationInfo> compilationInfos = new ArrayList<>();
	private final Set<Class<?>> seenSuperClasses = new HashSet<>();

	public TCompileSet(CompilerLog log, Class<?> mainClass) throws CompilerException
	{
		this.log = log;
		addCompilationInfos(mainClass);
	}

	public List<CompilationInfo> getCompilationInfos()
	{
		return compilationInfos;
	}

	public boolean isEmpty()
	{
		return compilationInfos.isEmpty();
	}

	private void addCompilationInfos(Class<?> cls) throws CompilerException
	{
		if (!cls.isAnnotationPresent(TIgnore.class)) {
			addTpnames(cls);
			TCompile a = cls.getAnnotation(TCompile.class);
			if (a != null)
				compilationInfos.add(new CompilationInfo(a, tpname(cls, a), cls, this));
		}
		for (Class<?> inner : cls.getDeclaredClasses())
			addCompilationInfos(inner);
	}

	private void addTpnames(Class<?> mainClass) throws CompilerException
	{
		for (Class<?> cls = mainClass; cls != null; cls = cls.getSuperclass()) {
			if (seenSuperClasses.contains(cls))
				continue;
			seenSuperClasses.add(cls);
			TCompile a = cls.getAnnotation(TCompile.class);
			if (a != null)
				add(cls, a);
		}
	}

	private void add(Class<?> cls, TCompile a) throws CompilerException
	{
		String tpname = tpname(cls, a);
		if (!CompilerUtils.isValidJavaIdentifier(tpname)) {
			log.log(LogLevel.Error, "Invalid tpname = '" + tpname + "', must be valid java classname without package");
			return;
		}
		String fullTpname = cls.getPackage().getName() + "." + tpname;
		Class<?> prev = map.get(fullTpname);
		if (prev != null) {
			if (prev == cls)
				log.log(LogLevel.Error, "tpname = '" + tpname + "' defined multiple times in " + cls.getSimpleName());
			else
				log.log(LogLevel.Error, "tpname = '" + tpname + "' defined in both " + cls.getSimpleName() + " and " + prev.getSimpleName());
			return;
		}
		map.put(fullTpname, cls);
	}

	public void checkTpnameIsKnown(CompilationInfo info, String tpname, AnnotationHelper helper)
	{
		String fullTpname = info.scopeClass.getPackage().getName() + "." + tpname;
		if (map.containsKey(fullTpname) || warnedUnknown.contains(fullTpname))
			return;
		warnedUnknown.add(fullTpname);
		log.log(LogLevel.Error, "Unknown tpname '" + tpname + "': " + helper.annotationToString());
	}

	public static String tpname(Class<?> cls, TCompile a)
	{
		String s = a.tpname();
		return !s.isEmpty() ? s : cls.getSimpleName() + "Tp";
	}
}
