/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.codeGeneration.impl;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.jabhts.runtime.lib.Opcode;
import org.jabhts.translator.compiler.application.scope.Scope;
import org.jabhts.translator.compiler.application.scope.ScopeImpl;
import org.jabhts.translator.compiler.codeGeneration.api.IteratorInfo;
import org.jabhts.translator.compiler.codeGeneration.api.PatchedGoto;
import org.jabhts.translator.compiler.codeGeneration.api.Program;
import org.jabhts.translator.compiler.codeGeneration.api.ProgramAddress;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.InternalCompilerException;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jabhts.translator.globals.CompilerOptions;

public class ProgramImpl implements Program
{
	final CompilerOptions options;
	final CompilerLog log;
	final ScopeImpl rootScope;
	//
	final StringTable stringTable;
	final List<Integer> opcodes;
	protected boolean isPostprocessed;
	int variableCount;
	private final Set<ProgramAddressImpl> nonpatchedAddresses = new HashSet<>();

	protected StringBuilder textSB = new StringBuilder();
	protected boolean textIsCommitted = true;

	final Map<String, Callback> callbackMap = new HashMap<>();
	final Map<String, List<ProgramAddressImpl>> callbackOpcodeAddrsMap = new HashMap<>();
	final LinkedList<Callback> subScopeStack = new LinkedList<>();

	public ProgramImpl(CompilerOptions options, CompilerLog log, ScopeImpl rootScope)
	{
		this.options = options;
		this.log = log;
		this.rootScope = rootScope;
		this.stringTable = new StringTable();
		this.opcodes = new ArrayList<>();
	}

	private void trace(String msg)
	{
		if (!options.traceProgram)
			return;
		log.log(LogLevel.Verbose, "Program[" + opcodes.size() + "]: " + msg);
	}

	public ProgramAddressImpl getNextAddress() throws CompilerException
	{
		commitText();
		ProgramAddressImpl addr = internalGetNextAddress();
		if (options.traceProgram)
			trace("getNextAddress -> " + addr.idx);
		return addr;
	}

	private ProgramAddressImpl internalGetNextAddress() throws CompilerException
	{
		commitText();
		return new ProgramAddressImpl(opcodes.size());
	}

	public void addGoto(ProgramAddress destination) throws CompilerException
	{
		commitText();
		ProgramAddressImpl dst = (ProgramAddressImpl)destination;
		if (options.traceProgram)
			trace("Goto(" + dst.idx + ")");
		opcodes.add(Opcode.GOTO.encoded(dst.idx));
	}

	public void addConditionalGoto(ProgramAddress destination) throws CompilerException
	{
		commitText();
		ProgramAddressImpl dst = (ProgramAddressImpl)destination;
		if (options.traceProgram)
			trace("ConditionalGoto(" + dst.idx + ")");
		opcodes.add(Opcode.CONDITIONAL_GOTO.encoded(dst.idx));
	}

	public PatchedGotoImpl addPatchedGoto() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PatchedGoto");
		return new PatchedGotoImpl(this, makePatchedOpcodeAddr(), Opcode.GOTO);
	}

	public PatchedGoto addPatchedConditionalGoto() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PatchedConditionalGoto");
		return new PatchedGotoImpl(this, makePatchedOpcodeAddr(), Opcode.CONDITIONAL_GOTO);
	}

	public void addNegateBoolean() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("NegateBoolean");
		opcodes.add(Opcode.NEGATE.encoded());
	}

	public void addTestIsStringNull() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("TestIsStringNull");
		opcodes.add(Opcode.TEST_IS_STRING_NULL.encoded());
	}

	public void addEmitChar(char ch) throws CompilerException
	{
		checkModifiable();
		textIsCommitted = false;
		textSB.append(ch);
		lastWasLF = (ch == '\n');
	}

	public void addEmitString(String s) throws CompilerException
	{
		checkModifiable();
		if (s == null || s.isEmpty())
			return;
		textIsCommitted = false;
		textSB.append(s);
		lastWasLF = (s.endsWith("\n"));
	}

	public void addSoftNewline() throws CompilerException
	{
		checkModifiable();
		if (!lastWasLF)
			addEmitChar('\n');
		//if (!endsWithSoftNewline())
		//addEmitChar('\n');
	}

	boolean lastWasLF = true;
	// even though this spans conditionals, output is much better due to use
	// of HtmlTags.elementsPrependNewline etc.
	/*
	private boolean endsWithSoftNewline()
	{
		if (textIsCommitted)
			return stringTable.isEmpty();
		else
			return textSB.length() > 0 && textSB.charAt(textSB.length() - 1) == '\n';
	}
	*/

	public void addCallback(Scope scope, Method method) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("Callback, scope = " + scope.scopeName + ", method = " + method.getName());
		makeCallback(scope, null, method, null, Callback.CallbackType.CallMethod);
	}

	public void addCallback(IteratorInfo iterInfo, Method method) throws CompilerException
	{
		commitText();
		if (options.traceProgram) {
			trace("Callback, iter = " + iterInfo.description +
					(iterInfo.scope != null ? ", scope = " + iterInfo.scope.scopeName : "") +
					", method = " + method.getName());
		}
		makeCallback(iterInfo.scope, null, method, iterInfo, Callback.CallbackType.CallMethod);
	}

	public void addIteratorStep(IteratorInfo iterInfo) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("IteratorStep, iter = " + iterInfo.description);
		makeCallback(iterInfo.scope, null, null, iterInfo, Callback.CallbackType.IteratorStep);
	}

	public void addGetStringField(Scope scope, Field field) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("GetStringField, scope = " + scope.scopeName + ", field = " + field.getName());
		makeCallback(scope, field, null, null, Callback.CallbackType.GetStringField);
	}

	public void addGetBooleanField(Scope scope, Field field) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("GetBooleanField, scope = " + scope.scopeName + ", field = " + field.getName());
		makeCallback(scope, field, null, null, Callback.CallbackType.GetBooleanField);
	}

	public void addGetIntegerField(Scope scope, Field field) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("GetIntegerField, scope = " + scope.scopeName + ", field = " + field.getName());
		makeCallback(scope, field, null, null, Callback.CallbackType.GetIntegerField);
	}

	public void addGetIteratorField(Scope scope, Field field, IteratorInfo iter) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("GetIteratorField, scope = " + scope.scopeName + ", field = " + field.getName() + ", iter = " + iter.description);
		makeCallback(scope, field, null, iter, Callback.CallbackType.GetIteratorField);
	}

	public void addPrintHtmlEncodedStringResult() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PrintHtmlEncodedStringResult");
		opcodes.add(Opcode.PRINT_HTML_ENCODED_STRING_VAR.encoded());
	}

	public void addPrintAttrEncodedStringResult() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PrintAttrEncodedStringResult");
		opcodes.add(Opcode.PRINT_ATTR_ENCODED_STRING_VAR.encoded());
	}

	public void addPrintRawStringResult() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PrintRawStringResult");
		opcodes.add(Opcode.PRINT_STRING_VAR.encoded());
	}

	public void addPrintIntegerResult() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("PrintIntegerResult");
		opcodes.add(Opcode.PRINT_INT_VAR.encoded());
	}

	public String getNextVariableNamePrefix()
	{
		return "v" + (++variableCount);
	}

	public void addCreateScopeInstance(Scope scope) throws CompilerException
	{
		commitText();
		if (scope == null)
			throw new InternalCompilerException("scope == null");
		if (options.traceProgram)
			trace("CreateScopeInstance: " + scope.scopeName);
		if (scope == rootScope)
			throw new InternalCompilerException("scope == rootScope");
		subScopeStack.add(makeCallback(scope, null, null, null, Callback.CallbackType.EnterScope));
		traceScopeStack();
	}

	public void addCreateScopeInstance(IteratorInfo iterInfo, Scope subScope) throws CompilerException
	{
		commitText();
		if (subScope == null)
			throw new InternalCompilerException("subScope == null");
		if (options.traceProgram) {
			trace("CreateScopeInstance: iter = " + iterInfo.description + ", subScope = " + subScope.scopeName
					+ " inside " + scopeStackToString());
		}
		if (subScope == rootScope)
			throw new InternalCompilerException("subScope == rootScope");
		subScopeStack.add(makeCallback(subScope, null, null, iterInfo, Callback.CallbackType.EnterScope));
		traceScopeStack();
	}

	public void addLeaveScopeInstance(Scope scope) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("LeaveScopeInstance: " + scope.scopeName + " inside " + scopeStackToString());
		if (subScopeStack.isEmpty())
			throw new InternalCompilerException("subScopeStack is empty");
		Callback inner = subScopeStack.getLast();
		if (inner.scope != scope)
			throw new InternalCompilerException("leaving non-current scope = " + scope.scopeName + ", current = " + inner.scope.scopeName);
		makeCallback(inner.scope, null, null, inner.iterInfo, Callback.CallbackType.LeaveScope);
		subScopeStack.removeLast();
		traceScopeStack();
	}

	private void traceScopeStack()
	{
		if (!options.traceProgram)
			return;
		trace("   => scope stack: " + scopeStackToString());
	}

	private String scopeStackToString()
	{
		StringBuilder sb = new StringBuilder();
		for (Callback cb : subScopeStack) {
			if (sb.length() > 0)
				sb.append(", ");
			sb.append(cb.scope.scopeName);
		}
		if (sb.length() == 0)
			sb.append("<<empty>>");
		return sb.toString();
	}

	public void addReturn() throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("Return");
		opcodes.add(Opcode.RETURN.encoded());
	}

	protected ProgramAddressImpl makePatchedOpcodeAddr() throws CompilerException
	{
		commitText();
		ProgramAddressImpl addr = internalGetNextAddress();
		opcodes.add(-1);
		nonpatchedAddresses.add(addr);
		//log.log(LogLevel.Notice, "makePatchedOpcodeAddr: " + addr);
		return addr;
	}

	void patchToOpcode(ProgramAddressImpl addr, int encoded) throws CompilerException
	{
		commitText();
		//log.log(LogLevel.Notice, "patches " + addr.idx + " to " + encoded);
		if (!nonpatchedAddresses.contains(addr))
			throw new CompilerException("Patching non-open address");
		if (opcodes.get(addr.idx) != -1)
			throw new CompilerException("Patching open address, but has encoded = " + opcodes.get(addr.idx));
		if (!nonpatchedAddresses.remove(addr))
			throw new CompilerException("nonpatchedAddresses did not remove " + addr);
		opcodes.set(addr.idx, encoded);
	}

	void patchToCurrentAddress(ProgramAddressImpl opcodeAddr, Opcode opcode) throws CompilerException
	{
		commitText();
		if (options.traceProgram)
			trace("Patch to this address of opcode at [" + opcodeAddr.idx + "]");
		patchToOpcode(opcodeAddr, opcode.encoded(internalGetNextAddress().idx));
	}

	protected Callback makeCallback(Scope scope, Field field, Method method, IteratorInfo iteratorInfo,
	                                Callback.CallbackType callbackType) throws CompilerException
	{
		commitText();
		String key = scope.scopeFieldName + ",  " + callbackType.name();
		if (iteratorInfo != null)
			key += ",  " + iteratorInfo.itemVar;
		if (method != null)
			key += ",  " + method.getName();
		if (field != null)
			key += ",  " + field.getName();
		Callback cb = callbackMap.get(key);
		List<ProgramAddressImpl> opcodeAddrsList;
		if (cb == null) {
			cb = new Callback(scope, field, method, iteratorInfo, callbackType, key);
			callbackMap.put(key, cb);
			opcodeAddrsList = new ArrayList<>();
			callbackOpcodeAddrsMap.put(key, opcodeAddrsList);
		} else {
			opcodeAddrsList = callbackOpcodeAddrsMap.get(key);
		}
		opcodeAddrsList.add(makePatchedOpcodeAddr());
		return cb;
	}

	protected void commitText() throws CompilerException
	{
		checkModifiable();
		if (textIsCommitted)
			return;
		String s = textSB.toString();
		int idx = stringTable.addText(s);
		if (options.traceProgram) {
			String s2 = (s.length() < 100)
					? s
					: s.substring(0, 50) + "[...]" + s.substring(s.length() - 50);
			trace("text[" + idx + "] = {{{" + s2 + "}}}");
		}
		textSB.setLength(0);
		opcodes.add(Opcode.PRINT_STRINGTABLE.encoded(idx));
		textIsCommitted = true;
	}

	protected void checkModifiable() throws CompilerException
	{
		if (isPostprocessed)
			throw new InternalCompilerException("Can't modify program at this point: it is already post-processed");
	}

	public ProgramData postProcess() throws CompilerException
	{
		commitText();
		Callback[] callbacks = callbackMap.values().toArray(new Callback[callbackMap.size()]);
		Arrays.sort(callbacks, new Comparator<Callback>()
		{
			public int compare(Callback a, Callback b)
			{
				return a.key.compareTo(b.key);
			}
		});
		for (int i = 0; i < callbacks.length; i++) {
			for (ProgramAddressImpl addr : callbackOpcodeAddrsMap.get(callbacks[i].key))
				patchToOpcode(addr, Opcode.CALLBACK.encoded(i));
		}
		if (!nonpatchedAddresses.isEmpty()) {
			List<ProgramAddressImpl> list = new ArrayList<>(nonpatchedAddresses);
			list.sort(null);
			throw new CompilerException(nonpatchedAddresses.size() + " patched-goto's have not been patched: " + list);
		}
		int[] ops = new int[opcodes.size()];
		for (int idx = 0; idx < ops.length; idx++)
			ops[idx] = opcodes.get(idx);
		isPostprocessed = true;
		List<ScopeImpl> scopes = new ArrayList<>();
		getFlatScopeList(scopes, rootScope);
		scopes.sort(new Comparator<ScopeImpl>()
		{
			public int compare(ScopeImpl a, ScopeImpl b)
			{
				return a.scopeFieldName.compareTo(b.scopeFieldName);
			}
		});
		return new ProgramData(log, ops, stringTable.toArray(),
				rootScope, scopes.toArray(new ScopeImpl[scopes.size()]),
				callbacks);
	}

	protected void getFlatScopeList(List<ScopeImpl> list, ScopeImpl scope)
	{
		list.add(scope);
		for (ScopeImpl sc : scope.iterSubScopes())
			getFlatScopeList(list, sc);
	}
}
