/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.codeGeneration.util;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import javax.annotation.processing.Generated;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.jetbrains.annotations.NotNull;

public class JavaClass
{
	@NotNull public final String fullClassName, classPackage, shortClassName;
	@NotNull public final File javaF;
	@NotNull public final File javaRootF;
	@NotNull public final StringWriter sw = new StringWriter();
	@NotNull public final PrintWriter pw = new PrintWriter(sw);
	@NotNull public final Set<String> imports = new HashSet<>();
	@NotNull public final ClassType type;
	public boolean isAbstract;
	public boolean isFinal;
	@NotNull public final Set<String> classImplements = new HashSet<>();
	public String classExtends = null;
	public boolean makeAutogeneratedBanner = true;
	public String commentsBeforeClass;
	@NotNull public final List<String> classAnnotations = new ArrayList<>();

	public String encoding = "UTF-8";

	public static final String autoGenerateMarker =
			"THIS IS AUTO" + "" + "GENERATED " + "CODE"; // avoid match by fgrep


	public JavaClass(@NotNull File javaRootF,
	                 @NotNull String fullClassName,
	                 @NotNull ClassType type)
	{
		this.javaRootF = javaRootF;
		this.type = type;
		this.fullClassName = fullClassName;
		int idx = fullClassName.lastIndexOf('.');
		this.shortClassName = fullClassName.substring(idx + 1);
		this.classPackage = fullClassName.substring(0, idx);
		this.javaF = new File(javaRootF, this.fullClassName.replace('.', '/') + ".java");
	}

	public void addGeneratedAnnotation(@NotNull String generatorPackage)
	{
		String s = "@Generated(\"" + generatorPackage + "\")";
		if (classAnnotations.contains(s))
			return;
		addImport(Generated.class); // 		javax.annotation.processing.Generated;
		classAnnotations.add(s);
	}

	public void addImport(Class<?> c)
	{
		imports.add(c.getName());
	}

	public String getCode() throws Exception
	{
		pw.flush();
		StringBuilder sb = new StringBuilder(500);
		getCopyright(sb);
		sb.append("package ").append(classPackage).append(";\n\n");
		sb.append("/*\n");
		sb.append(" *\n");
		if (makeAutogeneratedBanner) {
			sb.append(" *    " + autoGenerateMarker + "\n");
			sb.append(" *\n");
			sb.append(" *    DO NOT EDIT.\n");
		} else {
			sb.append(" *    This is just a skeleton").append(" for you to edit.\n"); // avoid match by fgrep
			sb.append(" *    It will not be overwritten").append(" by the generator.\n");
		}
		sb.append(" *\n");
		sb.append(" */\n\n");
		if (!imports.isEmpty()) {
			String[] imps = imports.toArray(new String[0]);
			Arrays.sort(imps, JavaImportComparator.comparator);
			for (String s : imps) {
				int idx = s.lastIndexOf('.');
				if (idx > 0 && s.substring(0, idx).equals(classPackage))
					continue;
				sb.append("import ").append(s).append(";\n");
			}
			sb.append("\n");
		}
		if (commentsBeforeClass != null && !commentsBeforeClass.isEmpty()) {
			sb.append("/*\n");
			sb.append(commentsBeforeClass);
			sb.append("\n*/\n\n");
		}
		for (String s : classAnnotations)
			sb.append(s).append("\n");
		sb.append("public ");
		if (isAbstract)
			sb.append("abstract ");
		if (isFinal)
			sb.append("final ");
		switch (type) {
			case Class -> sb.append("class");
			case Enum -> sb.append("enum");
			case Interface -> sb.append("interface");
			default -> throw new Exception("Unknown class type: " + type);
		}
		sb.append(" ").append(shortClassName);
		if (classExtends != null) {
			sb.append(" extends ");
			sb.append(classExtends);
		}
		sb.append("\n");
		if (!classImplements.isEmpty()) {
			String sep = "\t\timplements ";
			for (String s : classImplements) {
				sb.append(sep);
				sep = ", ";
				sb.append(s);
			}
			sb.append("\n");
		}
		sb.append("{\n");
		if (classPackage.equals("just-disabled-here"))
			sb.append("/*\n");
		sb.append(sw);
		sb.append("\n");
		sb.append("// magic comment */\n");
		sb.append("}\n");
		return sb.toString();
	}

	protected void getCopyright(StringBuilder sb) throws IOException
	{
		File dirF = javaF.getParentFile();
		while (dirF != null) {
			File f = new File(dirF, "copyright-template.txt");
			if (f.exists()) {
				String s = new String(FileUtils.readFile(f), encoding).trim();
				sb.append(s);
				sb.append("\n\n");
				return;
			}
			dirF = dirF.getParentFile();
		}
	}


	public static String javaNameFromSqlName(String s)
	{
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		boolean uc = true;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '_') {
				uc = true;
			} else {
				sb.append(uc ? Character.toUpperCase(ch) : ch);
				uc = false;
			}
		}
		return sb.toString();
	}

	/**
	 * @return true if file is created/updated
	 */
	public boolean save() throws Exception
	{
		String newCode = getCode();
		boolean exists = javaF.exists();
		if (exists) {
			String oldCode = new String(FileUtils.readFile(javaF), encoding);
			if (newCode.equals(oldCode))
				return false;
			if (!oldCode.contains(JavaClass.autoGenerateMarker)) {
				throw new IOException("Trying to overwrite" +
						"non-autogenerated java source " +
						"with auto-generated code: " + javaF.getPath());
			}
		}
		FileUtils.saveFile(javaF, newCode.getBytes(encoding));
		return true;
	}

	public void saveJavaClass() throws Exception
	{
		if (save())
			System.err.println("  -- Updated " + javaF.getName());
	}
}
