/* Copyright (c) 2007-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.utils;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

public class FileUtils
{
	/**
	 * Filename without the suffix
	 */
	public static String basename(String filename)
	{
		int i = filename.lastIndexOf('.');
		if (i >= Math.max(0, filename.length() - 7))
			filename = filename.substring(0, i);
		return filename;
	}

	/**
	 * Filename without the suffix
	 */
	public static String basename(File f)
	{
		return basename(f.getName());
	}


	/**
	 * @return the complete file contents
	 */
	public static byte[] readFile(File f) throws IOException
	{
		if (!f.isFile())
			throw new FileNotFoundException("Missing file: " + f.getPath());
		long size = f.length();
		if (size > 1024L * 1024L * 1024L)
			throw new IOException("File too big to read into memory: " + size + " bytes");
		byte[] data = new byte[(int)size];
		FileInputStream fis = new FileInputStream(f);
		try {
			int offset = 0;
			while (offset < data.length) {
				int num = fis.read(data, offset, data.length - offset);
				if (num < 0)
					throw new EOFException("End of stream reached prematurely.");
				offset += num;
			}
		}
		finally {
			fis.close();
		}
		return data;
	}

	/**
	 * Saves the data in the file.
	 */
	public static void saveFile(File f, byte[] data) throws IOException
	{
		File tmpF = new File(f.getParentFile(), f.getName() + "-" + System.currentTimeMillis() + ".tmp");
		FileOutputStream fos = new FileOutputStream(tmpF);
		try {
			fos.write(data);
		}
		catch (IOException ex) {
			fos.close();
			if (!tmpF.delete())
				System.err.println("Can't delete " + tmpF.getPath());
			throw ex;
		}
		finally {
			fos.close();
		}
		// Must delete file and not just overwrite it using renameTo() as
		// Intellij IDEA on Mac OS X doesn't detect filechanges otherwise...
		if (f.exists() && !f.delete())
			throw new IOException("Can't delete " + f.getPath());
		if (!tmpF.renameTo(f)) {
			if (!tmpF.delete())
				System.err.println("Can't delete " + tmpF.getPath());
			throw new IOException("Can't rename " + tmpF.getPath() + " to " + f.getPath());
		}
	}
}
