/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.exception;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.SocketException;
import java.util.regex.Pattern;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Utilities for handling Exceptions.
 */
public class ExceptionUtils
{
	/**
	 * Get a short and simple description of the Exception:
	 * returns the exception's message (if non-empty), otherwise the class name.
	 * The description is stripped of "java.lang" and "java.lang.Exception" prefixes.
	 */
	public static String getExMsg(@Nullable Throwable t)
	{
		if (t == null)
			return "(exception object is null)";
		String msg = t.getMessage();
		if (msg == null || msg.isEmpty())
			msg = javaPackageRE.matcher(t.getClass().getName()).replaceAll("");
		return msg;
	}

	private static final Pattern javaPackageRE = Pattern.compile("java\\.lang\\.(Exception)?");

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	public static String getStackTraceAsString()
	{
		return getStackTraceAsString(new Exception());
	}

	public static String getStackTraceAsString(@Nullable Throwable ex)
	{
		if (ex == null)
			return "(exception object is null)";
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		ex.printStackTrace(pw);
		pw.flush();
		return sw.toString();
	}

	public static boolean hasCause(@Nullable Throwable t, @NotNull Class<? extends Throwable> cause)
	{
		while (t != null) {
			if (cause.isAssignableFrom(t.getClass()))
				return true;
			t = t.getCause();
		}
		return false;
	}

	public static boolean isBrokenPipeException(Throwable t)
	{
		if (!(t instanceof SocketException))
			return false;
		String msg = t.getMessage();
		if (msg == null)
			return false;
		if (!msg.equals("Broken pipe") && !msg.equals("Connection reset"))
			return false;
//		System.err.println("Got broken pipe: ");
//		t.printStackTrace();
		return true;
	}
}
