/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.http;

import org.jetbrains.annotations.Nullable;

public class AcceptEncodingUtils
{
	/**
	 * OBS: IE6 doesn’t like compressed js files..
	 * http://paulstamatiou.com/how-to-optimize-your-apache-site-with-mod-deflate
	 */
	public static boolean allowsGzip(@Nullable String acceptEncodingHeader)
	{
		String s = acceptEncodingHeader;
		if (s == null || s.isEmpty())
			return false;
		//
		// Safari 4 sends: "gzip, deflate"
		// Firefox 3.5 sends: "gzip,deflate"
		// Chrome 2.0 sends "gzip,deflate,sdch"
		// MSIE 7 sends "gzip, deflate"
		// Opera 9.64 sends "deflate, gzip, x-gzip, identity, *;q=0"
		//
		if (s.equals("gzip") || s.startsWith("gzip,"))
			return true;
		s = s.toLowerCase();
		for (int idx = s.indexOf("gzip"); idx >= 0; idx = s.indexOf("gzip", idx + 1)) {
			if (idx > 0) {
				char ch = s.charAt(idx - 1);
				if (ch != ' ' && ch != '\t' && ch != ',' && ch != ';')
					continue;
			}
			if (idx < s.length() - 4) {
				char ch = s.charAt(idx + 4);
				if (ch != ' ' && ch != '\t' && ch != ',' && ch != ';')
					continue;
			}
			//System.err.println("Found gzip at " + idx + " in '" + s + "'");
			return true;
		}
		//System.err.println("accept-encoding without gzip: '" + s + "'\n   for user-agent: '" + getHeader("User-Agent") + "'");
		return false;
	}

/*
		http://www.mail-archive.com/libwww@perl.org/msg04150.html
		RE: Net::HTTP does not use compressed transfers when it should
		Paul Marquess
		Tue, 25 Mar 2003 06:30:58 -0800

		You need to be very careful with HTTP deflate encoding. Due to a common
		misreading of the HTTP spec, there are two distinct implementations of HTTP
		deflate out in the wild. The correct implementation uses RFC 1950. The
		incorrect implementation uses RFC 1951. The reason for this seems to be down
		to the overloaded use of the term "deflate". Both zlib and HTTP use the
		term, but unfortunately they don't mean the same thing. The incorrect
		implementations use the zlib definition of deflate.

		Last time I checked, Internet Explorer does it the wrong way. Opera can cope
		with either. Older Netscape/Mozilla got it wrong, but I think recent Mozilla
		has fixed it to handle both. I can't remember what Apache does.

		This makes things slightly tricky if you are going to process content that
		advertises itself as "Content-Encoding: deflate", because you can't tell
		which of the two deflate implementations it is without doing a quick test on
		the content. It is possible to do it though. You can test for RFC 1950 by
		checking for the presence of a ZLIB header. To detect RFC 1951, you have to
		attempt to decompress it. If it passes, you have RFC1951, if it fails, you
		don't.
		[...]

		----------------------------------------------------------------------------
		http://www.mail-archive.com/libwww@perl.org/msg04155.html
		Re: Net::HTTP does not use compressed transfers when it should
		Mike Simons
		Tue, 25 Mar 2003 22:16:17 -0800

		rfc1950 documents the 'ZLIB' format,
		  which is a 2 or 6 byte header, and 4 byte trailer,
		  around a DEFLATE data stream.

		rfc1951 documents the 'DEFLATE' format (a compression algorithm)
		  which has no header or trailer

		rfc1952 documents the 'GZIP' format,
		  which is a 10 or more byte header (lots of flexiblilty), an 8 byte
		  trailer, around a DEFLATE data stream.

		rfc2616 documents the HTTP/1.1 protocol,
		  it identifies three content-encodings key words:
			"gzip" == rfc1952, "compress" (no reference),
			"deflate" which is ZLIB format (rfc1950 format).

		----------------------------------------------------------------------------
		Sites that support deflate:

		curl  -H 'Accept-Encoding: deflate' --silent --dump-header /dev/stderr \
		'http://www.brazilianfightwear.com/store/stores_app/store.asp?Store_Id=117&page_id=5' \
		> /dev/null

		Server: Microsoft-IIS/6.0
		Content-Encoding: deflate


	 */
}
