/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.xml.xpath;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Pattern;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Attr;
import org.w3c.dom.CharacterData;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import dk.vitality.lib.foundation.misc.CollectionUtils;

/**
 * A simple layer on top of XPath which provides convenience access methods
 * to 'arguments' in an xml document.<br>
 * http://www.w3.org/TR/xpath <p>
 * <p/>
 * The value of a node is the attribute value for attributes, and for all other
 * node types: the concatenated string of all text node children.<p>
 * <p/>
 * The 'Opt' methods returns null (or the specified default value) in case the
 * parameter isn't specified.<p>
 * <p/>
 * The 'Req' methods throws an SAXException is the parameter isn't specfied
 * or is empty.<p>
 * <p/>
 * Example document:<br><code>
 * &lt;person&gt;<br>
 * &nbsp;&nbsp;&nbsp;	&lt;name&gt;Hobbes&lt;/name&gt;<br>
 * &nbsp;&nbsp;&nbsp;	&lt;height metric="cm"&gt;150&lt;/height&gt;<br>
 * &nbsp;&nbsp;&nbsp;	&lt;height metric="m"&gt;1.5&lt;/height&gt;<br>
 * &lt;/person&gt;</code></p>
 * <p/>
 * The document root node and the document element node is not the same node!
 * In the above example, the document node is an instance of "Document" and is
 * the parent to the document element which is an Element node named "person".
 * So the document root node sits one level above the document element node!
 * <p/>
 * If XmlArgs is constructed using the Document as argument, it uses the Document Element
 * as context element.<p>
 * <p/>
 * Constructing XmlArgs with the 'person' node as argument to the constructor, you get:<ul>
 * <li>	 getReqString("name") returns 'Hobbes'
 * <li>	 getReqString("height[@metric='cm']") returns '150'
 * <li>	 getReqString("height") throws an exception because there are more than one element
 * <li>	 getStringArray("height/@metric") returns {"cm", "m"}
 * </ul><p>
 */

public class XmlArgs
{
	protected Document doc;
	protected Node contextNode;
	private static int serial_g = 100;
	private final int serial = ++serial_g;
	//
	private static final Map<String, PathProgram> cache = new HashMap<>();

	/**
	 * enables logging of all getXXX() method calls.
	 */
	public static final boolean debugResolveFlag = false;
	/**
	 * enables statistics (timings and usage counts)
	 */
	public static final boolean enableStatisticsFlag = false;
	/**
	 * enables logging of compiled and cached paths
	 */
	public static final boolean debugCacheFlag = false;
	/**
	 * which implementation
	 */
	public static XmlArgsErrorLogger errorLogger = new DummyErrorLogger();


	protected static final Map<String, StatisticsEntry> statistics = new HashMap<>();

	/**
	 * If the argument is a document, then the context node is set
	 * to the document element, not the document root!
	 * This avoids having to know the name of the document root
	 * for most paths.
	 */
	public static XmlArgs create(Node theContextNode) //throws SAXException
	{
		return new XmlArgs(theContextNode);
	}

	protected static XmlArgs create(Document theDoc, Node theContextNode) throws SAXException
	{
		return new XmlArgs(theDoc, theContextNode);
	}

	public static String[] getListOfCachedXPaths() //throws SAXException
	{
		synchronized (cache) {
			Set<String> keys = cache.keySet();
			return keys.toArray(new String[keys.size()]);
		}
	}

	public static int getCacheLineSize(@NotNull String xpath) //throws SAXException
	{
		PathProgram pp;
		synchronized (cache) {
			pp = cache.get(xpath);
		}
		return pp != null ? pp.numTimesExecuted : 0;
	}

	public XmlArgs(Node theContextNode)
	{
		try {
			if (theContextNode == null)
				throw new NullPointerException("null Node passed to XmlArgs constructor");
			if (theContextNode instanceof Document) {
				Document d = (Document)theContextNode;
				initXmlArgs(d, d.getDocumentElement());
			} else {
				initXmlArgs(theContextNode.getOwnerDocument(), theContextNode);
			}
		}
		catch (NullPointerException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	public XmlArgs(Document theDoc, Node theContextNode) throws SAXException
	{
		if (theContextNode == null)
			throw makeSAXException("Can't do with null context node", null);
		if (theDoc == null)
			throw makeSAXException("Can't do with null document node", null);
		initXmlArgs(theDoc, theContextNode);
	}

	private void initXmlArgs(Document theDoc, Node theContextNode)
	{
		doc = theDoc;
		contextNode = theContextNode;
		//noinspection PointlessBooleanExpression
		if (debugResolveFlag || debugCacheFlag)
			System.err.println("XmlArgs[" + serial + "].created: " + contextNode);
	}

	public Document getDocument() throws SAXException
	{
		return doc;
	}

	public Node getContextNode()
	{
		return contextNode;
	}

	public Element getContextElement()
	{
		return (Element)contextNode;
	}

	public SAXException makeSAXException(@NotNull String msg, @Nullable String xpath) throws SAXException
	{
		if (doc != null) {
			System.err.println("XmlArgs[" + serial + "] error: " + msg);
			if (xpath != null)
				System.err.println("		xpath: " + xpath);
			reportCurrentPosition();
		}
		if (xpath != null)
			msg += ", xpath = '" + xpath + "'";
		return new SAXException(msg);
	}

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	public void reportCurrentPosition() throws SAXException
	{
		try {
// System.err.println("XmlArgs Document: --------------");
// dp.printToSystemErr(doc);
			errorLogger.logContextNodePath(this);
			new Exception().printStackTrace();
			System.err.println("---------------------------");
		}
		catch (SAXException sax) {
			throw sax;
		}
		catch (Exception ex) {
			throw new SAXException(ex);
		}
	}

	protected void log(@Nullable String what, @Nullable String xpath)
	{
		System.err.println("XmlArgs[" + serial + "]." + what + ": " + xpath);
	}

	protected void logAction(@Nullable String action, @Nullable String xpath)
	{
		log(action, xpath);
	}

	protected void logReturn(@Nullable String action, @Nullable String returnValue)
	{
		String r = (returnValue != null) ? "'" + returnValue + "'" : "null";
		log(action, "returns " + r);
	}

	public void gotoReqElement(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("gotoReqElement", xpath);
		contextNode = getReqElement(xpath);
		if (debugResolveFlag)
			logReturn("gotoReqElement", contextNode.toString());
	}

	@Nullable
	protected Node doGetOptNode(@NotNull String xpath) throws SAXException
	{
		List<Element> list = doGetElementArrayList(xpath);
		if (list == null)
			return null;
		int num = list.size();
		if (num > 1) {
			errorLogger.didntExpectMultipleNodes(this, xpath, list);
			throw makeSAXException("Matched " + num + " nodes in xml document", xpath);
		}
		if (num == 0)
			return null;
		return list.get(0);
	}

	@Nullable
	public Node getOptNode(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptNode", xpath);
		Node n = doGetOptNode(xpath);
		if (debugResolveFlag)
			logReturn("getOptNode", n != null ? n.toString() : null);
		return n;
	}

	@Nullable
	protected Element doGetOptElement(@NotNull String xpath) throws SAXException
	{
		Node n = doGetOptNode(xpath);
		if (n != null && !(n instanceof Element))
			throw makeSAXException("Found non-element node", xpath);
		return (Element)n;
	}

	@Nullable
	public Element getOptElement(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptElement", xpath);
		Element e = doGetOptElement(xpath);
		if (debugResolveFlag)
			logReturn("getOptElement", e != null ? e.toString() : null);
		return e;
	}

	@NotNull
	public Element getReqElement(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqElement", xpath);
		Element e = getOptElement(xpath);
		if (e == null) {
			if (1 == 1)
				errorLogger.didntGetRequiredElement(this, xpath);
			throw makeSAXException("Missing required element", xpath);
		}
		if (debugResolveFlag)
			logReturn("getReqElement", e.toString());
		return e;
	}

	@Nullable
	public XmlArgs getOptXmlArgs(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptXmlArgs", xpath);
		Node n = doGetOptElement(xpath);
		if (debugResolveFlag)
			logReturn("getOptXmlArgs", n != null ? n.toString() : null);
		return n != null ? new XmlArgs(doc, n) : null;
	}

	@NotNull
	public XmlArgs getReqXmlArgs(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptXmlArgs", xpath);
		Node n = doGetOptElement(xpath);
		if (debugResolveFlag)
			logReturn("getOptXmlArgs", n != null ? n.toString() : null);
		if (n == null)
			throw makeSAXException("Missing required element", xpath);
		return new XmlArgs(doc, n);
	}

	@Nullable
	protected String doGetOptString(@NotNull String xpath) throws SAXException
	{
		String[] array = getStringArray(xpath);
		if (array.length == 0)
			return null;
		if (array.length > 1)
			throw makeSAXException("Matched " + array.length + " nodes in xml document", xpath);
		String s = array[0];
		if (s == null || s.isEmpty())
			return null;
		else
			return s;
	}

	@Nullable
	public String getOptString(@NotNull String xpath) throws SAXException
	{
		return getOptString(xpath, null);
	}

	@Nullable
	public String getOptString(@NotNull String xpath, @Nullable String defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptString", xpath);
		String s = doGetOptString(xpath);
		if (debugResolveFlag)
			logReturn("getOptString", s);
		if (s == null || s.isEmpty())
			return defaultValue;
		else
			return s;
	}

	@NotNull
	public String getOptNonNullString(@NotNull String xpath, @NotNull String defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptNonNullString", xpath);
		String s = doGetOptString(xpath);
		if (debugResolveFlag)
			logReturn("getOptNonNullString", s);
		if (s == null || s.isEmpty())
			return defaultValue;
		else
			return s;
	}

	@NotNull
	public String getOptEmptyString(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptString", xpath);
		String s = doGetOptString(xpath);
		if (debugResolveFlag)
			logReturn("getOptString", s);
		if (s == null || s.isEmpty())
			return "";
		else
			return s;
	}

	@NotNull
	public String doGetReqString(@NotNull String xpath) throws SAXException
	{
		String s = getOptString(xpath, null);
		if (s == null || s.isEmpty())
			throw makeSAXException("Missing/empty required element", xpath);
		return s;
	}

	@NotNull
	public String getReqString(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqString", xpath);
		String s = doGetReqString(xpath);
		if (debugResolveFlag)
			logReturn("getReqString", s);
		return s;
	}

	public int getReqInt(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqInt", xpath);
		int i = parseInt(doGetReqString(xpath), xpath);
		if (debugResolveFlag)
			logReturn("getReqInt", Integer.toString(i));
		return i;
	}

	public int getOptInt(@NotNull String xpath, int defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptInt", xpath);
		String s = getOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getOptInt", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		} else {
			int i = parseInt(s, xpath);
			if (debugResolveFlag)
				logReturn("getOptInt", Integer.toString(i));
			return i;
		}
	}

	public long getReqLong(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqLong", xpath);
		long i = parseLong(doGetReqString(xpath), xpath);
		if (debugResolveFlag)
			logReturn("getReqLong", Long.toString(i));
		return i;
	}

	public long getOptLong(@NotNull String xpath, long defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptLong", xpath);
		String s = getOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getOptLong", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		} else {
			long i = parseLong(s, xpath);
			if (debugResolveFlag)
				logReturn("getOptLong", Long.toString(i));
			return i;
		}
	}

	public float getReqFloat(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqFloat", xpath);
		float f = (float)parseDouble(doGetReqString(xpath), xpath);
		if (debugResolveFlag)
			logReturn("getReqFloat", Float.toString(f));
		return f;
	}

	public float getOptFloat(@NotNull String xpath, float defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptFloat", xpath);
		String s = doGetOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getReqFloat", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		} else {
			float f = (float)parseDouble(s, xpath);
			if (debugResolveFlag)
				logReturn("getReqFloat", Float.toString(f));
			return f;
		}
	}

	public double getReqDouble(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqDouble", xpath);
		double d = parseDouble(doGetReqString(xpath), xpath);
		if (debugResolveFlag)
			logReturn("getReqDouble", Double.toString(d));
		return d;
	}

	public double getOptDouble(@NotNull String xpath, double defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptDouble", xpath);
		String s = doGetOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getReqDouble", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		} else {
			double d = parseDouble(s, xpath);
			if (debugResolveFlag)
				logReturn("getReqDouble", Double.toString(d));
			return d;
		}
	}

	@NotNull
	public BigDecimal getReqBigDecimal(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqBigDecimal", xpath);
		BigDecimal d = parseBigDecimal(doGetReqString(xpath), xpath);
		if (debugResolveFlag)
			logReturn("getReqBigDecimal", d.toString());
		return d;
	}

	@Nullable
	public BigDecimal getOptBigDecimal(@NotNull String xpath, @Nullable BigDecimal defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptBigDecimal", xpath);
		String s = doGetOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getReqBigDecimal", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		} else {
			BigDecimal d = parseBigDecimal(s, xpath);
			if (debugResolveFlag)
				logReturn("getReqBigDecimal", d.toString());
			return d;
		}
	}

	public boolean getReqBoolean(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getReqBoolean", xpath);
		String s = doGetReqString(xpath);
		boolean value = s.equals("true") || s.equals("y") || s.equals("1");
		if (debugResolveFlag)
			logReturn("getReqBoolean", value ? "true" : "false");
		return value;
	}

	public boolean getOptBoolean(@NotNull String xpath) throws SAXException
	{
		return getOptBoolean(xpath, false);
	}

	public boolean getOptBoolean(@NotNull String xpath, boolean defaultValue) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getOptBoolean", xpath);
		String s = doGetOptString(xpath);
		if (s == null) {
			if (debugResolveFlag)
				logReturn("getOptBoolean", "defaultValue(" + defaultValue + ")");
			return defaultValue;
		}
		boolean value = !s.equalsIgnoreCase("false") && !s.equalsIgnoreCase("n") && !s.equalsIgnoreCase("0");
		if (debugResolveFlag)
			logReturn("getOptBoolean", value ? "true" : "false");
		return value;
	}

	public boolean exists(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("exists", xpath);
		boolean value = count(xpath) > 0;
		if (debugResolveFlag)
			logReturn("exists", "value = " + value);
		return value;
	}

	public int count(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getNodeCount", xpath);
		PathRuntime r = getRuntime(xpath, PathRuntime.Interest.Count);
		if (debugResolveFlag)
			logReturn("getNodeCount", Integer.toString(r.matchCount));
		return r.matchCount;
	}

	@Nullable
	protected List<Element> doGetElementArrayList(@NotNull String xpath) throws SAXException
	{
		PathRuntime r = getRuntime(xpath, PathRuntime.Interest.Element);
		return r.elementResult;
	}

	@NotNull
	public List<Element> getElementList(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getElementArray", xpath);
		List<Element> list = CollectionUtils.unmodifiableOrEmptyList(doGetElementArrayList(xpath));
		if (debugResolveFlag)
			logReturn("getElementList", list.size() + " elements");
		return list;
	}

	@NotNull
	public Element[] getElementArray(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getElementArray", xpath);
		List<Element> list = doGetElementArrayList(xpath);
		Element[] array;
		if (list == null) {
			array = emptyElementArray;
		} else {
			array = new Element[list.size()];
			list.toArray(array);
		}
		if (debugResolveFlag)
			logReturn("getElementArray", array.length + " elements");
		return array;
	}

	@NotNull private static final Element[] emptyElementArray = new Element[0];

	@NotNull
	public List<String> getStringList(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getStringArray", xpath);
		PathRuntime r = getRuntime(xpath, PathRuntime.Interest.String);
		List<String> list = (r.stringResult != null) ? r.stringResult : Collections.<String>emptyList();
		if (debugResolveFlag)
			logReturn("getStringList", list.size() + " strings");
		return list;
	}

	@NotNull
	public String[] getStringArray(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getStringArray", xpath);
		PathRuntime r = getRuntime(xpath, PathRuntime.Interest.String);
		if (r.stringResult == null)
			return emptyStringArray;
		String[] array = new String[r.stringResult.size()];
		r.stringResult.toArray(array);
		if (debugResolveFlag)
			logReturn("getStringArray", array.length + " strings");
		return array;
	}

	private static final String[] emptyStringArray = new String[0];

	@NotNull
	public List<XmlArgs> getXmlArgsList(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getXmlArgsList", xpath);
		List<Element> eList = doGetElementArrayList(xpath);
		List<XmlArgs> xaList;
		if (eList == null) {
			xaList = Collections.emptyList();
		} else {
			xaList = new ArrayList<>(eList.size());
			for (Element e : eList)
				xaList.add(new XmlArgs(e));
		}
		if (debugResolveFlag)
			logReturn("getXmlArgsList", xaList.size() + " XmlArgs");
		return xaList;
	}

	@NotNull
	public XmlArgs[] getXmlArgsArray(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getXmlArgsArray", xpath);
		List<Element> list = doGetElementArrayList(xpath);
		XmlArgs[] array;
		if (list == null) {
			array = emptyXmlArgsArray;
		} else {
			int num = list.size();
			array = new XmlArgs[num];
			for (int i = 0; i < num; i++)
				array[i] = new XmlArgs(doc, list.get(i));
		}
		if (debugResolveFlag)
			logReturn("getXmlArgsArray", array.length + " XmlArgs");
		return array;
	}

	private static final XmlArgs[] emptyXmlArgsArray = new XmlArgs[0];

	@SuppressWarnings({"ManualArrayToCollectionCopy"})
	@NotNull
	public String[] getSortedUniqueStrings(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("getSortedUniqueStrings", xpath);
		String[] all = getStringArray(xpath);
		int numStrings = all.length;
		if (numStrings > 1) {
			TreeSet<String> tree = new TreeSet<>(usingCase);
			for (String s : all)
				tree.add(s);
			all = new String[tree.size()];
			Iterator<String> treeIter = tree.iterator();
			for (int i = 0; treeIter.hasNext(); i++)
				all[i] = treeIter.next();
		}
		if (debugResolveFlag)
			logReturn("getSortedUniqueStrings", numStrings + " total, " + all.length + " unique");
		return all;
	}

	public boolean hasNonEmptyArg(@NotNull String xpath) throws SAXException
	{
		if (debugResolveFlag)
			logAction("hasNonEmptyArg", xpath);
		boolean value = doGetOptString(xpath) != null;
		if (debugResolveFlag)
			logReturn("hasNonEmptyArg", value ? "true" : "false");
		return value;
	}

	/**
	 * Dette er en optimeret udgave af xpath'en "//tagName", som xalan
	 * er utroligt langsom til.
	 * xalan: 3171 msecs.
	 * denne kode: 5 msecs.
	 * For et enkelt opslag med en enkelt side...
	 */
	protected void getNodeArrayOfChildNodesWithNameLooper(Node parentN,
	                                                      String tagName,
	                                                      List<Node> items) throws SAXException
	{
		if (tagName == null || parentN.getNodeName().equals(tagName))
			items.add(parentN);
		Node n = parentN.getFirstChild();
		while (n != null) {
			if (n.getNodeType() == Node.ELEMENT_NODE)
				getNodeArrayOfChildNodesWithNameLooper(n, tagName, items);
			n = n.getNextSibling();
		}
	}

	@NotNull
	public static String getStatisticsHtml()
	{
		//noinspection PointlessBooleanExpression
		if (!enableStatisticsFlag)
			return "<P>Statistics are not enabled.</P>";
		StringBuilder sb;
		synchronized (statistics) {
			Set<String> keySet = statistics.keySet();
			if (keySet.isEmpty())
				return "<P>No xpaths executed yet.</P>";
			sb = new StringBuilder(500);
			sb.append("<table border=1>\n"
					+ "<tr>\n"
					+ "<th colspan=3>compile</th>"
					+ "<th colspan=3>exec</th>"
					+ "</tr>\n"
					+ "<tr>\n"
					+ "<th>num</th>"
					+ "<th>avag</th>"
					+ "<th>sum</th>"
					+ "<th>num</th>"
					+ "<th>avag</th>"
					+ "<th>sum</th>"
					+ "<th>xpath</th>"
					+ "</tr>\n");
			String[] xpaths = new String[keySet.size()];
			keySet.toArray(xpaths);
			Arrays.sort(xpaths, usingCase);
			int sumTimesCompiled = 0, sumTimesExecuted = 0;
			long sumSumCompile = 0, sumSumExecute = 0;
			for (String xpath : xpaths) {
				StatisticsEntry e = statistics.get(xpath);
				if (e == null)
					continue;
				sb.append("<tr>");
				//
				sb.append("<td align=right>");
				sb.append(e.numTimesCompiled);
				sumTimesCompiled += e.numTimesCompiled;
				sb.append("</td>");
				sb.append("<td align=right>");
				sb.append(formatAvagMilliSecs(e.sumCompileTime, e.numTimesCompiled));
				sb.append("</td>");
				sb.append("<td align=right>");
				sb.append(e.sumCompileTime);
				sb.append("</td>");
				//
				sb.append("<td align=right>");
				sb.append(e.numTimesExecuted);
				sumTimesExecuted += e.numTimesExecuted;
				sb.append("</td>");
				sb.append("<td align=right>");
				sb.append(formatAvagMilliSecs(e.sumExecutionTime, e.numTimesExecuted));
				sb.append("</td>");
				sb.append("<td align=right>");
				sb.append(e.sumExecutionTime);
				sb.append("</td>");
				//
				sb.append("<td>");
				for (int i = 0; i < xpath.length(); i++) {
					char ch = xpath.charAt(i);
					switch (ch) {
						case '<':
							sb.append("lt;");
							break;
						case '>':
							sb.append("gt;");
							break;
						case '&':
							sb.append("&amp;");
							break;
						default:
							sb.append(ch);
					}
				}
				sb.append("</td>");
				sb.append("</tr>\n");
				sumSumCompile += e.sumCompileTime;
				sumSumExecute += e.sumExecutionTime;
			}
			sb.append("<tr><td height=4></td></tr>");
			sb.append("<tr>");
			//
			sb.append("<td align=right>").append(sumTimesCompiled).append("</td><td></td>");
			sb.append("<td align=right>").append(sumSumCompile).append("</td>");
			//
			sb.append("<td align=right>").append(sumTimesExecuted).append("</td><td></td>");
			sb.append("<td align=right>").append(sumSumExecute).append("</td>");
			//
			sb.append("</td><td>SUM</td>");
			sb.append("</tr>\n");
		}
		sb.append("</table>\n");
		sb.append("<P>Flags:<BR>\n");
		sb.append(" &nbsp; &nbsp; debugResolveFlag = " + debugResolveFlag + "<BR>\n");
		sb.append(" &nbsp; &nbsp; debugCacheFlag = " + debugCacheFlag + "<BR>\n");
		sb.append(" &nbsp; &nbsp; enableStatisticsFlag = " + enableStatisticsFlag + "<BR>\n");
		sb.append("</P>");
		return sb.toString();
	}

	private static String formatAvagMilliSecs(long sumTime, long numTimes)
	{
		long t = (sumTime * 100) / numTimes;
		String s = Long.toString(t);
		int len = s.length();
		if (len == 1)
			return "0.0" + s;
		if (len == 2)
			return "0." + s;
		return s.substring(0, len - 2) + "." + s.substring(len - 2);
	}

	protected boolean isSimpleAttributePath(@NotNull String xpath)
	{
		return isSimplePrefixedToken(xpath, "@");
	}

	protected boolean isSimplePrefixedToken(@NotNull String xpath, @NotNull String prefix)
	{
		if (!xpath.startsWith(prefix))
			return false;
		int startIndex = prefix.length();
		int len = xpath.length();
		if (len < startIndex + 1)
			return false;
		for (int i = startIndex; i < len; i++) {
			char ch = xpath.charAt(i);
			if (ch >= 'a' && ch <= 'z')
				continue;
			if (ch >= 'A' && ch <= 'Z')
				continue;
			if (ch >= '0' && ch <= '9')
				continue;
			if (ch == '-' || ch == '_')
				continue;
			return false;
		}
		return true;
	}

	protected int parseInt(@Nullable String s, @NotNull String xpath) throws NumberFormatException
	{
		if (s == null)
			throw new NullPointerException("integer-string is null from xpath '" + xpath + "'");
		if (s.isEmpty())
			throw new NullPointerException("integer-string is zero length from xpath '" + xpath + "'");
		try {
			return Integer.parseInt(s);
		}
		catch (NumberFormatException ex) {
			String msg = getExMsg(ex);
			//noinspection ThrowInsideCatchBlockWhichIgnoresCaughtException
			throw new NumberFormatException("Invalid integer: '" + s + "' from xpath '" + xpath + "': " + msg);
		}
	}

	protected long parseLong(@Nullable String s, @NotNull String xpath) throws NumberFormatException
	{
		if (s == null)
			throw new NullPointerException("long-string is null from xpath '" + xpath + "'");
		if (s.isEmpty())
			throw new NullPointerException("long-string is zero length from xpath '" + xpath + "'");
		try {
			return Long.parseLong(s);
		}
		catch (NumberFormatException ex) {
			String msg = getExMsg(ex);
			//noinspection ThrowInsideCatchBlockWhichIgnoresCaughtException
			throw new NumberFormatException("Invalid long: '" + s + "' from xpath '" + xpath + "': " + msg);
		}
	}

	protected double parseDouble(@Nullable String s, @NotNull String xpath) throws NumberFormatException
	{
		if (s == null)
			throw new NullPointerException("double-string is null from xpath '" + xpath + "'");
		if (s.isEmpty())
			throw new NullPointerException("double-string is zero length from xpath '" + xpath + "'");
		try {
			return Double.parseDouble(s);
		}
		catch (NumberFormatException ex) {
			String msg = getExMsg(ex);
			//noinspection ThrowInsideCatchBlockWhichIgnoresCaughtException
			throw new NumberFormatException("Invalid double: '" + s + "'"
					+ " from xpath '" + xpath + "', " + msg);
		}
	}

	@NotNull
	protected BigDecimal parseBigDecimal(@Nullable String s, @NotNull String xpath) throws NumberFormatException
	{
		if (s == null)
			throw new NullPointerException("BigDecimal-string is null from xpath '" + xpath + "'");
		if (s.isEmpty())
			throw new NullPointerException("BigDecimal-string is zero length from xpath '" + xpath + "'");
		try {
			return new BigDecimal(s);
		}
		catch (NumberFormatException ex) {
			String msg = getExMsg(ex);
			//noinspection ThrowInsideCatchBlockWhichIgnoresCaughtException
			throw new NumberFormatException("Invalid BigDecimal: '" + s + "'"
					+ " from xpath '" + xpath + "', " + msg);
		}
	}

	@NotNull
	protected StatisticsEntry getStatisticsEntry(@NotNull String xpath)
	{
		synchronized (statistics) {
			StatisticsEntry statEntry = statistics.get(xpath);
			if (statEntry == null) {
				statEntry = new StatisticsEntry();
				statistics.put(xpath, statEntry);
			}
			return statEntry;
		}
	}


	static class StatisticsEntry
	{
		int numTimesCompiled, numTimesExecuted;
		long sumExecutionTime, sumCompileTime;
	}

	public static String getOptNodeData(Node node)
	{
		Node firstChild;
		if (node.getNodeType() == Node.ELEMENT_NODE
				&& ((firstChild = node.getFirstChild()) != null)
				&& (firstChild.getNextSibling() == null)
				&& (firstChild.getNodeType() == Node.TEXT_NODE)) {
			return ((CharacterData)firstChild).getData();
		}
		StringBuilder buf = new StringBuilder(25);
		subGetNodeData(node, buf);
		return (buf.length() > 0) ? buf.toString() : "";
	}

	private static void subGetNodeData(Node node, StringBuilder buf)
	{
		switch (node.getNodeType()) {
			case Node.DOCUMENT_NODE:
			case Node.ELEMENT_NODE:
				for (Node child = node.getFirstChild(); child != null; child = child.getNextSibling())
					subGetNodeData(child, buf);
				break;
			case Node.TEXT_NODE:
			case Node.CDATA_SECTION_NODE:
				buf.append(node.getNodeValue());
				break;
			case Node.ATTRIBUTE_NODE:
				buf.append(node.getNodeValue());
				break;
		}
	}

	@NotNull
	public static String getReqNodeData(Node node) throws SAXException
	{
		String s = getOptNodeData(node);
		if (s == null || s.isEmpty())
			throw new SAXException("Missing/empty required node text value");
		return s;
	}

	@Nullable
	public final String getOptContextNodeData()
	{
		return getOptNodeData(contextNode);
	}

	@NotNull
	public final String getReqContextNodeData() throws SAXException
	{
		String s = getOptNodeData(contextNode);
		if (s == null || s.isEmpty())
			throw makeSAXException("Missing/empty required string for context node", null);
		return s;
	}

	@Nullable
	public static String escapeStringArg(@Nullable String s)
	{
		if (s == null || !needsToEscapeStringArg(s))
			return s;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 5);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch < 32 || ch == '\'' || ch == '\\' || ch == '"')
				sb.append('\\');
			sb.append(ch);
		}
		return sb.toString();
	}

	private static boolean needsToEscapeStringArg(@Nullable String s)
	{
		if (s == null)
			return false;
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch < 32 || ch == '\'' || ch == '\\' || ch == '"')
				return true;
		}
		return false;
	}

	@NotNull
	protected PathProgram getProgram(@NotNull String xpath) throws SAXException
	{
		PathProgram pp;
		synchronized (cache) {
			pp = cache.get(xpath);
		}
		if (pp != null) {
			pp.numTimesExecuted++;
			return pp;
		}
		if (enableStatisticsFlag) {
			StatisticsEntry statEntry = getStatisticsEntry(xpath);
			long st0 = System.currentTimeMillis();
			pp = PathCompiler.compile(xpath);
			statEntry.numTimesCompiled++;
			statEntry.sumCompileTime += System.currentTimeMillis() - st0;
		} else {
			pp = PathCompiler.compile(xpath);
		}
		synchronized (cache) {
			pp.numTimesExecuted++;
			cache.put(xpath, pp);
		}
		return pp;
	}

	@NotNull
	protected PathRuntime getRuntime(@NotNull String xpath, PathRuntime.Interest interest) throws SAXException
	{
		PathProgram pp = getProgram(xpath);
		PathRuntime r = new PathRuntime(pp, interest);
		long st0;
		StatisticsEntry statEntry = null;
		if (enableStatisticsFlag) {
			statEntry = getStatisticsEntry(xpath);
			st0 = System.currentTimeMillis();
		}
		r.execute((Element)contextNode);
		if (enableStatisticsFlag) {
			statEntry.numTimesExecuted++;
			statEntry.sumExecutionTime += System.currentTimeMillis() - st0;
		}
		return r;
	}

	/**
	 * Get a short and simple description of the Exception:
	 * returns the exception's message (if non-empty), otherwise the class name.
	 * The description is stripped of "java.lang" and "java.lang.Exception" prefixes.
	 */
	@NotNull
	public static String getExMsg(Throwable t)
	{
		if (t == null)
			return "(exception object is null)";
		String msg = t.getMessage();
		if (msg == null || msg.isEmpty())
			msg = javaPackageRE.matcher(t.getClass().getName()).replaceAll("");
		return msg;
	}

	@NotNull
	private static final Pattern javaPackageRE = Pattern.compile("(java\\.lang(\\.Exception)?)");

	@NotNull
	public static final Comparator<String> usingCase = new Comparator<String>()
	{
		public int compare(String s1, String s2)
		{
			return s1.compareTo(s2);
		}
	};

	public static class DummyErrorLogger implements XmlArgsErrorLogger
	{
		public void didntGetRequiredElement(@NotNull XmlArgs xa, @NotNull String xpath) throws SAXException
		{
			System.err.println("Can't resolve xpath: " + xpath);
			logContextNodePath(xa);
		}

		public void didntExpectMultipleNodes(@NotNull XmlArgs xa,
		                                     @NotNull String xpath,
		                                     @NotNull List<Element> list) throws SAXException
		{
			System.err.println("Didn't expect " + list.size() + " nodes from: " + xpath);
			logContextNodePath(xa);
		}

		public void logContextNodePath(@NotNull XmlArgs xa) throws SAXException
		{
			Node n = xa.getContextNode();
			System.err.println("ContextNode: -------------- ");
			while (n != null) {
				if (n instanceof Element)
					System.err.println("   <" + n.getNodeName() + ">");
				else if (n instanceof Attr)
					System.err.println("   @" + n.getNodeName() + " = '" + n.getNodeValue() + "'");
				else
					System.err.println("  type = " + n.getNodeType() + ", " + n.getNodeName());
				n = n.getParentNode();
			}
			System.err.println("-----------------------------");
		}
	}
}

// END OF FILE
