/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "InstallerApp.h"
#import "GBDebug.h"
#import "AlertHelper.h"
#import "XmlDoc.h"
#import "AutoZap.h"
#import "VersionListJsonParser.h"
#include <sys/file.h>

@interface InstallerApp ()

@property (weak) IBOutlet NSMenuItem* menuItemAbout;
@property (weak) IBOutlet NSMenuItem* menuItemQuit;

@property (weak) IBOutlet NSWindow* mainWindow;
@property BOOL mainWindowIsPositioned;

@property IBOutlet NSView* progressView;
@property (weak) IBOutlet NSProgressIndicator *progressBar;
@property (weak) IBOutlet NSTextField *progressHeader;
@property (weak) IBOutlet NSTextField *progressMessage;
@property (weak) IBOutlet NSButton *progressOkButton;
@property (weak) IBOutlet NSButton *progressAllVersionsButton;

@property IBOutlet NSView* versionView;
@property (weak) IBOutlet NSTextField *versionListHeader;
@property (weak) IBOutlet NSPopUpButton *versionListPop;
@property (weak) IBOutlet NSTextField *versionListPopLabel;
@property (weak) IBOutlet NSTextField *versionListOldLabel;
@property (weak) IBOutlet NSTextField *versionCurrentLabel;
@property (weak) IBOutlet WebView *versionWebView;
@property (weak) IBOutlet NSButton *versionButtonSkip;
@property (weak) IBOutlet NSButton *versionButtonRemind;
@property (weak) IBOutlet NSButton *versionButtonDownload;

@property SInt32 osxMajor, osxMinor, osxBugfix;

@property NSFileManager* fm;
@property NSString* perlPath;
@property NSBundle* mainBundle;
@property NSString* filterDataPath;
@property NSURLSession* urlSession;
@property UpgradeChannel wantedChannel;
@property NSString* wantedVersion;
@property NSArray* versionList;

@property NSString* installedVersion; // NULL if none.
@property NSURL* versionInfoListURL;
@property NSMutableDictionary* versionInfoByVersion;

@property NSURL* tmpDirPath;
@property NSURL* applypatchPath;

@property AlertHelper* alertHelper;

@property GbVersionInfo* wantedVersionInfo;
@property double progressBase, progressSlice;

@property NSMutableArray* downloadUrls;  // NSURL
@property NSMutableArray* downloadSizes; // NSNumber
@property int downloadIndex;
@property int64_t downloadSizeOfAllPatches, downloadSizeCompleted;

@property AuthorizationRef authRef;
@property BOOL installationSuccess;

@end

@implementation InstallerApp

static NSString* gbBundlePath = @"/Library/PreferencePanes/GlimmerBlocker.prefPane";


- (instancetype)init
{
	if (!(self = [super init]))
		return nil;
	[AutoZap initGlobals];
	[AutoZap addSticky:self];
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	Gestalt(gestaltSystemVersionMajor, &_osxMajor);
	Gestalt(gestaltSystemVersionMinor, &_osxMinor);
	Gestalt(gestaltSystemVersionBugFix, &_osxBugfix);
	// NSOperatingSystemVersion osv = [processInfo operatingSystemVersion]; only in 10.10 and later.
#pragma GCC diagnostic pop
	_fm = [NSFileManager defaultManager];
	_perlPath = [GBDebug findCommandAbsPath:@"perl"];
	_mainBundle = [NSBundle mainBundle];
	NSAssert(_mainBundle, @"mainBundle");
	if ([_fm fileExistsAtPath:gbBundlePath]) {
		NSString* path = [gbBundlePath stringByAppendingString:@"/Contents/Info.plist"];
		NSDictionary *plistData = [NSDictionary dictionaryWithContentsOfFile:path];
		_installedVersion = [plistData objectForKey:@"CFBundleShortVersionString"];
		NSAssert(_installedVersion, @"installedVersion");
	}
	_wantedChannel = [VersionListJsonParser wantedChannel];
	//
	NSArray *args = [[NSProcessInfo processInfo] arguments];
	for (NSString* s in args) {
		//DebugNSLog(@"arg: %@", s);
		if ([s hasPrefix:@"/Library/GlimmerBlocker/Filter subscriptions"] && [s hasSuffix:@".xml"])
			_filterDataPath = s;
	}
	[[NSApplication sharedApplication] setDelegate:self];
	//_topLevelObjects = [AutoZap loadNibNamed:@"UpgradeApp2" bundle:_mainBundle owner:self];
	NSURLSessionConfiguration *config = [NSURLSessionConfiguration ephemeralSessionConfiguration];
	config.HTTPMaximumConnectionsPerHost = 1;
	_urlSession = [NSURLSession sessionWithConfiguration:config delegate:self delegateQueue: [NSOperationQueue mainQueue]];
	NSAssert(_urlSession, @"urlSession");
	//
	_tmpDirPath = [NSURL fileURLWithPath:[NSTemporaryDirectory() stringByAppendingPathComponent:@"GlimmerBlockerUpdater"] isDirectory:YES];
	NSError* error = NULL;
	[_fm createDirectoryAtURL:_tmpDirPath withIntermediateDirectories:YES attributes:nil error:&error];
	if (error)
		NSLog(@"Can't create tmp files: %@", error);
	NSLog(@"tmp dir: %@", _tmpDirPath);
	_applypatchPath = [_tmpDirPath URLByAppendingPathComponent:@"applypatch"];
	NSLog(@"applypatchPath: %@", _applypatchPath);
	[[NSAppleEventManager sharedAppleEventManager] setEventHandler:self
													   andSelector:@selector(handleEvent:withReplyEvent:)
													 forEventClass:kInternetEventClass
														andEventID:kAEGetURL];
	//
	return self;
}

- (void)dealloc
{
}

- (void)awakeFromNib
{
	//NSLog(@"awakeFromNib");
	NSAssert(_mainWindow, @"mainWindow");
	
	NSAssert(_progressView, @"progressView");
	NSAssert(_progressBar, @"progressBar");
	NSAssert(_progressHeader, @"progressHeader");
	NSAssert(_progressMessage, @"progressMessage");
	//
	NSAssert(_versionView, @"versionView");
	NSAssert(_versionListPop, @"versionListPop");
	NSAssert(_versionCurrentLabel, @"versionCurrentLabel");
	NSAssert(_versionWebView, @"versionWebView");
	//
	NSAssert(_versionButtonSkip, @"versionButtonSkip");
	NSAssert(_versionButtonRemind, @"versionButtonRemind");
	NSAssert(_versionButtonDownload, @"versionButtonDownload");
	//
	[_mainWindow setRestorable:NO];
	[_mainWindow orderOut:self];
	NSView* parent = _mainWindow.contentView;
	[self setWindowToView:_progressView];
	[_progressOkButton setHidden:YES];
	[_progressAllVersionsButton setHidden:YES];
	_alertHelper = [[AlertHelper alloc] initWindowView:parent withMainBundle:_mainBundle];
	//
	if (_wantedChannel == ChannelFinal) {
		[_versionListPop setHidden:YES];
		[_versionListPopLabel setHidden:YES];
	}
	[_versionListOldLabel setHidden:YES];
	if (!_installedVersion) {
		[_mainWindow setTitle:@"Installing GlimmerBlocker"];
	}
}

- (IBAction)orderFrontStandardAboutPanel:(id)sender
{
	[NSApp orderFrontStandardAboutPanel:sender];
}

- (IBAction)quit:(id)sender
{
	if (_tmpDirPath) {
		NSLog(@"Zaps temp dir");
		[[NSFileManager defaultManager] removeItemAtURL:_tmpDirPath error:NULL];
	}
	[NSApp terminate:self];
}

- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
	return menuItem != _menuItemAbout;
}

//=========================================================================================================

- (void)showFilterDialog
{
	NSAssert(_filterDataPath, @"_filterDataPath");
	NSURL* url = [NSURL fileURLWithPath:_filterDataPath];
	NSData* data = [NSData dataWithContentsOfURL:url];
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:url withData:data];
	[NSApp requestUserAttention:NSInformationalRequest];
	NSString* title = [NSString stringWithFormat:@"An update is available to the filter: “%@”", [doc stringForXPath:@"filter-data/@name"]];
	//
	NSMutableString* info = [NSMutableString stringWithCapacity:100];
	NSString* s = [doc dateFromMillisecsXPath:@"available-update-last-modified/@msecs"];
	if (s)
		[info appendFormat:@"Available update: %@\n", s];
	s = [doc dateFromMillisecsXPath:@"last-modified/@msecs"];
	if (s)
		[info appendFormat:@"Currently using: %@\n", s];
	//
	[_alertHelper prepareInformativeAlertWithTitle:title
					 withInformativeText:[info length] ? [info substringToIndex:[info length] - 1] : info];
	[[_alertHelper addButtonWithTitle:@"Show filter"] setKeyEquivalent:@"\r"];
	[[_alertHelper addButtonWithTitle:@"Ignore update"] setKeyEquivalent:@"\e"];
	[_alertHelper showPreparedAlert];
	NSInteger rc = [_alertHelper lastAlertReturnCode];
	if (rc != NSAlertFirstButtonReturn) {
		[self quit:self]; // clean temps.
		exit((int)rc);
		return;
	}
	NSData* magicData = [[[_filterDataPath lastPathComponent] stringByDeletingPathExtension] dataUsingEncoding:NSUTF8StringEncoding];
	NSString* path = [NSString stringWithFormat:@"/tmp/gb-edit-filter-%d", getuid()];
	NSDictionary* attrs = @{NSFilePosixPermissions: @0644L};
	BOOL b = [_fm createFileAtPath:path contents:magicData attributes:attrs];
	if (!b)
		NSLog(@"Couldn't create file: %@", path);
	//
	[[NSWorkspace sharedWorkspace] openFile:gbBundlePath];
	sleep(2); // wait a bit with quitting to avoid seeming like nothing happends while the System Preferences app launches.
}

//=========================================================================================================

- (void)setWindowToView:(NSView*)wanted
{
	[_progressView removeFromSuperview];
	[_versionView removeFromSuperview];
	//
	NSSize z = _progressView.frame.size;
	NSSize newSize = wanted.bounds.size;
	NSRect windowFrame = [_mainWindow contentRectForFrameRect:[_mainWindow frame]];
	NSRect newContentFrame = NSMakeRect( NSMinX(windowFrame), NSMaxY(windowFrame) - newSize.height, newSize.width, newSize.height);
	NSRect newWindowFrame = [_mainWindow frameRectForContentRect:newContentFrame];
	newWindowFrame.origin.y -= (windowFrame.size.height - newWindowFrame.size.height) / 2;
	DebugNSLog(@"setWindowToView: oldSize = %@, newSize = %@", NSStringFromSize(_mainWindow.frame.size), NSStringFromSize(newSize));
//	[_mainWindow orderFrontRegardless];
	[_mainWindow setFrame:newWindowFrame display:YES animate:YES];
	[_mainWindow.contentView addSubview:wanted];
	if (!_mainWindowIsPositioned) {
		_mainWindowIsPositioned = YES;
		[_mainWindow center];
	}
	[_mainWindow orderFront:self];
	NSUInteger sm = ([_mainWindow styleMask] & ~NSResizableWindowMask);
	if (wanted == _versionView) {
		sm |= NSResizableWindowMask;
		_mainWindow.contentMinSize = NSMakeSize(573, 300);
	} else {
		_mainWindow.contentMinSize = NSMakeSize(573, 100);
	}
	[_mainWindow setStyleMask:sm];
	DebugNSLog(@"   progress:  oldSize = %@, newSize = %@", NSStringFromSize(z), NSStringFromSize(_progressView.frame.size));
}

//=========================================================================================================

- (void)handleGenericError:(NSString*)message
{
	NSLog(@"handleGenericError: %@", message);
	message = [message stringByAppendingString:@"\n\nTry download full version from\nhttps://glimmerblocker.org"];
	[_alertHelper prepareCriticalAlertWithTitle:@"Upgrade failed!" withInformativeText:message];
	[_alertHelper setDelegate:self
				 withSelector:@selector(quit:)];
	[_alertHelper showPreparedAlert];
}

- (void)handleException:(NSException*)ex withMessage:(NSString*)message
{
	NSString* msg = [NSString stringWithFormat:@"%@\n\n%@", message, [ex reason]];
	NSLog(@"Got exception: %@", ex);
	[self handleGenericError:msg];
}

- (BOOL)checkNSError:(NSError*)error forUrl:(NSURL *)url
{
	if (!error)
		return YES;
	NSString* message = error.localizedDescription;
	if (url)
		message = [NSString stringWithFormat:@"%@\n%@", message, url];
	[self handleGenericError:message];
	return NO;
}

//=========================================================================================================

- (NSURLRequest*)makeURLRequestFor:(NSURL*)url
						allowCache:(BOOL)allowCache
{
	NSString *gbVersion = [_mainBundle objectForInfoDictionaryKey:@"CFBundleVersion"];
	NSString *ua = [NSMutableString stringWithFormat:@"GlimmerBlockerInstaller/%@ OSX/%d.%d.%d",
					gbVersion,
					_osxMajor, _osxMinor, _osxBugfix];
	NSURLRequestCachePolicy policy = allowCache ? NSURLRequestUseProtocolCachePolicy : NSURLRequestReloadIgnoringLocalCacheData;
	NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
	[request setCachePolicy:policy];
	[request setHTTPShouldHandleCookies:NO];
	[request setValue:ua forHTTPHeaderField:@"user-agent"];
	return request;
}

typedef void (^GbDownloadHandler)(NSData*);

- (void)startDownloadOf:(NSURL*)url
			 allowCache:(BOOL)allowCache
		 wantedMimeType:(NSString*)wantedMime
				handler:(GbDownloadHandler)handler
{
	void (^internalHandler)(NSData *data, NSURLResponse *response, NSError *error);
	internalHandler = ^(NSData *data, NSURLResponse *response, NSError *error) {
		if (![self checkNSError:error forUrl:response.URL])
			return;
		if (![response isKindOfClass:[NSHTTPURLResponse class]]) {
			NSString* msg = [NSString stringWithFormat:@"Download response type is not HTTP for:\n%@", url];
			[self handleGenericError:msg];
			return;
		}
		NSHTTPURLResponse *hresp = (NSHTTPURLResponse*)response;
		if (hresp.statusCode != 200 && hresp.statusCode != 304) {
			NSString* msg = [NSString stringWithFormat:@"Could not download from server (http status %ld) for url:\n%@", hresp.statusCode, url];
			[self handleGenericError:msg];
			return;
		}
		if (wantedMime) {
			NSString* actual = [response MIMEType];
			if (![actual hasPrefix:wantedMime]) {
				NSString* msg = [NSString stringWithFormat:@"Server returned wrong kind of data for:\n%@\nExpected %@, got %@",
								 url, wantedMime, actual];
				[self handleGenericError:msg];
		return;
			}
		}
		//NSLog(@"calls handler for %@", url);
		handler(data);
	};
	NSURLRequest *request = [self makeURLRequestFor:url allowCache:allowCache];
	NSURLSessionDataTask *task = [_urlSession dataTaskWithRequest:request completionHandler:internalHandler];
	[task resume];
}

- (BOOL)gunzipAndWriteData:(NSData*)data toPath:(NSURL*)path
{
	NSURL *gz = [NSURL URLWithString:[[path absoluteString] stringByAppendingString:@".gz"]];
	[_fm removeItemAtURL:gz error:NULL];
	NSError *error = NULL;
	[data writeToURL:gz options:NSDataWritingAtomic error:&error];
	if (![self checkNSError:error forUrl:NULL])
		return NO;
	[_fm removeItemAtURL:path error:NULL];
	//
	int x = [GBDebug executeCommandlineTool:@"gunzip" withArgs:@[@"-f", [gz path]]];
	if (x != 0) {
		[self handleGenericError:@"Failed unzipping updater."];
		return NO;
	}
	if (![path checkResourceIsReachableAndReturnError:&error] || error)
		NSLog(@"gunzipped file, but it is gone: %@", error);
	return YES;
}

//=========================================================================================================

- (void)startProgressSlice:(double)size
{
	//NSLog(@"startProgressSlice:  %.1f + %.1f -> %.1f  ++  %.1f", _progressBase, _progressSlice, _progressBase + _progressSlice, size);
	_progressBase += _progressSlice;
	[_progressBar setDoubleValue:_progressBase];
	_progressSlice = size;
}

- (void)setProgress:(double)fraction
{
	//NSLog(@"  setProgress %.2f:  %.1f + %.1f  / %.1f", fraction, _progressBase, fraction * _progressSlice, _progressBase + _progressSlice);
	if (fraction < 0)
		fraction = 0;
	else if (fraction > 1)
		fraction = 1;
	double v = _progressBase + fraction * _progressSlice;
	[_progressBar setDoubleValue:v];
}

- (void)commitProgress
{
	//NSLog(@"commitProgress");
	[self startProgressSlice:0];
}

- (IBAction)handleProgressOkButton:(id)sender
{
	[[NSWorkspace sharedWorkspace] launchApplication:gbBundlePath];
	[self quit:self];
}

//=========================================================================================================

- (NSURL*)getBaseUrlForVersion:(NSString*)version
{
	NSDictionary *dict = [_versionInfoByVersion objectForKey:version];
	NSString *s = [dict objectForKey:@"url"];
	if (!s.length) {
		NSLog(@"Missing baseUrl for version '%@'", version);
		return NULL;
	}
	if ([s hasPrefix:@"/"])
		s = [@"https://glimmerblocker.org/upgrades" stringByAppendingString:s];
	return [NSURL URLWithString:s];
}

- (NSURL*)getUrlFor:(NSString*)key forVersion:(NSString*)version
{
	NSURL *baseUrl = [self getBaseUrlForVersion:version];
	if (!baseUrl)
		return NULL;
	NSDictionary *dict = [_versionInfoByVersion objectForKey:version];
	NSString *relUrl = [dict objectForKey:@"notes"];
	if (!relUrl.length) {
		NSLog(@"Missing %@ for version '%@'", key, version);
		return NULL;
	}
	return [NSURL URLWithString:[[baseUrl absoluteString] stringByAppendingString:relUrl]];
}

- (void)quitSystemPreferences
{
	for (NSRunningApplication *app in [NSRunningApplication runningApplicationsWithBundleIdentifier:@"com.apple.systempreferences"]) {
		//NSLog(@"Running app: %@", dict);
		NSLog(@"Quits System Preferences with pid %d", app.processIdentifier);
		[app terminate];
		NSLog(@"Done quitting System Preferences app");
	}
}

- (void)runInstaller
{
	[self quitSystemPreferences];
	// If the $TMPDIR directory doesn't exist (or is unwriteable),
	// AuthorizationExecuteWithPrivileges fails with errAuthorizationInternal = -60008
	// so we must ensure it points to an existing directory which this process has write permissions to.
	char* origTmpDirPath = getenv("TMPDIR");
	setenv("TMPDIR", [[_tmpDirPath path] UTF8String], 1);
	const char* args[] = { "-script",
		[_perlPath UTF8String],
		"-apply",
		[[_tmpDirPath path] UTF8String],
		[[_applypatchPath path] UTF8String],
		NULL};
	FILE *msgPipe = NULL;
	NSLog(@"Calls AuthorizationExecuteWithPrivileges");
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	OSStatus status = AuthorizationExecuteWithPrivileges(_authRef,
														 [[_applypatchPath path] UTF8String],
														 kAuthorizationFlagDefaults,
														 (char**)args,
														 &msgPipe);
#pragma GCC diagnostic pop
	if (status == errAuthorizationSuccess) {
		NSLog(@"Echoes stderr from installer.");
		NSMutableString* allOutput = [NSMutableString stringWithCapacity:512];
		char buffer[128];
		NSMutableData* msgBuffer = [NSMutableData dataWithCapacity:80];
		while (1) {
			long num = read(fileno(msgPipe), buffer, sizeof(buffer));
			if (num < 1)
				break;
			[msgBuffer appendBytes:buffer length:num];
			NSInteger len = [msgBuffer length];
			int idx = 0;
			while (idx < len) {
				char ch = ((const char*)[msgBuffer bytes])[idx];
				if (ch != '\n') {
					idx++;
					continue;
				}
				NSString* s = [[NSString alloc] initWithBytes:[msgBuffer bytes] length:idx encoding:NSUTF8StringEncoding];
				[allOutput appendString:s];
				[allOutput appendString:@"\n"];
				NSLog(@"%@", s);
				[msgBuffer replaceBytesInRange:NSMakeRange(0, idx + 1) withBytes:"" length:0];
				len -= idx + 1;
				idx = 0;
			}
		}
		NSLog(@"installer done.");
		if ([allOutput rangeOfString:@"INSTALLATION-SUCCESS"].length > 0) {
			_installationSuccess = YES;
			NSLog(@"Got success-marker.");
		}
	} else {
		NSLog(@"AuthorizationExecuteWithPrivileges failed with statuscode %d", status);
	}
	fclose(msgPipe);
	if (origTmpDirPath)
		setenv("TMPDIR", origTmpDirPath, 1);
	else
		unsetenv("TMPDIR");
	if (status != errAuthorizationSuccess || !_installationSuccess) {
		[_progressMessage setStringValue:@""];
		[_progressHeader setStringValue:@"Upgrade FAILED."];
		[_progressBar setHidden:YES];
		[self handleGenericError:@"Installer failed."];
		return;
	}
	// ping server with installed version, replacement for stats of downloaded DMG versions.
	NSString *notifyUrl = [@"https://glimmerblocker.org/site/install-notify.txt?v=" stringByAppendingString:_wantedVersionInfo.version];
	if (_installedVersion)
		notifyUrl = [NSString stringWithFormat:@"%@&old=%@", notifyUrl, _installedVersion];
	NSURLRequest *request = [self makeURLRequestFor:[NSURL URLWithString:notifyUrl] allowCache:NO];
	NSURLSessionDataTask *task = [_urlSession dataTaskWithRequest:request
												completionHandler:^(NSData *data, NSURLResponse *response, NSError *error) {}];
	[task resume];
	//
	NSLog(@"Show success feedback.");
	[_progressBar setDoubleValue:100.0];
	[_progressMessage setStringValue:@""];
	[_progressHeader setStringValue:@"Upgrade complete."];
	[_progressOkButton setHidden:NO];
	[_progressBar setHidden:YES];
	[self performSelector:@selector(quit:) withObject:self afterDelay:(15 * 60)]; // auto-quit so installer doesn't hang around forever for people who never quit apps.
}

- (void)prepareAuth
{
	AuthorizationRef authRef = NULL;
	AuthorizationFlags authFlags = kAuthorizationFlagDefaults | kAuthorizationFlagInteractionAllowed | kAuthorizationFlagExtendRights;
	OSStatus status = AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, authFlags, &authRef);
	if (status != errAuthorizationSuccess) {
		NSLog(@"GlimmerBlocker: AuthorizationCreate failed");
		[self handleGenericError:@"Can't get authorization to run installer."];
		return;
	}
	DebugNSLog(@"GlimmerBlocker: AuthorizationCreate ok");
	AuthorizationItem myItems = {kAuthorizationRightExecute, 0, NULL, 0};
	AuthorizationRights myRights = {1, &myItems};
	authFlags = kAuthorizationFlagInteractionAllowed | kAuthorizationFlagExtendRights;
	status = AuthorizationCopyRights(authRef, &myRights, NULL, authFlags, NULL );
	if (status != errAuthorizationSuccess) {
		NSLog(@"GlimmerBlocker: AuthorizationCopyRights failed: %ld", (long)status);
		[self handleGenericError:@"Could not get authentication to run installer."];
		return;
	}
	DebugNSLog(@"GlimmerBlocker: AuthorizationCopyRights ok");
	_authRef = authRef;
	[_progressMessage setStringValue:@"Running installer. Please wait."];
	[self performSelector:@selector(runInstaller) withObject:nil afterDelay:0.1]; // get UI updated.
}

- (void)updateDownloadProgress:(int64_t)bytesReceivedForThisFile
{
	int64_t fileSize = [[_downloadSizes objectAtIndex:_downloadIndex] longLongValue];
	if (bytesReceivedForThisFile > fileSize)
		bytesReceivedForThisFile = fileSize;
	double totalBytesReceived = bytesReceivedForThisFile + _downloadSizeCompleted;
	double totalBytesWanted = _downloadSizeOfAllPatches;
	double fractionReceived = totalBytesReceived / totalBytesWanted;
	NSString* progress;
	if (totalBytesWanted > 5 * 1024 * 1024) {
		progress = [NSString stringWithFormat:@"%.1f of %.1f MB downloaded",
						totalBytesReceived / 1024 / 1024, totalBytesWanted / 1024 / 1024];
	} else {
		progress = [NSString stringWithFormat:@"%.0f of %.0f kB downloaded",
					totalBytesReceived / 1024, totalBytesWanted / 1024];
	}
	[_progressMessage setStringValue:[NSString stringWithFormat:@"%@  (file %d of %ld)",
									  progress, _downloadIndex + 1, _downloadUrls.count]];
	[self setProgress:fractionReceived];
}

- (void)URLSession:(NSURLSession *)session
			  task:(NSURLSessionTask *)task
didCompleteWithError:(NSError *)error
{
	if ([self checkNSError:error forUrl:task.originalRequest.URL])
		return;
}

- (void)URLSession:(NSURLSession *)session
	  downloadTask:(NSURLSessionDownloadTask *)downloadTask
	  didWriteData:(int64_t)bytesWritten
 totalBytesWritten:(int64_t)totalBytesWritten
totalBytesExpectedToWrite:(int64_t)totalBytesExpectedToWrite
{
	[self updateDownloadProgress:totalBytesWritten];
	//NSLog(@"  got %lld of %lld / %lld", totalBytesWritten, totalBytesExpectedToWrite, [[_downloadSizes objectAtIndex:_downloadIndex] longLongValue]);
}

- (void)URLSession:(NSURLSession *)session
	  downloadTask:(NSURLSessionDownloadTask *)downloadTask
 didResumeAtOffset:(int64_t)fileOffset
expectedTotalBytes:(int64_t)expectedTotalBytes
{
	NSLog(@"downloadTask didResumeAtOffset: %lld", fileOffset);
}

- (void)startNextDownload
{
	if (_downloadIndex >= _downloadUrls.count) {
		NSLog(@"All needed files downloaded. Proceeding to installation.");
		[self commitProgress];
		[_progressMessage setStringValue:@"Ready to install. Please authenticate."];
		[self performSelector:@selector(prepareAuth) withObject:nil afterDelay:0.1]; // get UI updated.
		return;
	}
	NSURL* serverUrl = [_downloadUrls objectAtIndex:_downloadIndex];
	//NSNumber* sz = [_downloadSizes objectAtIndex:_downloadIndex];
	//NSLog(@"Downloads file #%d, %lld bytes: %@", _downloadIndex + 1, [sz longLongValue], serverUrl);
	[self updateDownloadProgress:0];
	NSURLRequest *request = [self makeURLRequestFor:serverUrl allowCache:YES];
	[[_urlSession downloadTaskWithRequest:request] resume];
}

- (void)URLSession:(NSURLSession *)session
	  downloadTask:(NSURLSessionDownloadTask *)downloadTask
didFinishDownloadingToURL:(NSURL *)location
{
	//NSLog(@"Did finish downloading %@", downloadTask.originalRequest.URL);
	NSURL* serverUrl = [_downloadUrls objectAtIndex:_downloadIndex];
	NSURL* dstFileUrl = [_tmpDirPath URLByAppendingPathComponent:[serverUrl lastPathComponent]];
	NSError *error = NULL;
	[_fm removeItemAtURL:dstFileUrl error:&error];
	error = NULL; // ignore remove errors. Rename would fail then.
	[_fm moveItemAtURL:location toURL:dstFileUrl error:&error];
	if (![self checkNSError:error forUrl:serverUrl])
		return;
	//NSLog(@"File saved as: %@", [dstFileUrl path]);
	_downloadSizeCompleted += [[_downloadSizes objectAtIndex:_downloadIndex] longLongValue];
	[self updateDownloadProgress:0];
	_downloadIndex++;
	[self performSelector:@selector(startNextDownload) withObject:nil afterDelay:0.1];
}

- (NSArray*)getStringMatches:(NSString*)s forRegexp:(NSString*)regexText
{
	NSRegularExpression *regex = [NSRegularExpression regularExpressionWithPattern:regexText options:0 error:nil];
	return [regex matchesInString:s options:0 range:NSMakeRange(0, s.length)];
}

- (void)handleDownloadedApplypatch:(NSData*)data
{
	NSLog(@"Received applypatch, %ld bytes, for path %@", [data length], _applypatchPath);
	if (![self gunzipAndWriteData:data toPath:_applypatchPath])
		return;
	//
	NSMutableDictionary* attrs = [NSMutableDictionary dictionaryWithCapacity:1];
	[attrs setValue:[NSNumber numberWithShort:0777] forKey:NSFilePosixPermissions];
	[_fm setAttributes:attrs ofItemAtPath:[_applypatchPath path] error:nil];
	[self commitProgress];
	[_progressMessage setStringValue:@"Prepares download."];
	//
	NSMutableString *stdout = [NSMutableString stringWithCapacity:256];
	int x = [GBDebug executeCommandlineTool:[_applypatchPath path]
								   withArgs:@[@"-script", _perlPath, @"-urls", [_tmpDirPath path], [_applypatchPath path]]
							   stdoutString:stdout];
	if (x != 0) {
		NSLog(@"stdout from failed process: %@", stdout);
		[self handleGenericError:@"Failed determining changed files."];
		return;
	}
	NSRegularExpression *regex = [NSRegularExpression regularExpressionWithPattern:@"^DOWNLOAD\\s+(\\d+)\\s+(\\S+)\\n*"
																		   options:0
																			 error:nil];
	_downloadUrls = [NSMutableArray arrayWithCapacity:10];
	_downloadSizes = [NSMutableArray arrayWithCapacity:10];
	for (NSString *line in [stdout componentsSeparatedByString:@"\n"]) {
		NSTextCheckingResult *r = [regex firstMatchInString:line options:0 range:NSMakeRange(0, line.length)];
		//NSLog(@"line (%d): %@", !!r, line);
		if (!r || r.range.location == NSNotFound) {
			if (line.length)
				NSLog(@"%@", line);
			continue;
		}
		long size = [[line substringWithRange:[r rangeAtIndex:1]] integerValue];
		NSString *url = [line substringWithRange:[r rangeAtIndex:2]];
		_downloadSizeOfAllPatches += size;
		if (![url hasPrefix:@"http"])
			url = [[[_wantedVersionInfo baseUrl] absoluteString] stringByAppendingString:url];
		[_downloadUrls addObject:[NSURL URLWithString:url]];
		[_downloadSizes addObject:[NSNumber numberWithLong:size]];
	}
	NSLog(@"Got %ld patches to download, total size = %lld MB", _downloadUrls.count, _downloadSizeOfAllPatches / 1024 / 1024);
	_downloadIndex = 0;
	_downloadSizeCompleted = 0;
	[self startProgressSlice:80];
	[self startNextDownload];
}

- (IBAction)handleButtonInstallUpdate:(id)sender
{
	NSLog(@"handleButtonInstallUpdate");
	_wantedVersionInfo = [_versionInfoByVersion objectForKey:[_versionListPop selectedItem].title];
	if (!_wantedVersionInfo) {
		[self handleGenericError:@"Info for selected version missing!!!"];
		return;
	}
	NSLog(@"Fetches updater: %@", _wantedVersionInfo.applyPatchUrl);
	[_progressHeader setStringValue:@"Updates GlimmerBlocker"];
	[_progressMessage setStringValue:@"Downloads updater"];
	[self setWindowToView:_progressView];
	[self startProgressSlice:5]; // initially fraction
	[self commitProgress];
	[self startProgressSlice:5]; // updater
	[_progressBar setIndeterminate:NO];
	GbDownloadHandler handler = ^(NSData* data) { [self handleDownloadedApplypatch:data]; };
	[self startDownloadOf:_wantedVersionInfo.applyPatchUrl
			   allowCache:(_wantedChannel != ChannelDev)
		   wantedMimeType:NULL handler:handler];
}

- (IBAction)handleButtonSkipThisVersion:(id)sender
{
	NSString *version = [_versionListPop selectedItem].title;
	NSLog(@"handleButtonSkipThisVersion: %@", version);
	NSUserDefaults* defaults = [[NSUserDefaults alloc] initWithSuiteName:PREFS_SUITE_ID_INSTALLER];
	[defaults setObject:version forKey:PREFSKEY_SKIPPED_VERSION];
	[defaults synchronize];
	[self quit:self];
}

- (IBAction)handleButtonRemindMeLater:(id)sender
{
	NSLog(@"handleButtonRemindMeLater");
	[self quit:self];
}

- (void)loadReleaseNotes:(GbVersionInfo *)info
{
	DebugNSLog(@"loadReleaseNotes: %@", info.releaseNotesUrl);
	[[_versionWebView mainFrame] loadRequest:[NSURLRequest requestWithURL:info.releaseNotesUrl]];
}


- (void)handleSelectedVersion:(id)sender
{
	NSString *version = [_versionListPop selectedItem].title;
	GbVersionInfo *info = [_versionInfoByVersion objectForKey:version];
	if (!info) {
		[_versionButtonSkip setEnabled:NO];
		[_versionButtonDownload setEnabled:NO];
		NSURL* siteUrl = [NSURL URLWithString:@"https://glimmerblocker.org"];
		[[_versionWebView mainFrame] loadHTMLString:@"Missing version!!!" baseURL:siteUrl];
		return;
	}
	if (info.isUpgrade) {
		[_versionListPopLabel setStringValue:@"Upgrade to:"];
		[_versionListPopLabel setTextColor:[NSColor blackColor]];
		[_versionListOldLabel setHidden:YES];
	} else if (info.isInstalled) {
		[_versionListPopLabel setStringValue:@"Version:"];
		[_versionListPopLabel setTextColor:[NSColor blackColor]];
		[_versionListOldLabel setHidden:NO];
		[_versionListOldLabel setStringValue:@"(this is your installed version)"];
		[_versionListOldLabel setTextColor:[NSColor blackColor]];
	} else {
		[_versionListPopLabel setStringValue:@"DOWNGRADE to:"];
		[_versionListPopLabel setTextColor:[NSColor redColor]];
		[_versionListOldLabel setHidden:NO];
		[_versionListOldLabel setStringValue:@"(this is an OLDER VERSION)"];
		[_versionListOldLabel setTextColor:[NSColor redColor]];
	}
	[_versionButtonSkip setEnabled:!info.isInstalled];
	[_versionButtonDownload setEnabled:!info.isInstalled];
	NSString *html = [NSString stringWithFormat:@"<i style='color: #aaa;'>Loads release notes for %@</i>", info.version];
	[[_versionWebView mainFrame] loadHTMLString:html baseURL:info.releaseNotesUrl];
	[self performSelector:@selector(loadReleaseNotes:) withObject:info afterDelay:(0.05)]; // auto-quit so installer doesn't hang around forever for people who never quit apps.
}

- (void)showVersionList:(BOOL)showAllVersions
{
	NSMenu *menu = _versionListPop.menu;
	[menu removeAllItems];
	_versionInfoByVersion = [NSMutableDictionary dictionaryWithCapacity:_versionList.count];
	BOOL hasSeenFinalVersion = NO;
	BOOL hasSeenWantedVersion = NO;
	NSString *newestUpgrade = NULL;
	for (GbVersionInfo *info in _versionList) {
		if (![_wantedVersion isEqualToString:info.version] && _wantedChannel == ChannelFinal && !info.isUpgrade)
			continue;
		if (info.isUpgrade && !newestUpgrade)
			newestUpgrade = info.version;
		[_versionInfoByVersion setObject:info forKey:info.version];
		NSMenuItem *item = [menu addItemWithTitle:info.version action:@selector(handleSelectedVersion:) keyEquivalent:@""];
		if (info.channel == ChannelFinal)
			hasSeenFinalVersion = YES;
		if ([_wantedVersion isEqualToString:info.version]) {
			[_versionListPop selectItem:item];
			hasSeenWantedVersion = YES;
		}
		if (_wantedChannel != ChannelDev && !info.isUpgrade && hasSeenFinalVersion)
			break;
	}
	if (_wantedVersion && !hasSeenWantedVersion) {
		[_progressHeader setStringValue:@"Version not available."];
		[_progressMessage setStringValue:[NSString stringWithFormat:@"Version %@ is no longer available for download.", _wantedVersion]];
		[_progressOkButton setHidden:NO];
		[_progressBar setHidden:YES];
		return;
	}
	if (_wantedChannel != ChannelDev && !newestUpgrade && !showAllVersions) {
		[_progressHeader setStringValue:@"You have the latest available version installed."];
		[_progressMessage setStringValue:@""];
		[_progressOkButton setHidden:NO];
		[_progressBar setHidden:YES];
		[_progressAllVersionsButton setHidden:(_wantedChannel == ChannelFinal || showAllVersions)];
		return;
	}
	if (_installedVersion) {
		NSString *msg;
		if (newestUpgrade)
			msg = [NSString stringWithFormat:@"GlimmerBlocker %@ is now available—you have version %@ installed", newestUpgrade, _installedVersion];
		else
			msg = [NSString stringWithFormat:@"You have version %@ installed", _installedVersion];
		[_versionCurrentLabel setStringValue:msg];
		if (newestUpgrade)
			[_versionListHeader setStringValue:@"An update to GlimmerBlocker is available"];
		else
			[_versionListHeader setStringValue:@"No upgrade available."];
	} else {
		[_versionListHeader setStringValue:@"Get the latest GlimmerBlocker"];
		NSString *msg;
		if (newestUpgrade)
			msg = [NSString stringWithFormat:@"GlimmerBlocker %@ is now available", newestUpgrade];
		else
			msg = @"";
		[_versionCurrentLabel setStringValue:msg];
		[_versionButtonSkip setHidden:YES];
		[_versionButtonRemind setHidden:YES];
	}
	if (menu.itemArray.count) {
		[self handleSelectedVersion:self];
	} else {
		[_versionListOldLabel setHidden:YES];
		[_versionListPop setHidden:YES];
		[_versionButtonSkip setEnabled:NO];
		[_versionButtonDownload setEnabled:NO];
		NSURL* siteUrl = [NSURL URLWithString:@"https://glimmerblocker.org"];
		[[_versionWebView mainFrame] loadHTMLString:@"" baseURL:siteUrl];
	}
	[self setWindowToView:_versionView];
}

- (IBAction)handleProgressAllVersionsButton:(id)sender
{
	[self showVersionList:YES];
}

- (void)handleVersionListResponseData:(NSData*)data
{
	NSError* error = NULL;
	VersionListJsonParser *parser = [[VersionListJsonParser alloc] initWithData:data
													installedVersion:_installedVersion
															 verbose:YES
															   error:&error];
	if (![self checkNSError:error forUrl:_versionInfoListURL])
		return;
	if (!parser) {
		[self handleGenericError:@"Can't parse version list info from server."];
		return;
	}
	_versionList = parser.versionList;
	[self showVersionList:NO];
}


- (BOOL)checkFreeSpace
{
	NSError *error = nil;
	NSDictionary *attr = [_fm attributesOfFileSystemForPath:@"/Library/PreferencePanes" error:&error];
	int64_t bytesFree = [[attr objectForKey:NSFileSystemFreeSize] longLongValue];
	int64_t mbFree = bytesFree / 1024 / 1024;
	NSLog(@"free space: %lld bytes = %lld MB", bytesFree, mbFree);
	int64_t minMb = 500;
	if (mbFree >= minMb)
		return YES;
	[self handleGenericError:[NSString stringWithFormat:
							  @"You have too little free space available (%lld MB) for GlimmerBlocker to upgrade. At least %lld MB must be available.",
							  mbFree, minMb]];
	return NO;
}

- (void)getListOfNewVersions
{
	if (![self checkFreeSpace])
		return;
	_versionInfoListURL = [VersionListJsonParser versionListURLForChannel:_wantedChannel];
	DebugNSLog(@"versionUrl: %@", _versionInfoListURL);
	[_progressHeader setStringValue:@"Fetching update info"];
	[_progressMessage setStringValue:@""];
	[self setWindowToView:_progressView];
	[_progressBar setMinValue:0];
	[_progressBar setMaxValue:100];
	[_progressBar setIndeterminate:YES];
	GbDownloadHandler handler = ^(NSData* data) { [self handleVersionListResponseData:data]; };
	[self startDownloadOf:_versionInfoListURL
			   allowCache:NO
		   wantedMimeType:@"application/json"
				  handler:handler];
	NSDictionary* attrs = @{NSFilePosixPermissions: @0666L};
	[_fm createFileAtPath:LAST_UPDATE_CHECK_MARKER_PATH contents:[NSData data] attributes:attrs];
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	DebugNSLog(@"applicationDidFinishLaunching");
	if (_filterDataPath) {
		[self showFilterDialog];
		return;
	}
	[self getListOfNewVersions];
}

- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
	DebugNSLog(@"openFile: %@", filename);
	if ([filename hasPrefix:@"/Library/GlimmerBlocker/Filter subscriptions"] && [filename hasSuffix:@".xml"])
		_filterDataPath = filename;
	return NO;
}

- (BOOL)applicationOpenUntitledFile:(NSApplication *)theApplication
{
	DebugNSLog(@"applicationOpenUntitledFile");
	return NO;
}

- (void)handleEvent:(NSAppleEventDescriptor *)event withReplyEvent:(NSAppleEventDescriptor *)replyEvent;
{
	NSURL *url = [NSURL URLWithString:[[event paramDescriptorForKeyword:keyDirectObject] stringValue]];
	NSLog(@"handleEvent.url = %@", url);
	if (_versionInfoListURL) {
		NSString *msg = [NSString stringWithFormat:@"%@\n\n%@",
						 @"The installer cannot change version midstream",
						 url];
		[_alertHelper prepareInformativeAlertWithTitle:@"Quit installer before clicking URL"
								   withInformativeText:msg];
		[_alertHelper showPreparedAlert];
		return;
	}
	NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
	for (NSString *qs in [url.query componentsSeparatedByString:@"&"]) {
		NSString *key = [[qs componentsSeparatedByString:@"="] objectAtIndex:0];
		NSString *value = [[qs componentsSeparatedByString:@"="] objectAtIndex:1];
		value = [value stringByReplacingOccurrencesOfString:@"+" withString:@" "];
		value = [value stringByReplacingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
		dict[key] = value;
	}
	NSString *ch = dict[@"ch"];
	if ([ch isEqualToString:@"dev"]) {
		_wantedChannel = ChannelDev;
	} else if ([ch isEqualToString:@"beta"]) {
		_wantedChannel = ChannelBeta;
	} else if ([ch isEqualToString:@"final"]) {
		_wantedChannel = ChannelFinal;
	} else {
		NSString *msg = [NSString stringWithFormat:@"%@\n\n%@",
						 @"The clicked URL cannot be parsed (invalid channel).",
						 url];
		[_alertHelper prepareInformativeAlertWithTitle:@"Unsupported / invalid URL"
								   withInformativeText:msg];
		[_alertHelper showPreparedAlert];
		return;
		
	}
	NSString *v = dict[@"v"];
	NSRegularExpression *regex = [NSRegularExpression regularExpressionWithPattern:@"^\\d+(\\.\\d+)+([a-z]\\d+)?$"
																		   options:0
																			 error:nil];
	if ([regex numberOfMatchesInString:v options:0 range:NSMakeRange(0, v.length)] == 0) {
		NSString *msg = [NSString stringWithFormat:@"%@\n\n%@",
						 @"The clicked URL cannot be parsed (invalid version).",
						 url];
		[_alertHelper prepareInformativeAlertWithTitle:@"Unsupported / invalid URL"
								   withInformativeText:msg];
		[_alertHelper showPreparedAlert];
		return;
	}
	_wantedVersion = v;
	NSLog(@"Valid URL clicked, ch = '%@', v = '%@'", ch, v);
}

@end

static void exitIfAlreadyRunning()
{
	NSString* path = [NSString stringWithFormat:@"%@/Contents/Info.plist", [[NSBundle mainBundle] bundlePath]];
	NSFileHandle *fh = [NSFileHandle fileHandleForReadingAtPath:path];
	if (!fh) {
		NSLog(@"Updater can't open its Info.plist: %@", path);
		return;
	}
	BOOL ok = flock([fh fileDescriptor], (LOCK_EX | LOCK_NB)) >= 0;
	if (ok)
		return;
	NSLog(@"GlimmerBlocker updater already running. Quits.");
	exit(0);
}

int main(int argc, const char *argv[])
{
	@autoreleasepool {
		@try {
			exitIfAlreadyRunning();
			NSLog(@"GlimmerBlocker updater running, checks for updates.");
			return NSApplicationMain(argc, argv);
		}
		@catch (NSException* ex) {
			NSLog(@"Stops due to exception: '%@' / '%@'", [ex name], [ex reason]);
			exit(1);
		}
	}
	return 0;
}
