/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterListView.h"
#import "FilterTabController.h"
#import "Prefs.h"
#import "Filter.h"
#import "ServerCommunication.h"
#import "Xml.h"
#import "FilterRuleTableView.h"

#ifdef DEBUG
//#define DEBUG_DD
#endif

@interface FilterListDropView : NSView
{
	NSColor* _color;
}

- (instancetype)initWithFrame:(NSRect)frameRect;
- (void)drawRect:(NSRect)r;

@end

@implementation FilterListDropView

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	_color = [NSColor colorWithDeviceRed:0.23f green:0.33f blue:0.85f alpha:1];
	return self;
}


- (void)drawRect:(NSRect)r
{
	[super drawRect:r];
	[_color set];
	NSRectFill(r);
}

@end

@interface FilterListView()

- (void)releaseDragInfo;

@end


@implementation FilterListView

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	[self registerForDraggedTypes:@[[Filter draggingDataType]]];
	return self;
}

- (void)dealloc
{
	[self releaseDragInfo];
	[self setController:NULL];
}

- (void)setController:(FilterTabController*)controller
{
	_controller = controller;
	_prefs = [controller prefs];
}

- (void)zapDropViewFrame
{
	if (!_dropDestinationView)
		return;
#ifdef DEBUG_DD
	NSLog(@"Zaps _dropDestinationView");
#endif
	[_dropDestinationView removeFromSuperview];
	_dropDestinationView = NULL;
}

- (NSPoint)findDropDestinationFilter:(NSPoint)mousePos withDestinationIndex:(NSUInteger*)dstIndex
{
	if (dstIndex)
		*dstIndex = NSNotFound;
	NSArray* rows = [self rows];
	NSInteger num = [rows count];
	for (int i = 0; i < num; i++) {
		Filter* filter = rows[i];
		NSRect r = [filter frame];
		if (mousePos.y > r.origin.y + r.size.height && i != num - 1) // always accept pos below last item in list.
			continue;
		CGFloat y = r.origin.y - 1;
		if (mousePos.y > r.origin.y + r.size.height / 2) {
			y += r.size.height;
			i++;
		} else {
			if (i == 0)
				y += 1;
		}
		if (NO) {
			NSLog(@"dst mp = %.0f, idx = %d, r = %.0f to %.0f to %.0f -> %ld",
				  mousePos.y, i, r.origin.y,
				  r.origin.y + r.size.height / 2, r.origin.y + r.size.height,
				  (long)(dstIndex ? *dstIndex : 66666));
		}
		if (dstIndex)
			*dstIndex = i;
		return NSMakePoint(0, y);
	}
	return NSMakePoint(-1, -1);
}

- (NSDragOperation)setDropViewFrame:(NSPoint)pos
{
	pos = [self findDropDestinationFilter:pos withDestinationIndex:NULL];
	if (pos.x < 0) {
		[self zapDropViewFrame];
		return NSDragOperationNone;
	}
	if (_dropDestinationView) {
		[_dropDestinationView setFrameOrigin:pos];
	} else {
		NSRect r = NSMakeRect(0, pos.y, [self frame].size.width, 2);
#ifdef DEBUG_DD
		NSLog(@"Creates _dropDestinationView: %@", NSStringFromRect(r));
#endif
		_dropDestinationView = [[FilterListDropView alloc] initWithFrame:r];
		[self addSubview:_dropDestinationView positioned:NSWindowAbove relativeTo:NULL];
	}
	return NSDragOperationMove;
}

- (void)releaseDragInfo
{
	[self zapDropViewFrame];
	_srcFilter = NULL;
}

- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
	NSPasteboard* pboard = [sender draggingPasteboard];
#ifdef DEBUG_DD
	NSLog(@"draggingEntered: %@", [pboard types]);
#endif
	if (![[pboard types] containsObject:[Filter draggingDataType]])
		return NSDragOperationNone;
	NSPoint pos = [self convertPoint:[sender draggingLocation] fromView:NULL];
	return [self setDropViewFrame:pos];
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
#ifdef DEBUG_DD
	NSLog(@"performDragOperation: %@", sender);
#endif
	NSPoint draggingLocationLocal = [self convertPoint:[sender draggingLocation] fromView:NULL];
	NSUInteger srcIndex = [[self rows] indexOfObject:_srcFilter];
	NSUInteger dstIndex = NSNotFound;
	[self findDropDestinationFilter:draggingLocationLocal withDestinationIndex:&dstIndex];
#ifdef DEBUG_DD
	NSLog(@"index: %d -> %d", srcIndex, dstIndex);
#endif
	if (dstIndex == NSNotFound || srcIndex == NSNotFound || dstIndex == srcIndex || dstIndex == srcIndex + 1) {
		[self releaseDragInfo];
		return NO;
	}
	ServerCommunication* comm = [_srcFilter comm];
	[self removeRow:_srcFilter];
	[self addRow:_srcFilter atIndex:dstIndex - (dstIndex > srcIndex)];
	[self releaseDragInfo];
	NSMutableArray* a = [NSMutableArray arrayWithCapacity:10];
	for (Filter* filter in [self rows])
		[a addObject:[filter filterE]];
	[Xml reorderChilds:a];
	[_prefs markAsDirty];
	[comm saveDirtyPrefsAndNotify];
	return YES;
}

- (BOOL)prepareForDragOperation:(id < NSDraggingInfo >)sender
{
#ifdef DEBUG_DD
	NSLog(@"prepareForDragOperation: %@", sender);
#endif
	return YES;
}

- (NSDragOperation)draggingUpdated:(id < NSDraggingInfo >)sender
{
#ifdef DEBUG_DD
//	NSLog(@"draggingUpdated: %@", sender);
#endif
	NSPoint pos = [self convertPoint:[sender draggingLocation] fromView:NULL];
	return [self setDropViewFrame:pos];
}

- (void)draggingExited:(id < NSDraggingInfo >)sender
{
#ifdef DEBUG_DD
	NSLog(@"draggingUpdated: %@", sender);
#endif
	[self zapDropViewFrame];
}


- (NSDragOperation)draggingSession:(NSDraggingSession *)session sourceOperationMaskForDraggingContext:(NSDraggingContext)context;
{
	return NSDragOperationMove;
}


- (void)startDrag:(Filter*)filter withImage:(NSImage*)image withEvent:(NSEvent*)event
{
	[self releaseDragInfo];
	_srcFilter = filter;
	//
	NSPasteboardItem *pbItem = [NSPasteboardItem new];
	NSString* s = [NSString stringWithFormat:@"%d", [filter filterId]];
	NSData* data = [s dataUsingEncoding:NSUTF8StringEncoding];
	[pbItem setData:data forType:[Filter draggingDataType]];
	//
	NSDraggingItem *dragItem = [[NSDraggingItem alloc] initWithPasteboardWriter:pbItem];
	//
	NSDraggingSession *session = [self beginDraggingSessionWithItems:@[dragItem] event:event source:self];
	session.animatesToStartingPositionsOnCancelOrFail = YES;
	session.draggingFormation = NSDraggingFormationNone;
}

- (void)draggedImage:(NSImage*)anImage endedAt:(NSPoint)aPoint operation:(NSDragOperation)operation
{
#ifdef DEBUG_DD
	NSLog(@"draggedImage:endedAt = %d", operation);
#endif
	[self releaseDragInfo];
}

#pragma mark ------------------- Paste of rules

- (BOOL)validateMenuItem:(NSMenuItem*)item
{
	if ([item action] != @selector(paste:))
		return [super validateMenuItem:item];
	return [[_controller rulesTableView] validateMenuItem:item];
}

- (void)paste:(id)sender
{
	[[_controller rulesTableView] paste:sender];
}

@end
