/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterRuleEditUtils.h"
#import "FilterRuleEdit.h"
#import "Xml.h"
#import "GBDebug.h"

@implementation FilterRuleTextField

- (void)highlightAttributedString:(NSMutableAttributedString*)txt
{
	if (_isRegexp) {
		//DebugNSLog(@"FilterRuleTextField: textDidChange (%d: %d = %d): %@", !!_ruleController, !!storage, !!tv, [self stringValue]);
		[_ruleController syntaxHighlightRegExp:txt];
	} else {
		//NSLog(@"isEnabled = %d: %@", [self isEnabled], [self stringValue]);
		NSColor* color = [self isEnabled] ? [NSColor blackColor] : [NSColor grayColor];
		NSRange range = NSMakeRange(0, [txt length]);
		[txt addAttribute:NSForegroundColorAttributeName value:color range:range];
		[txt addAttribute:NSFontAttributeName value:[self font] range:range];
		[txt removeAttribute:NSParagraphStyleAttributeName range:range];
		[txt removeAttribute:NSUnderlineStyleAttributeName range:range];
		[txt removeAttribute:NSSuperscriptAttributeName range:range];
		[txt removeAttribute:NSBackgroundColorAttributeName range:range];
		[txt removeAttribute:NSLinkAttributeName range:range];
	}
}

- (void)updateTextColor
{
	NSText* t = [self currentEditor];
	if (!t) {
		//DebugNSLog(@"updateTextColor WITHOUT editor: %@", [self stringValue]);
		NSMutableAttributedString* as = [[NSMutableAttributedString alloc] initWithAttributedString:[self attributedStringValue]];
		[self highlightAttributedString:as];
		[self setAttributedStringValue:as];
		return;
	}
	if (![t isKindOfClass:[NSTextView class]]) {
		DebugNSLog(@"FilterRuleTextField.currentEditor isn't NSTextView!");
		return;
	}
	//DebugNSLog(@"updateTextColor with editor: %@", [self stringValue]);
	NSTextView* tv = (NSTextView*)t;
	NSTextStorage* storage = [tv textStorage];
	[self highlightAttributedString:storage];
}

- (void)setRuleController:(FilterRuleEdit*)ctrl
{
	_ruleController = ctrl;
}

- (void)setIsRegExp:(BOOL)isRegexp
{
	BOOL e = [self isEnabled];
	if (isRegexp == _isRegexp && e == _wasEnabled)
		return;
	//DebugNSLog(@"setIsRefExp %d -> %d,  enabled %d -> %d", _isRegexp, isRegexp, _wasEnabled, e);
	_isRegexp = isRegexp;
	_wasEnabled = e;
	[self updateTextColor];
}

- (void)textDidEndEditing:(NSNotification *)notification
{
	//DebugNSLog(@"textDidEndEditing: %@", [self stringValue]);
	[self performSelectorOnMainThread:@selector(updateTextColor)
						   withObject:NULL
						waitUntilDone:NO];
}

- (BOOL)becomeFirstResponder
{
	BOOL became = [super becomeFirstResponder];
	if (became)
		[[self currentEditor] setDelegate:self];
	[self updateTextColor];
	return became;
}

- (BOOL)resignFirstResponder
{
	BOOL resigned = [super resignFirstResponder];
	if (resigned)
		[[self currentEditor] setDelegate:nil];
	return resigned;
}

// prohibit newlines in host/path fields.
- (BOOL)textView:(NSTextView *)aTextView shouldChangeTextInRange:(NSRange)affectedCharRange replacementString:(NSString *)replacementString
{
	if ([_ruleController handlePasteUrl:replacementString withTextField:self])
		return NO;
	NSString* s = replacementString;
	NSUInteger len = [s length];
	NSCharacterSet* cs = [NSCharacterSet newlineCharacterSet];
	for (NSUInteger i = 0; i < len; i++) {
		if ([cs characterIsMember:[s characterAtIndex:i]])
			return NO;
	}
	return YES;
}

- (void)textDidChange:(NSNotification *)aNotification
{
	[self updateTextColor];
	[_ruleController textDidChange:aNotification];
}

#ifdef DEBUG_RESPONDS_TO_SELECTOR
- (BOOL)respondsToSelector:(SEL)aSelector
{
	BOOL rts = [super respondsToSelector:aSelector];
	DebugNSLog(@"FilterRuleTextField, %@ respondsToSelector  %s", (rts ? @"does   " : @"does not"), aSelector);
	return rts;
}
#endif

@end

//==============================================================================

@implementation FilterRuleFieldEditor


- (void)paste:(id)sender
{
	// Pasting as plain text has 2 advantages:
	// 1) The font doesn't change.
	// 2) When the pasteboard contains both rich text and URLs
	//     (e.g. after "copy link" right-click in Safari)
	//     it pastes the URL instead of the rich text description of the URL.
	[super pasteAsPlainText:sender];
}

@end
