/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.admin;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.http.data.UrlParams;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.UserErrorException;
import dk.vitality.lib.foundation.io.FileContentsUtils;

public abstract class AbstractAdminHandler implements Logger
{
	@NotNull public final GlimmerGlobals globals;
	@NotNull public final ProxyRequestHandler req;
	@NotNull public final UrlParams urlParams;
	@Nullable private Logger logger;

	protected AbstractAdminHandler(@NotNull ProxyRequestHandler req) throws Exception
	{
		this.req = req;
		this.globals = req.globals;
		this.urlParams = new UrlParams();
		int idx = req.request.path.indexOf('?');
		if (idx >= 0)
			urlParams.addFromQueryString(req.request.path.substring(idx + 1));
		readWwwFormUrlencoded();
	}

	protected abstract void main() throws Exception;

	public void ln(@Nullable String s)
	{
		if (logger == null)
			logger = new PrefixedLogger(req, getClass().getSimpleName());
		logger.ln(s);
	}

	// ================================================================================
	protected void readWwwFormUrlencoded() throws Exception
	{
		if (!req.request.method.equals("POST"))
			return;
		String ct = req.request.headers.contentType;
		if (ct == null) {
			ln("POST request is missing content-type.");
			return;
		}
		int semicolon = ct.indexOf(';');
		String s = (semicolon >= 0) ? ct.substring(0, semicolon).trim() : ct.trim();
		if (!s.equals("application/x-www-form-urlencoded")) {
			if (globals.debugFlag)
				ln("POST request has bad content-type = '" + ct + "'");
			return;
		}
		String charset = "ISO-8859-1";
		if (semicolon > 0) {
			Matcher m = charsetRE.matcher(ct.substring(semicolon));
			if (m.find())
				charset = m.group(1);
		}
		long contentLength = req.request.requestBodyContentLength;
		if (contentLength <= 0)
			return;
		byte b[] = new byte[(int)contentLength];
		try {
			FileContentsUtils.readFully(req.request.getInputStream(), b);
		}
		catch (IOException ex) {
			throw new UserErrorException("Browser aborted sending of www-form-urlencoded http-request", ex);
		}
		s = new String(b, charset);
		//System.err.println("POSTDATA: " + s);
		urlParams.addFromQueryString(s);
	}

	private static final Pattern charsetRE = Pattern.compile("\\bcharset=([-A-Za-z0-9]+)", Pattern.CASE_INSENSITIVE);
}
