/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.admin.filter;

import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import org.glimmerblocker.proxy.admin.AbstractAdminXmlHandler;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.rule.HostMatchType;
import org.glimmerblocker.proxy.rule.KeywordMatchType;
import org.glimmerblocker.proxy.rule.PathMatchType;
import org.glimmerblocker.proxy.rule.QueryMatchType;
import org.jetbrains.annotations.NotNull;

public class AdminFilterValidate extends AbstractAdminXmlHandler
{
	public AdminFilterValidate(@NotNull ProxyRequestHandler req) throws Exception
	{
		super(req);
	}

	protected void processAdminXml() throws Exception
	{
		validateHost();
		validatePath();
		validateQuery();
		validateKeyword();
	}

	//================================================================================

	private void validateHost() throws Exception
	{
		String host = urlParams.optString("host");
		if (host == null)
			return;
		HostMatchType type = HostMatchType.fromXmlValue(urlParams.reqString("host-type"));
		switch (type) {
			case Ignored:
				return;
			case Domain:
			case Is:
				commit(getPlainHostError(host, type), "host");
				break;
			case Regexp:
				if (globals.debugVerbose)
					ln("Validates host regexp: " + host);
				commit(validateRegExp(host), "host");
		}
	}

	private Error getPlainHostError(String host, HostMatchType hmt)
	{
		if (globals.debugVerbose)
			ln("Validates plain host (" + hmt + "): " + host);
		String what = (hmt == HostMatchType.Domain) ? "Domain name" : "Host name";
		if (hmt == HostMatchType.Is && ipRE.matcher(host).matches())
			return null;
		if (host.charAt(0) == '.')
			return new Error(0, what + " must not start with '.'");
		if (host.charAt(0) == '-')
			return new Error(0, what + " must not start with '-'");
		int len = host.length();
		if (host.charAt(len - 1) == '.')
			return new Error(len - 1, what + " must not end with '.'");
		if (host.charAt(len - 1) == '-')
			return new Error(0, what + " must not end with '-'");
		for (String s : invalidHostSequences) {
			int idx = host.indexOf(s);
			if (idx >= 0)
				return new Error(idx, what + " must not contain '" + s + "'");
		}
		for (int i = 0; i < len; i++) {
			char ch = host.charAt(i);
			if (ch == '.' || ch == '-')
				continue;
			if (ch >= '0' && ch <= '9')
				continue;
			if (ch >= 'a' && ch <= 'z')
				continue;
			if (ch >= 'A' && ch <= 'Z')
				continue;
			return new Error(i, what + " must not contain '" + ch + "'");

		}
		return null;
	}

	private static final Pattern ipRE = Pattern.compile("^\\d+\\.\\d+\\.\\d+\\.\\d+$");

	private static final String[] invalidHostSequences = {
			"..", ".-", "-."  // "--" is used in puny code.
	};

	//================================================================================

	private void validatePath() throws Exception
	{
		String path = urlParams.optString("path");
		if (path == null)
			return;
		if (globals.debugVerbose)
			ln("Validates host: " + path);
		PathMatchType type = PathMatchType.fromXmlName(urlParams.reqString("path-type"));
		switch (type) {
			case Is:
			case StartsWith:
				if (globals.debugVerbose)
					ln("Validates plain path (" + type + "): " + path);
				if (!path.startsWith("/"))
					commit(new Error(0, "Paths always starts with '/'"), "path");
				break;
			case Regexp:
			case NotRegexp:
				if (globals.debugVerbose)
					ln("Validates path regexp: " + path);
				commit(validateRegExp(path), "path");
				return;
			case Contains:
			default:
		}
	}

	//================================================================================

	private void validateQuery() throws Exception
	{
		String query = urlParams.optString("query");
		if (query == null)
			return;
		QueryMatchType hmt = QueryMatchType.fromXmlValue(urlParams.reqString("query-type"));
		switch (hmt) {
			case Regexp:
				if (globals.debugVerbose)
					ln("Validates query regexp: " + query);
				commit(validateRegExp(query), "query");
				return;
			default:
		}
	}

	//================================================================================

	private void validateKeyword() throws Exception
	{
		String keyword = urlParams.optString("keyword");
		if (keyword == null)
			return;
		KeywordMatchType mt = KeywordMatchType.fromXmlValue(urlParams.reqString("keyword-type"));
		switch (mt) {
			case Regexp:
				if (globals.debugVerbose)
					ln("Validates keyword regexp: " + keyword);
				commit(validateRegExp(keyword), "keyword");
				return;
			default:
		}
	}

	//================================================================================

	Error validateRegExp(String s)
	{
		try {
			//noinspection ResultOfMethodCallIgnored
			Pattern.compile(s, Pattern.CASE_INSENSITIVE);
			return null;
		}
		catch (PatternSyntaxException ex) {
			String msg;
			if (ex.getMessage() == null)
				msg = "unknown error";
			else
				msg = ex.getDescription();
			if (globals.debugVerbose)
				ln("  Regexp error at index " + ex.getIndex() + ": " + msg);
			return new Error(ex.getIndex(), "Error: " + msg);
		}
	}

	static class Error
	{
		final int idx;
		final String msg;

		Error(int idx, String msg)
		{
			this.idx = idx;
			this.msg = msg;
		}
	}

	void commit(Error error, String prefix)
	{
		if (error == null)
			return;
		rootE.setAttribute("has-error", "true");
		rootE.setAttribute(prefix + "-error-message", error.msg);
		rootE.setAttribute(prefix + "-index", Integer.toString(error.idx));
	}
}
