/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.BlockerRule;
import org.glimmerblocker.proxy.rule.KeywordRule;
import org.glimmerblocker.proxy.rule.ModificationRule;
import org.glimmerblocker.proxy.rule.RequestRule;
import org.glimmerblocker.proxy.rule.WhitelistRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.StringUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

class FilterLoader extends PrefixedLogger
{
	private final GlimmerGlobals globals;
	private final RuleCollection ruleCollection;
	private int numSets = 0, numRules = 0;

	private int previewFilterId = defaultPreviewFilterId;
	private int previewRuleId = defaultPreviewRuleId;

	private static final int defaultPreviewFilterId = -1;
	private static final int defaultPreviewRuleId = -10;
	private static final int defaultMissingRuleId = -11;

	FilterLoader(@NotNull GlimmerGlobals globals, @NotNull RuleCollection ruleCollection) throws Exception
	{
		super(globals, "filter-loader");
		this.globals = globals;
		this.ruleCollection = ruleCollection;
		loadPreviewRule();
		Set<Integer> seenFilterIds = new HashSet<>();
		for (XmlArgs filterXA : globals.panelSettingsXA.getXmlArgsList("filters/filter")) {
			try {
				Integer filterId = filterXA.getReqInt("@id");
				if (seenFilterIds.contains(filterId)) {
					ln("DUPLICATE FILTER ID: " + filterId);
					continue;
				}
				seenFilterIds.add(filterId);
				boolean enabled = filterXA.getOptBoolean("@enabled");
				if (!enabled)
					continue;
				String url = filterXA.getOptString("@url");
				if (url != null && !filterXA.exists("filter-data")) {
					boolean allowContentModification = "modification".equals(filterXA.getOptString("@safety-policy"));
					Map<Integer, Boolean> userStates = new HashMap<>();
					decodeUserStates(filterXA.getOptString("@rules-enabled"), true, userStates);
					decodeUserStates(filterXA.getOptString("@rules-disabled"), false, userStates);
					XmlArgs filterContainerXA = globals.filterSubscriptions.getOptFilterContainerXA(filterId);
					if (filterContainerXA == null)
						continue;
					XmlArgs filterDataXA = filterContainerXA.getReqXmlArgs("filter-data");
					new FilterRuleLoader(filterId, filterDataXA, userStates, allowContentModification);
				} else {
					XmlArgs filterDataE = filterXA.getReqXmlArgs("filter-data");
					new FilterRuleLoader(filterId, filterDataE, null, true);
				}
			}
			catch (Exception ex) {
				ln("Failed to load filter: " + ExceptionUtils.getExMsg(ex));
				ln(ExceptionUtils.getStackTraceAsString(ex));
				//ln(new DocumentPrinter(filterXA.getDocument()).printToString());
			}
		}
		if (globals.debugFilters)
			ln("Loaded " + numSets + " filters with " + numRules + " rules total.");
	}

	private void loadPreviewRule() throws Exception
	{
		try {
			XmlArgs rulePreviewXA = globals.panelSettingsXA.getOptXmlArgs("rule-preview");
			if (rulePreviewXA == null)
				return;
			XmlArgs ruleXA = rulePreviewXA.getOptXmlArgs("rule");
			if (ruleXA == null)
				return;
			if (globals.debugFilters)
				ln("Loads preview rule.");
			previewFilterId = rulePreviewXA.getOptInt("@for-filter-id", defaultPreviewFilterId);
			previewRuleId = ruleXA.getOptInt("@rule-id", defaultPreviewRuleId);
			RuleLoaderBase rlb = new RuleLoaderBase(false, previewFilterId, defaultPreviewRuleId,
					"(Rule preview)", null, true);
			rlb.parseRule(ruleXA);
			rlb.commitRules();
		}
		catch (Exception ex) {
			ln("Failed to load PREVIEW rule: " + ExceptionUtils.getExMsg(ex));
			ln(ExceptionUtils.getStackTraceAsString(ex));
		}
	}

	private void decodeUserStates(@Nullable String encoded,
	                              @NotNull Boolean value,
	                              @NotNull Map<Integer, Boolean> userStates)
	{
		if (encoded == null)
			return;
		for (String s : StringUtils.splitByChar(encoded, ';'))
			userStates.put(Integer.parseInt(s), value);
	}

	class RuleLoaderBase
	{
		final boolean skipPreviewRule;
		final int filterId;
		final int defaultRuleId;
		@NotNull final String filterName;
		@Nullable final Map<Integer, Boolean> userEnableStates;
		final boolean allowContentModification;

		final List<AbstractRule> ruleList = new ArrayList<>();
		int numModificatorsSkipped = 0;

		RuleLoaderBase(boolean skipPreviewRule,
		               int filterId,
		               int defaultRuleId,
		               @NotNull String filterName,
		               @Nullable Map<Integer, Boolean> userEnableStates,
		               boolean allowContentModification)
		{
			this.skipPreviewRule = skipPreviewRule;
			this.filterId = filterId;
			this.defaultRuleId = defaultRuleId;
			this.filterName = filterName;
			this.userEnableStates = userEnableStates;
			this.allowContentModification = allowContentModification;
		}

		void parseRule(XmlArgs ruleXA) throws Exception
		{
			int ruleId = ruleXA.getOptInt("@rule-id", defaultRuleId);
			boolean isPreview = (ruleId == previewRuleId) && (filterId == previewFilterId);
			if (isPreview && skipPreviewRule) {
				if (globals.debugFilters)
					ln("      rule(" + filterId + "-" + ruleId + ") overridden by preview, thus ignored.");
				return;
			}
			try {
				Boolean enabled = !ruleXA.getOptBoolean("@disabled");
				if (userEnableStates != null) {
					Boolean b = userEnableStates.get(ruleId);
					if (b != null)
						enabled = b;
				}
				if (!enabled)
					return;
				String type = ruleXA.getReqString("@type");
				AbstractRule rule;
				if (type.equals("block")) {
					rule = new BlockerRule(isPreview, filterId, filterName, ruleXA);
				} else if (type.equals("whitelist")) {
					rule = new WhitelistRule(isPreview, filterId, filterName, ruleXA);
				} else if (!allowContentModification) {
					numModificatorsSkipped++;
					return;
				} else if (type.equals("modify")) {
					rule = new ModificationRule(isPreview, filterId, filterName, ruleXA);
				} else if (type.equals("request")) {
					rule = new RequestRule(isPreview, filterId, filterName, ruleXA);
				} else if (type.equals("keyword")) {
					rule = new KeywordRule(isPreview, filterId, filterName, ruleXA);
				} else {
					ln("Unknown rule type = '" + type + "' in " + filterName);
					return;
				}
				if (globals.debugFilters && globals.debugVerbose)
					ln("    Parsed " + rule);
				ruleList.add(rule);
			}
			catch (Throwable ex) {
				ln("Failed to load rule " + ruleId + " in filter " + filterId + ": " + filterName);
				ln("  due to: " + ExceptionUtils.getExMsg(ex));
			}
		}

		void commitRules()
		{
			for (AbstractRule rule : ruleList) {
				ruleCollection.add(rule);
				numRules++;
			}
			numSets++;
		}
	}

	class FilterRuleLoader extends RuleLoaderBase
	{
		FilterRuleLoader(int filterId,
		                 @NotNull XmlArgs filterDataXA,
		                 @Nullable Map<Integer, Boolean> userEnableStates,
		                 boolean allowContentModification) throws Exception
		{
			super(true, filterId, defaultMissingRuleId,
					filterDataXA.getReqString("@name"),
					userEnableStates, allowContentModification);
			for (XmlArgs ruleXA : filterDataXA.getXmlArgsList("rule"))
				parseRule(ruleXA);
			commitRules();
			if (globals.debugFilters) {
				ln("    Loaded " + ruleList.size() + " rules, skipped " +
						numModificatorsSkipped + " content modifying rules.");
			}
		}
	}
}
