/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network;

import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolverConfig;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.misc.StringUtils;

/**
 * Immutable.  Hask-key able.
 */
public class NetworkConfig
{
	public final long date = System.currentTimeMillis();
	@Nullable public final HostnamePort httpProxy;
	@Nullable public final HostnamePort socksProxy;
	@NotNull public final DnsResolverConfig dnsConfig;
	@NotNull public final String interfacesJSON;
	@NotNull public final String locationName;

	@NotNull public static final String emptyStateJSON = "{}";

	public NetworkConfig()
	{
		httpProxy = null;
		socksProxy = null;
		dnsConfig = DnsResolverConfig.NATIVE_RESOLVER;
		interfacesJSON = emptyStateJSON;
		locationName = "*unknown*";
	}

	public NetworkConfig(@Nullable HostnamePort httpProxy,
	                     @Nullable HostnamePort socksProxy,
	                     @NotNull DnsResolverConfig dnsConfig,
	                     @NotNull String interfacesJSON,
	                     @NotNull String locationName)
	{
		this.httpProxy = httpProxy;
		this.socksProxy = socksProxy;
		this.dnsConfig = dnsConfig;
		this.interfacesJSON = interfacesJSON;
		this.locationName = locationName;
	}

	public NetworkConfig withDnsConfig(@NotNull DnsResolverConfig dnsConfig)
	{
		if (dnsConfig.equals(this.dnsConfig))
			return this;
		return new NetworkConfig(httpProxy, socksProxy, dnsConfig, interfacesJSON, locationName);
	}

	public NetworkConfig withHttpProxy(@Nullable HostnamePort httpProxy)
	{
		if (httpProxy == null ? this.httpProxy == null : httpProxy.equals(this.httpProxy))
			return this;
		return new NetworkConfig(httpProxy, socksProxy, dnsConfig, interfacesJSON, locationName);
	}

	public NetworkConfig withSocksProxy(@Nullable HostnamePort socksProxy)
	{
		if (socksProxy == null ? this.socksProxy == null : socksProxy.equals(this.socksProxy))
			return this;
		return new NetworkConfig(httpProxy, socksProxy, dnsConfig, interfacesJSON, locationName);
	}

	public NetworkConfig withInterfacesJSON(@NotNull String interfacesJSON)
	{
		if (interfacesJSON.equals(this.interfacesJSON))
			return this;
		return new NetworkConfig(httpProxy, socksProxy, dnsConfig, interfacesJSON, locationName);
	}

	public NetworkConfig withLocationName(@NotNull String locationName)
	{
		if (locationName.equals(this.locationName))
			return this;
		return new NetworkConfig(httpProxy, socksProxy, dnsConfig, interfacesJSON, locationName);
	}

	public boolean hasFullInterfacesJSON()
	{
		return !interfacesJSON.equals(emptyStateJSON);
	}

	public void logChanges(@NotNull Logger logger,
	                       @NotNull NetworkConfig prev)
	{
		if (prev == this)
			return;
		if (!HostnamePort.equals(prev.httpProxy, httpProxy)) {
			logger.ln("Changed http proxy from " + toString(prev.httpProxy) +
					" to " + toString(httpProxy));
		}
		//
		if (!HostnamePort.equals(prev.socksProxy, socksProxy)) {
			logger.ln("Changed socks proxy from " + toString(prev.socksProxy) +
					" to " + toString(socksProxy));
		}
		//
		if (!prev.dnsConfig.javascriptText.equals(dnsConfig.javascriptText)) {
			logger.ln("Changed DNS servers from '" + prev.dnsConfig.description + "'" +
					" to '" + dnsConfig.description + "'");
		}
		//
		if (!StringUtils.triEquals(prev.locationName, locationName)) {
			logger.ln("Changed location from '" + prev.locationName + "'" +
					" to '" + locationName + "'");
		}
	}

	private String toString(@Nullable HostnamePort hp)
	{
		return (hp != null) ? hp.hostnamePort : "-none-";
	}

	@Override
	public boolean equals(Object o)
	{
		if (this == o)
			return true;
		if (o == null || getClass() != o.getClass())
			return false;
		NetworkConfig k = (NetworkConfig)o;
		return dnsConfig.equals(k.dnsConfig)
				&& (httpProxy != null ? httpProxy.equals(k.httpProxy) : k.httpProxy == null)
				&& (socksProxy != null ? socksProxy.equals(k.socksProxy) : k.socksProxy == null)
				&& interfacesJSON.equals(k.interfacesJSON)
				&& locationName.equals(k.locationName);
	}

	@Override
	public int hashCode()
	{
		int result = (httpProxy != null) ? httpProxy.hashCode() : 0;
		result = 31 * result + (socksProxy != null ? socksProxy.hashCode() : 0);
		result = 31 * result + dnsConfig.hashCode();
		result = 31 * result + interfacesJSON.hashCode();
		result = 31 * result + locationName.hashCode();
		return result;
	}
}
