/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.connection;

import java.io.BufferedInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.net.Socket;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.ByteBuffer;
import dk.vitality.lib.foundation.ref.FatRef;

class HttpResponseHeaderReader implements Logger
{
	@NotNull private final String stateDescPrefix;
	@NotNull private final FatRef<Boolean> hasCompletedSendingRequest;
	@NotNull private final FatRef<String> stateDescRef;
	@NotNull private final BufferedInputStream is;
	private final boolean debugFlag;
	private final Logger parentLogger;
	private String name, value;
	private ByteBuffer bb = new ByteBuffer(50);
	private int headerSize = 0;
	private HttpResponseInfo responseInfo;

	private boolean http09;
	private String keepAliveDisabledReason;
	private final HeadersBuilder hb;
	private String transferEncoding = null;
	private long keepAliveTimeoutSeconds = -1;
	private long contentLength = -1;
	private int crlfCount;

	private static final Pattern keepAliveTimeoutRE = Pattern.compile("\\btimeout=(\\d+)");

	/**
	 * @param is Must be BufferedInputStream as it must support mark/reset.
	 */
	HttpResponseHeaderReader(@NotNull String method,
	                         @NotNull Socket skt,
	                         boolean isQuicktimeRtspPost,
	                         @NotNull BufferedInputStream is,
	                         boolean debugFlag,
	                         Logger logger,
	                         @NotNull String stateDescPrefix,
	                         @NotNull FatRef<Boolean> hasCompletedSendingRequest,
	                         @NotNull FatRef<String> stateDescRef) throws Exception
	{
		this.is = is;
		this.debugFlag = debugFlag;
		this.parentLogger = logger;
		this.stateDescPrefix = stateDescPrefix;
		this.hasCompletedSendingRequest = hasCompletedSendingRequest;
		this.stateDescRef = stateDescRef;
		try {
			if (isQuicktimeRtspPost) {
				// server will never respond, so use long timeout so we just wait until socket is closed by
				// RequestBodySender
				skt.setSoTimeout(24 * 60 * 60 * 1000);
			} else {
				skt.setSoTimeout(30 * 60 * 1000);
				//if (debugFlag)
				//	skt.setSoTimeout(30 * 1000);
			}
		}
		catch (Exception ex) {
			throw new KeepAliveException(ex.getMessage(), ex);
		}
		hb = HeadersBuilder.makeEmpty();
		readStatusLine();
		long startTime = System.currentTimeMillis();
		if (!http09)
			readHeaderLines();
		if (transferEncoding != null && contentLength >= 0) {
			if (debugFlag) {
				ln("Response has specified both Content-Length and Transfer-Encoding which is illegal." +
						" Ignores Content-Length.");
			}
			contentLength = -1;
		}
		if (debugFlag)
			ln("Done reading response header.");
		responseInfo = new HttpResponseInfo(startTime, method, hb.toHeaders(),
				transferEncoding, keepAliveTimeoutSeconds, contentLength);
		stateDescRef.set(stateDescPrefix + "Done reading HTTP response headers.");
	}

	private void readStatusLine() throws Exception
	{
		// some servers send too many LF in keep-alive connections.
		bb.clear();
		is.mark(500);
		String statusLine;
		int emptyLineCount = 0;
		boolean hasSetState = false;
		long lastExceptionTime = 0;
		long startReadTime = System.currentTimeMillis();
		while (true) {
			int ch;
			try {
				ch = read();
			}
			catch (IOException ex) {
				long now = System.currentTimeMillis();
				// timeout exceptions are not properly subclasses, so GB can't reliable
				// detect timeouts versus other socket problems. So just assume, that if
				// the time since last exception is large, then it is a timeout because
				// all other problems (e.g. socket closed) will reported immediately again.
				if (debugFlag) {
					ln("Got IO exception while reading headers: '" + ex.getMessage() + "', " + ex.getClass().getName());
					ln("Has sent all of request: " + hasCompletedSendingRequest.get());
					ln("Total time since start: " + (now - startReadTime) / 1000 + " secs." +
							" Time since last exception: " + (now - lastExceptionTime) / 1000 + " secs.");
				}
				if (!hasCompletedSendingRequest.get() && now - lastExceptionTime > 15 * 1000L) {
					lastExceptionTime = now;
					if (debugFlag)
						ln("Retries read.");
					continue;
				}
				throw new KeepAliveException(ex.getMessage(), ex);
			}
			if (ch < 0) {
				if (bb.length() == 0)
					throw new KeepAliveException("EOF before status line.");
				else
					throw new KeepAliveException("EOF while sending status line.");
			}
			if (ch == '<' && bb.length() == 0) {
				if (debugFlag)
					ln("  Got '<' as 1st char in status line => HTTP/0.9");
				http09 = true;
				statusLine = "HTTP/0.9 200 OK";
				break;
			}
			if (ch == 13) {
				skipLF();
				ch = 10;
			}
			if (ch == 10) {
				if (bb.length() > 0) {
					statusLine = decode();
					break;
				}
				if (++emptyLineCount < 5) {
					is.mark(500);
					continue;
				}
				throw new IOException("Invalid empty-lines response from server when reading http header.");
			}
			if (!hasSetState) {
				stateDescRef.set(stateDescPrefix + "Reads HTTP response status line.");
				hasSetState = true;
			}
			if (ch < 32)
				ch = 32;
			bb.append(ch);
			if (bb.length() == 5) {
				String s = bb.toString("ISO-8859-1");
				if (!s.startsWith("HTTP/") && !s.startsWith("ICY")) {
					if (debugFlag) {
						ln("  Got status doesn't start with HTTP/ nor ICY => HTTP/0.9: '" +
								(s.length() < 250 ? s : s.substring(0, 200) + "...") + "'");
					}
					http09 = true;
					statusLine = "HTTP/0.9 200 OK";
					break;
				}
			}
		}
		if (!http09 && !parseStatusLine(statusLine)) {
			if (debugFlag)
				ln("  Got leading garbage in status line => HTTP/0.9");
			http09 = true;
			statusLine = "HTTP/0.9 200 OK";
		}
		if (http09) {
			killKeepAlive("HTTP/0.9");
			hb.contentType = "text/html";
			is.reset();
		}
		is.mark(0);
		if (debugFlag)
			ln("Response status: '" + statusLine + "'");
	}

	private static final Pattern responseLineRE = Pattern.compile("^HTTP/1\\.(\\d+)\\s+(\\d+)(?:\\s+(.+))?\\s*$");
	private static final Pattern akamaiResponseLineRE = Pattern.compile("^(\\d{3}) (.*?) HTTP/1\\.(\\d+)\\s*$");
	private static final Pattern icyResponseLineRE = Pattern.compile("^ICY\\s+(\\d+)(?:\\s+(.+))?\\s*$");

	private boolean parseStatusLine(String statusLine) throws Exception
	{
		Matcher m = responseLineRE.matcher(statusLine);
		if (m.matches()) {
			boolean http11 = Integer.parseInt(m.group(1)) >= 1;
			if (!http11)
				killKeepAlive("Response is HTTP/1.0");
			hb.status = Integer.parseInt(m.group(2));
			String statusMessage = m.group(3);
			hb.statusMessage = (statusMessage != null) ? statusMessage : "No statusmessage";
			return true;
		}
		m = akamaiResponseLineRE.matcher(statusLine);
		if (m.matches()) {
			boolean http11 = Integer.parseInt(m.group(3)) >= 1;
			if (!http11)
				killKeepAlive("Response is HTTP/1.0");
			hb.status = Integer.parseInt(m.group(1));
			String statusMessage = m.group(2);
			hb.statusMessage = (statusMessage != null) ? statusMessage : "No statusmessage";
			return true;
		}
		m = icyResponseLineRE.matcher(statusLine);
		if (m.matches()) {
			killKeepAlive("Response is ICY");
			hb.icyProtocol = true;
			hb.status = Integer.parseInt(m.group(1));
			String statusMessage = m.group(2);
			hb.statusMessage = (statusMessage != null) ? statusMessage : "No statusmessage";
			return true;
		}
		return false;
	}

	private void readHeaderLines() throws Exception
	{
		stateDescRef.set(stateDescPrefix + "Reads HTTP response headers.");
		while (readSingleHeader()) {
			if (debugFlag)
				ln("  " + name + ": " + value);
			if (name.length() == 8 && name.startsWith("HTTP/1.")) {
				if (debugFlag)
					ln("Invalid http status line within http header ignored.");
				continue;
			}
			String nameLC = name.toLowerCase();
			switch (nameLC) {
				case "content-encoding":
					hb.contentEncoding = value;
					break;
				case "content-type":
					hb.contentType = value;
					break;
				case "content-length":
					String s = zapOptions(value);
					try {
						long len = Long.parseLong(s);
						if (contentLength >= 0) {
							// Twart HTTP Request Smuggling
							// http://www.owasp.org/index.php/HTTP_Request_Smuggling
							if (len != contentLength) {
								ln("Server tried HTTP Request Smuggling: Got multiple Content-Length headers, " +
										len + " and " + contentLength);
							}
							contentLength = Math.max(contentLength, len);
						} else {
							contentLength = len;
						}
					}
					catch (Exception ex) {
						ln("Invalid content-length: '" + s + "': " + ExceptionUtils.getExMsg(ex));
					}
					break;
				case "transfer-encoding":
					if (value.equalsIgnoreCase("identity"))
						transferEncoding = null; // explicit value is just noise as identity is the default.
					else
						transferEncoding = value;
					break;
				case "keep-alive":
					Matcher m = keepAliveTimeoutRE.matcher(value);
					if (m.find())
						keepAliveTimeoutSeconds = Long.parseLong(m.group(1));
					break;
				case "connection":
					if (!zapOptions(value).equalsIgnoreCase("keep-alive"))
						killKeepAlive("Got Connection: " + value);
					break;
				default:
					hb.add(nameLC, value);
					break;
			}
		}
	}

	private boolean readSingleHeader() throws Exception
	{
		// skip (illegal) leading spaces.
		int ch = read();
		while (ch == 32 || ch == '\t')
			ch = read();
		if (ch < 0)
			return false;
		// collect the name.
		bb.clear();
		while (ch >= 0 && ch != 32 && ch != '\t') {
			if (ch == 13) {
				if (!skipLF()) {
					if (crlfCount > 2) {
						// Some, and only some, Akamai servers send invalid headers such as
						// ", 1.1 nwk-www-cache01: 80\r\n\r, 1.1 cup-www-cache02: 80\r\n"
						// Both spaces and \r\n\r are invalid.
						if (debugFlag) {
							ln("Ignores invalid 'CR without LF' header-end." +
									" Has got " + crlfCount + " CRLFs at this point.");
						}
						ch = read();
						continue;
					}
					if (debugFlag)
						ln("Got CR without LF, taken as end-of-header.");
				}
				ch = 10;
			}
			if (ch == 10) {
				if (bb.length() == 0 || !isContinuation())
					break;
				ch = 32; // Section 2.2: All linear white space, including folding, has the same semantics as SP.
			}
			if (ch == ':') {
				if (bb.length() == 0)
					bb.appendLatin1("x-empty-header-name");
				break;
			}
			bb.append(ch <= 32 ? '-' : ch);
			ch = read();
		}
		if (bb.length() == 0)
			return false;
		name = decode();
		while (ch == ' ') // some buggy webservers sends space before colon.
			ch = read();
		if (ch == ':') {
			ch = read();
			// skip space after colon
			while (ch == 32 || ch == '\t')
				ch = read();
		} else if (debugFlag) {
			ln("Invalid http header: name '" + name + "' not terminated by ':'");
		}
		if (ch < 0 || ch == 10) {
			value = "";
			return true;
		}
		// collect the value.
		bb.clear();
		while (ch >= 0) {
			if (ch == 13) {
				skipLF();
				ch = 10;
			}
			if (ch == 10) {
				if (!isContinuation())
					break;
				ch = 32; // Section 2.2: All linear white space, including folding, has the same semantics as SP.
			}
			bb.append(Math.max(32, ch));
			ch = read();
		}
		value = decode();
		return true;
	}

	private boolean isContinuation() throws Exception
	{
		is.mark(1);
		int ch2 = is.read();
		if (ch2 == 32 || ch2 == '\t') {
			// folded header line, rfc 2616, 4.2
			do {
				is.mark(1);
				ch2 = is.read();
			} while (ch2 == 32 || ch2 == '\t');
			is.reset();
			is.mark(0);
			return true;
		}
		is.reset();
		return false;
	}

	private boolean skipLF() throws Exception
	{
		is.mark(2);
		int ch = is.read();
		if (ch == 10) {
			is.mark(0);
			crlfCount++;
			return true;
		}
		if (ch == 13 && is.read() == 10) {
			// squid/2.6.STABLE16 sends invalid http header:
			// Vary: Accept-Language,\r\r\nAccept-Encoding
			is.mark(0);
			crlfCount++;
			return true;
		}
		is.reset();
		return false;
	}

	public void ln(@Nullable String s)
	{
		parentLogger.ln(s);
	}

	@NotNull
	HttpResponseInfo getResponseInfo()
	{
		return responseInfo;
	}

	/**
	 * @return null if keep-alive allowed.
	 */
	@Nullable
	String getKeepAliveDisableReason()
	{
		return keepAliveDisabledReason;
	}

	private String zapOptions(String s)
	{
		int idx = s.indexOf(';');
		return (idx < 0) ? s.trim() : s.substring(0, idx).trim();
	}

	private int read() throws IOException
	{
		int ch = is.read();
		if (ch < 0)
			throw new EOFException("Server closed connection while sending the HTTP header.");
		if (headerSize++ >= 65536)
			throw new IOException("Too huge HTTP response header (64k)");
		return (char)ch;
	}

	private String decode() throws Exception
	{
		return bb.toString("ISO-8859-1");
	}

	void killKeepAlive(String why)
	{
		if (keepAliveDisabledReason != null)
			return;
		if (debugFlag)
			ln("  -  disallows keep-alive: " + why);
		keepAliveDisabledReason = why;
	}
}
