/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.connection;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.Socket;
import javax.net.ssl.SSLSocketFactory;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCacheKey;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;

class HttpSocketConnector implements Logger
{
	@NotNull final Logger parentLogger;
	@NotNull final HttpClientCacheKey key;
	BufferedInputStream is;
	OutputStream os;
	Socket skt;

	HttpSocketConnector(@NotNull Logger parentLogger,
	                    @NotNull InetAddress iaddr,
	                    @NotNull HttpClientCacheKey key,
	                    boolean useCONNECT,
	                    boolean longTimeout,
	                    String debugAddressDesc,
	                    FatRef<String> stateDescRef,
	                    boolean debugFlag) throws Exception
	{
		this.parentLogger = parentLogger;
		this.key = key;
		InetSocketAddress saddr = new InetSocketAddress(iaddr, key.tcpHost.port);
		skt = new Socket(Proxy.NO_PROXY);
		final int timeout = longTimeout ? 20 * 1000 : 5 * 1000;
		try {
			skt.setTcpNoDelay(true);
			skt.setSoTimeout(600 * 1000);
			if (debugFlag)
				ln("Connects to " + debugAddressDesc);
			stateDescRef.set("Connects to " + debugAddressDesc);
			long st0 = System.currentTimeMillis();
			//if (debugHttpClient.get() && (i == 0 || Math.random() > 0.4))
			//	throw new IOException("Bad hair day");
			skt.connect(saddr, timeout);
			if (debugFlag) {
				long dur = System.currentTimeMillis() - st0;
				ln("Got connection to " + debugAddressDesc + " in " + dur + " msecs.");
			}
			if (key.socksProxy != null) {
				makeStreams(key.ssl);
				if (debugFlag)
					ln("makes socks5 handshake with " + debugAddressDesc);
				stateDescRef.set("makes socks5 handshake with " + debugAddressDesc);
				performSocksHandshake();
			}
			if (useCONNECT && key.httpProxy != null) {
				if (is == null)
					makeStreams(key.ssl);
				if (debugFlag)
					ln("Sends CONNECT request to " + debugAddressDesc);
				stateDescRef.set("Sends CONNECT request to " + debugAddressDesc);
				// GB must send a CONNECT request before starting SSL.
				if (key.requestHostPort == null)
					throw new Exception("key.requestHostPort is null");
				send("CONNECT " + key.requestHostPort.hostnamePort + " HTTP/1.1\r\n");
				send("Host: " + key.requestHostPort.hostnamePort + "\r\n");
				send("Connection: close\r\n");
				send("\r\n");
				os.flush();
				if (debugFlag)
					ln("Reads response from http-proxy for CONNECT to " + key.requestHostPort.hostnamePort);
				stateDescRef.set("Waits for CONNECT request from " + debugAddressDesc);
				final boolean isQuicktimeRtspPost = false;
				FatRef<Boolean> hasCompletedSendingRequest = new FatRef<>(true);
				HttpResponseHeaderReader r = new HttpResponseHeaderReader("CONNECT", skt, isQuicktimeRtspPost,
						is, debugFlag, this, "CONNECT request: ",
						hasCompletedSendingRequest, stateDescRef);
				HttpResponseInfo ri = r.getResponseInfo();
				if (ri.headers.status != HttpURLConnection.HTTP_OK) {
					Headers h = new Headers(ri.headers.status,
							"Cannot connect to " + key.requestHostPort.hostnamePort + ", " + ri.headers.statusMessage);
					throw new HttpException(h);
				}
			}
			if (key.ssl) {
				stateDescRef.set("Starts SSL to " + debugAddressDesc);
				if (debugFlag) {
					ln("Starts SSL.");
					st0 = System.currentTimeMillis();
				}
				if (key.requestHostPort == null)
					throw new Exception("key.requestHostPort is null");
				SSLSocketFactory factory = (SSLSocketFactory)SSLSocketFactory.getDefault();
				skt = factory.createSocket(skt, key.requestHostPort.hostname, key.requestHostPort.port, true);
				makeStreams(false);
				if (debugFlag) {
					long dur = System.currentTimeMillis() - st0;
					ln("SSL started in " + dur + " msecs.");
				}
			} else if (is == null) {
				makeStreams(false);
			}
			stateDescRef.set("Fully connected to " + debugAddressDesc);
		}
		catch (Exception ex) {
			if (debugFlag) {
				ln("Connect failed. iaddr=" + iaddr +
						", socketAddr=" + saddr +
						", useCONNECT=" + useCONNECT +
						", ssl=" + key.ssl +
						", timeout=" + timeout +
						", prefer4=" + (System.getProperty("java.net.preferIPv4Stack") != null) +
						", prefer6=" + (System.getProperty("java.net.preferIPv6Addresses") != null) +
						", is4=" + (iaddr instanceof Inet4Address) +
						", is6=" + (iaddr instanceof Inet6Address) +
						"");
			}
			try {
				skt.close();
			}
			catch (Exception ex2) {
				ln("HttpSocketConnector failed closing semi-open socket: " + ExceptionUtils.getExMsg(ex2));
			}
			throw ex;
		}
	}

	private void makeStreams(boolean small) throws Exception
	{
		is = new BufferedInputStream(skt.getInputStream(), small ? 100 : 4096);
		os = new BufferedOutputStream(skt.getOutputStream(), small ? 200 : 4096);
	}

	private void send(String s) throws IOException
	{
		os.write(s.getBytes("UTF-8"));
	}

	// the advantage of doing this manuall instead of using
	// Socket s = new Socket(new Proxy(Proxy.Type.SOCKS, new InetSocketAddress("socks.mydom.com", 1080)));
	// is that we can make sure we don't try to resolve the http-hostname locally, but sends it to
	// the proxy as a name without having performed a DNS lookup.
	void performSocksHandshake() throws Exception
	{
		// http://en.wikipedia.org/wiki/SOCKS
		os.write(5); // socks version
		os.write(1); // one auth method
		os.write(0); // auth method = no auth
		os.flush();
		int r = readSocksByte();
		if (r != 5)
			throw new Exception("Unsupported SOCKS server version: " + (r & 255));
		r = readSocksByte();
		if (r != 0)
			throw new Exception("Unsupported SOCKS auth method = " + r);
		//
		os.write(5); // socks version
		os.write(1); // cmd = tcp connect
		os.write(0); // reserved
		os.write(3); // adress type = domain name
		os.write(key.firstHttpHost.hostname.length()); // hostname
		os.write(key.firstHttpHost.hostname.getBytes("ISO-8859-1"));
		os.write(key.firstHttpHost.port >> 8);
		os.write(key.firstHttpHost.port & 255);
		os.flush();
		//
		r = readSocksByte();
		if (r != 5)
			throw new Exception("Unsupported SOCKS server version: " + (r & 255));
		r = readSocksByte();
		switch (r) {
			case 0:
				break; // connected.
			case 1:
				throw new Exception("SOCKS server said: general failure");
			case 2:
				throw new Exception("SOCKS server said: connection not allowed by ruleset");
			case 3:
				throw new Exception("SOCKS server said: network unreachable");
			case 4:
				throw new Exception("SOCKS server said: host unreachable");
			case 5:
				throw new Exception("SOCKS server said: connection refused by destination host");
			case 6:
				throw new Exception("SOCKS server said: TTL expired");
			case 7:
				throw new Exception("SOCKS server said: command not supported / protocol error");
			case 8:
				throw new Exception("SOCKS server said: address type not supported");
			default:
				throw new Exception("SOCKS server returned error code " + r);
		}
		r = readSocksByte();
		if (r != 0)
			ln("SOCKS server returned non-zero byte after status byte.");
		r = readSocksByte();
		int skip;
		switch (r) {
			case 1:
				skip = 4;
				break;
			case 3:
				skip = readSocksByte();
				break;
			case 4:
				skip = 6;
				break;
			default:
				throw new Exception("SOCKS server returned unknown address type = " + r);
		}
		for (int i = 0; i < skip; i++)
			readSocksByte();
		int p2 = readSocksByte() << 8;
		p2 |= readSocksByte();
		if (p2 < 0)
			ln("SOCKS port: " + p2);

	}

	private int readSocksByte() throws IOException
	{
		int x = is.read();
		if (x < 0)
			throw new EOFException("SOCKS server closed connection in initial handshake.");
		return x;
	}

	public void ln(@Nullable String s)
	{
		parentLogger.ln(s);
	}
}
