/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.connection;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.stream.ChunkedInputStream;
import org.glimmerblocker.proxy.http.client.stream.DevNullInputStream;
import org.glimmerblocker.proxy.http.client.stream.FixedLengthInputStream;
import org.glimmerblocker.proxy.http.client.stream.RequestData;
import org.glimmerblocker.proxy.http.client.stream.SocketTimeoutableInputStream;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.ArrayUtils;
import dk.vitality.lib.foundation.ref.FatRef;

public class HttpTransaction implements Logger
{
	@NotNull final HttpClientConnection conn;
	@NotNull final RequestData requestData;
	@NotNull Logger logger;
	final long transactionStartTime = System.currentTimeMillis();
	final boolean isQuicktimeUserAgent;
	final boolean isQuicktimeRtspPost;
	boolean isFailedQuicktimeRtspPost;
	final boolean hasExpect100Continue;

	volatile boolean hasStartedSendingRequest; // read by other thread when getting state.
	final FatRef<Boolean> hasCompletedSendingRequest = new FatRef<>(false);

	volatile boolean hasReadResponseHeaders; // checked by body sender
	boolean isReadingResponseBody, hasReadResponseBody;
	boolean isDone;
	private boolean allowKeepAlive;
	boolean debugHttpClient;
	final FatRef<String> transactionStateDescRef = new FatRef<>("virgin");

	HttpResponseInfo responseInfo;

	/**
	 * Initial "HTTP/1.1 100 Continue" before the proper response.
	 */
	HttpResponseInfo continueResponseInfo;

	public HttpTransaction(@NotNull HttpClientConnection conn,
	                       @NotNull RequestData requestData,
	                       @NotNull Logger logger,
	                       boolean debugHttpClient)
	{
		this.conn = conn;
		this.requestData = requestData;
		this.logger = logger;
		this.debugHttpClient = debugHttpClient;
		this.isQuicktimeRtspPost = isQuicktimeRtspPost(requestData);
		this.allowKeepAlive = true;
		String s = requestData.headers.getLC("expect");
		hasExpect100Continue = (s != null) && s.toLowerCase().startsWith("100-continue");
		s = requestData.headers.getLC("user-agent");
		isQuicktimeUserAgent = (s != null) && s.startsWith("QuickTime");
	}

	public void sendRequest() throws Exception
	{
		if (hasStartedSendingRequest)
			throw new Exception("Has already sent request");
		hasStartedSendingRequest = true;
		if (isQuicktimeRtspPost)
			killKeepAlive("Is QuickTime RTSP POST which never has any reply.");
		transactionStateDescRef.set("Sends HTTP request");
		try {
			new HttpTransactionSender(this);
		}
		catch (IOException ex) {
			throw new KeepAliveException(ExceptionUtils.getExMsg(ex) + " while sending request.", ex);
		}
		transactionStateDescRef.set("Done sending HTTP request");
	}

	@NotNull
	public HttpResponseInfo getResponseHeaderInfo() throws Exception
	{
		if (responseInfo == null)
			makeResponseInfo();
		return responseInfo;
	}

	@Nullable
	public HttpResponseInfo getOpt100ContinueResponseHeaderInfo() throws Exception
	{
		getResponseHeaderInfo();
		return continueResponseInfo;
	}

	private void makeResponseInfo() throws Exception
	{
		transactionStateDescRef.set("Waits for response");
		readResponseHeader();
		// http://haacked.com/archive/2004/05/15/http-web-request-expect-100-continue.aspx
		while (responseInfo.headers.status == 100) {
			if (debugHttpClient) {
				if (hasExpect100Continue) {
					ln("Got expected 100-continue header.  Reads next header from server.");
				} else {
					// Bug in IIS: sends "100 Continue" without "Expect:" request header.
					// It continues with sending the proper response, so just eat the 100-continue.
					ln("Got 100-Continue without having sent 'expect:' header, thus reads next http response header.");
				}
			}
			if (hasExpect100Continue)
				continueResponseInfo = responseInfo;
			transactionStateDescRef.set("Waits for response HTTP headers after '100 Continue'");
			readResponseHeader();
		}
		String transferEncoding = responseInfo.transferEncoding;
		//ln("makeResponseInfo: canHaveNonEmptyBody = " + responseInfo.canHaveNonEmptyBody +
		//		", transferCoding=" + transferEncoding +
		//		", contentLength=" + responseInfo.contentLength);
		if (!responseInfo.canHaveNonEmptyBody) {
			hasReadResponseBody = true;
			if (transferEncoding != null && !transferEncoding.equalsIgnoreCase("identity")) {
				if (debugHttpClient)
					ln("Response has transfer-encoding in content-less response type. Tries to read empty response");
				InputStream is = getTransferEncodingInputStream();
				//noinspection StatementWithEmptyBody
				while (is != null && is.read() >= 0) {
					//
				}
				if (debugHttpClient)
					ln("Done reading empty response");
			}
		} else if (transferEncoding == null && responseInfo.contentLength < 0) {
			killKeepAlive("No transfer-encoding and no content-length header and not body-less response.");
		}
		boolean doneSendingRequest = hasCompletedSendingRequest.get();
		hasReadResponseHeaders = true;
		if (!doneSendingRequest) {
			killKeepAlive("Got reponse headers before all of request body has been sent.");
			for (int i = 0; i < 1000; i++) {
				if (hasCompletedSendingRequest.get())
					break;
				Thread.sleep(50);
			}
			if (debugHttpClient) {
				ln("Done waiting for request body sender to finish." +
						" Completed = " + hasCompletedSendingRequest.get() + ".");
			}
		}
		checkTransactionDone();
	}

	private void readResponseHeader() throws Exception
	{
		try {
			HttpResponseHeaderReader r = new HttpResponseHeaderReader(requestData.method, conn.skt,
					isQuicktimeRtspPost, conn.is, debugHttpClient, this, "",
					hasCompletedSendingRequest, transactionStateDescRef);
			String s = r.getKeepAliveDisableReason();
			if (s != null)
				killKeepAlive(s);
			responseInfo = r.getResponseInfo();
			if (isQuicktimeUserAgent
					&& responseInfo.headers.status == HttpURLConnection.HTTP_UNAUTHORIZED
					&& requestData.url.getHost().equals("publish.me.com")) {
				killKeepAlive("publish.me.com always drops the connection after a 401 response" +
						" even when it sends connection:keep-alive.");
			}
		}
		catch (KeepAliveException ex) {
			if (isQuicktimeRtspPost) {
				isFailedQuicktimeRtspPost = true;
				if (debugHttpClient)
					ln("Fakes RTSP POST response content as server closed connection.");
				// Fake an empty response instead of retrying the POST request.
				final Headers h = new Headers(HttpURLConnection.HTTP_OK, "OK");
				responseInfo = new HttpResponseInfo(System.currentTimeMillis(), "POST", h, null, 0, 0);
				return;
			}
			endTransaction();
			throw ex;
		}
	}

	@Nullable
	public TimeoutableInputStream getBodyStream() throws IOException
	{
		if (responseInfo.isContentLessResponse)
			throw new IOException("getBodyStream for content-less response.");
		if (isReadingResponseBody)
			throw new IOException("Is already reading response body");
		isReadingResponseBody = true;
		if (isFailedQuicktimeRtspPost) {
			if (debugHttpClient)
				ln("Returns empty body stream for failed RTSP POST");
			hasReadResponseBody = true;
			checkTransactionDone();
			return new SocketTimeoutableInputStream(new ByteArrayInputStream(ArrayUtils.emptyByteArray), conn.skt);
		}
		transactionStateDescRef.set("Reads response body");
		InputStream is = getTransferEncodingInputStream();
		if (is != null) {
			// do not add FixedLengthInputStream on top of ChunkedInputStream
			// as the latter then won't read the complete output stream.
			return new SocketTimeoutableInputStream(is, conn.skt);
		}
		final long size = responseInfo.contentLength < 0 ? Long.MAX_VALUE : responseInfo.contentLength;
		if (size == 0) {
			hasReadResponseBody = true;
			is = new DevNullInputStream();
			return new SocketTimeoutableInputStream(is, conn.skt);
		}
		is = new FixedLengthInputStream(conn.is, size)
		{
			protected void notifyReachedEndOfExpectedSize()
			{
				hasReadResponseBody = true;
				checkTransactionDone();
			}

			protected void handleUnexpectedEndOfStream() throws IOException
			{
				if (responseInfo.contentLength >= 0) {
					super.handleUnexpectedEndOfStream();
					return;
				}
				hasReadResponseBody = true;
				checkTransactionDone();
			}
		};
		return new SocketTimeoutableInputStream(is, conn.skt);
	}

	@Nullable
	private InputStream getTransferEncodingInputStream() throws IOException
	{
		String s = responseInfo.transferEncoding;
		if (s != null)
			s = s.trim().toLowerCase();
		if (s == null || s.equals("identity"))
			return null;
		if (s.equals("chunked")) {
			if (debugHttpClient)
				ln("Chunked transfer-encoding.");
			return new ChunkedInputStream(conn.is, conn.skt, this)
			{
				@Override
				protected void notifyReachedEndOfExpectedSize() throws IOException
				{
					hasReadResponseBody = true;
					checkTransactionDone();
					super.notifyReachedEndOfExpectedSize();
				}

				@Override
				protected void handleUnexpectedEndOfStream() throws IOException
				{
					killKeepAlive("Got EOF from server inside chunked stream");
					hasReadResponseBody = true;
					checkTransactionDone();
					super.handleUnexpectedEndOfStream();
				}
			};
		}
		throw new IOException("Can't decode transfer-encoding: '" + responseInfo.transferEncoding + "'");
	}


	public void endTransaction()
	{
		if (isDone)
			return;
		transactionStateDescRef.set("Transaction done");
		isDone = true;
		if (!allowKeepAlive) {
			conn.transactionDone(0);
			return;
		}
		if (!hasCompletedSendingRequest.get()) {
			killKeepAlive("EndTransaction: Isn't done sending request body.");
			conn.transactionDone(0);
			return;
		}
		if (responseInfo == null) {
			killKeepAlive("EndTransaction: Hasn't read response headers.");
			conn.transactionDone(0);
			return;
		}
		if (!hasReadResponseBody) {
			killKeepAlive("EndTransaction: Hasn't read (all of) response body.");
			conn.transactionDone(0);
			return;
		}
		if (responseInfo.keepAliveTimeoutSeconds == 0) {
			killKeepAlive("Got 'Keep-Alive: timeout=0' header.");
			conn.transactionDone(0);
			return;
		}
		long timeout = responseInfo.keepAliveTimeoutSeconds * 1000;
		if (timeout < 0)
			timeout = 10 * 1000L;
		conn.transactionDone(responseInfo.startTime + timeout);
	}

	void checkTransactionDone()
	{
		if (!isDone && hasCompletedSendingRequest.get() && hasReadResponseBody)
			endTransaction();
	}


	void killKeepAlive(String why)
	{
		if (!allowKeepAlive)
			return;
		if (debugHttpClient)
			ln("  -  disallows keep-alive: " + why);
		allowKeepAlive = false;
	}

	boolean allowKeepAlive()
	{
		return allowKeepAlive;
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	/**
	 * <h2>How RTSP-over-HTTP works using QuickTime:</h2>
	 * (as observed by the behavior of QuickTime and Darwin Streaming Server)
	 * <p/>
	 * The client opens one GET request which is used to receive all data from the server.
	 * It keeps this connection during the full streaming session.
	 * The request must contain an "x-sessioncookie" header with a unique value. If the
	 * value is already known by the server (e.g. from an earlier request), the GET request fails
	 * without proper status code nor message.
	 * <p/>
	 * After the GET request has been sent, the client issues a POST request each time it wants
	 * to send a command to the server.  The request must contain the same "x-sessioncookie" header
	 * as the GET request.
	 * The RTSP command is base64 encoded and sent as the request-content.
	 * It does not use the standard URL-arg encoding.<br>
	 * When the server has received the POST content, it just keeps the connection open
	 * <i>and never sends a response nor closes the connection</i>. The client has thus to close
	 * the connection, very much against normal http pratice.
	 * <p/>
	 * A GET request looks like:
	 * <pre>
	 * GET http://streamer-01.dr.nordija.dk:80/dr1lowmp4 HTTP/1.0
	 * User-Agent: QuickTime/7.4.5 (qtver=7.4.5;cpu=IA32;os=Mac 10.5.2)
	 * x-sessioncookie: upYRxLgBAAD8fgkBPYcBAQ9
	 * Accept: application/x-rtsp-tunnelled
	 * Pragma: no-cache
	 * Cache-Control: no-cache
	 * </pre>
	 * and a POST request like:
	 * <pre>
	 * POST http://streamer-01.dr.nordija.dk:80/dr1lowmp4 HTTP/1.0
	 * User-Agent: QuickTime/7.4.5 (qtver=7.4.5;cpu=IA32;os=Mac 10.5.2)
	 * x-sessioncookie: upYRxLgBAAD8fgkBPYcBAQ9
	 * Content-Type: application/x-rtsp-tunnelled
	 * Pragma: no-cache
	 * Cache-Control: no-cache
	 * Content-Length: 32767
	 * Expires: Sun, 9 Jan 1972 00:00:00 GMT
	 * <b></b>
	 * REVTQ1JJQkUgcnRzcDovL3N0cmVhbWVyLTAxLmRyLm5vcmRpamEuZGs6ODAvZHIxbG93bXA0IFJUU1Av
	 * MS4wDQpDU2VxOiAxDQpBY2NlcHQ6IGFwcGxpY2F0aW9uL3NkcA0KQmFuZHdpZHRoOiAyMTQ3NDgzNjQ3
	 * DQpBY2NlcHQtTGFuZ3VhZ2U6IGVuLVVTDQpVc2VyLUFnZW50OiBRdWlja1RpbWUvNy40LjUgKHF0dmVy
	 * PTcuNC41O2NwdT1JQTMyO29zPU1hYyAxMC41LjIpDQoNCg==
	 * </pre>
	 * <a href='http://developer.apple.com/documentation/QuickTime/QTSS/Concepts/chapter_2_section_14.html'>Apple's documentation</a>
	 * says that the client might send several requests thru the same POST request.
	 * This happends mostly when changing location in the movie, e.g. seeking.
	 * <p/>
	 * The documentation states that the content-length header of the request
	 * is bogous, so we can't depend on being able to read that many bytes from the client.
	 * The client will asynchronously send request data, and just close the connection when done.
	 * <p/>
	 * So the GET conforms to standard http practice (assuming proper sessioncookie) while the POST doesn't.
	 * <p/>
	 * VLC doesn't have this problem as it uses a very different pattern of requests.
	 * <p/>
	 * http://developer.apple.com/quicktime/icefloe/dispatch028.html
	 */
	private static boolean isQuicktimeRtspPost(@NotNull RequestData requestData)
	{
		return requestData.method.equals("POST")
				&& requestData.headers.isContentTypeBaseLC("application/x-rtsp-tunnelled");
	}
}
