/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.util.List;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.HostnamePort;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCacheKey;
import org.glimmerblocker.proxy.http.client.connection.HttpClientConnection;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.BlockerRule;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;

class ConnectMethodHandler implements Logger
{
	@NotNull private final GlimmerGlobals globals;
	@NotNull private final HttpServerRequest req;
	private final int remotePortNumber;
	@NotNull private final String remoteHostname;
	//final Socket skt;

	ConnectMethodHandler(@NotNull HttpServerRequest req) throws Exception
	{
		this.globals = req.globals;
		this.req = req;
		if (globals.debugFlag)
			ln("CONNECT: " + req.path);
		req.conn.disableKeepalive("Method = CONNECT");
		int colon = req.path.lastIndexOf(':');
		if (colon < 0) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST, "Missing colon/portnumber");
			throw new HttpException(h);
		}
		try {
			String s = req.path.substring(colon + 1);
			int slash = s.indexOf('/'); // some clients sends invalid "hostname:port/some/path/here"
			if (slash > 0)
				s = s.substring(0, slash);
			remotePortNumber = Integer.parseInt(s);
		}
		catch (Throwable ex) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST, "Invalid portnumber");
			throw new HttpException(h, ex);
		}
		try {
			String hn = req.path.substring(0, colon);
			if (hn.equalsIgnoreCase("ihost.netflix.com")) {
				if (globals.debugFlag)
					ln("Uses client addr as destination instead of '" + hn + "'.");
				hn = req.getClientAddress().getHostAddress();
			}
			remoteHostname = hn;
		}
		catch (Throwable ex) {
			Headers h = new Headers(HttpURLConnection.HTTP_GATEWAY_TIMEOUT, "Unknown host: " + req.path);
			throw new HttpException(h, ex);
		}
	}

	@Nullable
	Headers getBlockedHeaders()
	{
		if (remotePortNumber != 443)  // don't try block non-https requests.
			return null;
		List<AbstractRule> ruleList = globals.ruleCollection.getRules(remoteHostname, "/https-proxy", "", false);
		for (AbstractRule r : ruleList) {
			if (r instanceof BlockerRule) {
				HistoryEntryBuilder heb = globals.historyCache.makeBuilder("https://" + remoteHostname, null);
				heb.addAction(r, "blocked");
				heb.commit();
				return new Headers(HttpURLConnection.HTTP_FORBIDDEN,
						remoteHostname + " blocked by rule");
			}
		}
		return null;
	}

	void main() throws Exception
	{
		String oldThreadName = Thread.currentThread().getName();
		try {
			Thread.currentThread().setName("CONNECT " + req.path + " server->client sender");
			internalMain();
		}
		finally {
			Thread.currentThread().setName(oldThreadName);
		}
	}

	private void internalMain() throws Exception
	{
		if (globals.debugHttpServer)
			ln("Port: " + remotePortNumber);
		if (!isPortAllowed()) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST,
					"Port " + remotePortNumber + " not allowed in CONNECT: " + req.path);
			throw new HttpException(h);
		}
		NetworkConfig networkConfig = globals.networkManager.getConfig();
		HttpClientCacheKey key = new HttpClientCacheKey(
				networkConfig.dnsConfig,
				networkConfig.httpProxy,
				networkConfig.socksProxy,
				new HostnamePort(remoteHostname, remotePortNumber),
				false);
		FatRef<Integer> ipAddressIndex = new FatRef<>(0);
		HttpClientConnection conn;
		try {
			conn = new HttpClientConnection(globals.dnsResolver, key, true, this,
					ipAddressIndex, globals.debugHttpClient);
			conn.skt.setSoTimeout(60 * 60 * 1000); // 1 hour.
		}
		catch (Throwable ex) {
			ln("Got connect exception: " + ExceptionUtils.getExMsg(ex));
			//ln(ExceptionUtils.getStackTraceAsString(ex));
			Headers h = new Headers(HttpURLConnection.HTTP_GATEWAY_TIMEOUT, "Can't connect to host: " + req.path);
			throw new HttpException(h, ex);
		}
		if (globals.debugHttpServer)
			ln("Did get connection.");
		req.sendResponseHeaders(new Headers(HttpURLConnection.HTTP_OK, "Did get connection"), -1);
		req.conn.os.flush();
		if (globals.debugHttpServer)
			ln("Starts sender thread");
		req.conn.is.setReadTimeout(60 * 60 * 1000); // 1 hour.
		Thread t = new Thread(new Sender(conn.os), "CONNECT " + req.path + " client->server sender");
		t.start();
		//
		long sumReceived = 0;
		try {
			byte[] buffer = new byte[4096];
			while (true) {
				int num = conn.is.read(buffer);
				if (num < 0) {
					if (globals.debugHttpServer)
						ln("Receiver got EOF");
					break;
				}
				if (globals.debugHttpServer)
					ln("Receiver writes " + num + " bytes server->client.");
				req.conn.os.write(buffer, 0, num);
				req.conn.os.flush();
				sumReceived += num;
			}
		}
		catch (Throwable ex) {
			ln("receiver got exception: " + ExceptionUtils.getExMsg(ex));
		}
		if (globals.debugHttpServer)
			ln("Receiver read " + sumReceived + " bytes total.");
		try {
			conn.skt.close();
		}
		catch (Throwable ex) {
			ln("Got exception when closing remote socket: " + ExceptionUtils.getExMsg(ex));
		}
		try {
			req.conn.skt.shutdownInput();
			conn.forcedSocketClose();
		}
		catch (Throwable ex) {
			ln("Got exception when closing client socket: " + ExceptionUtils.getExMsg(ex));
		}
		t.join();
		if (globals.debugHttpServer)
			ln("Sender thread done.");
	}

	private boolean isPortAllowed()
	{
		// http://httpd.apache.org/docs/2.2/mod/mod_proxy.html#AllowCONNECT
		switch (remotePortNumber) {
			case 80: // http
			case 83:// often alternative port for https.
			case 311: // Apple Server Admin (https)
			case 443: // https
			case 631: // IPP
				return true;
			case 2049: // common IRC ports
			case 3659: // http://trac.webkit.org/browser/trunk/LayoutTests/security/block-test.html
			case 4045: // WebKit disallowes a bunch of ports below 1024 instead of having a white-list.
			case 6000:
			case 6665:
			case 6666:
			case 6667:
			case 6668:
			case 6669:
				return false;
			default:
				return remotePortNumber >= 1024;
		}
	}

	private class Sender implements Runnable
	{
		private final OutputStream os;

		Sender(OutputStream os)
		{
			this.os = os;
		}

		public void run()
		{
			long sumSent = 0;
			try {
				byte[] buffer = new byte[4096];
				while (true) {
					int num = req.conn.is.read(buffer);
					if (num <= 0) {
						if (globals.debugHttpServer)
							ln("Sender got EOF");
						break;
					}
					if (globals.debugHttpServer)
						ln("Sender writes " + num + " bytes client->server.");
					os.write(buffer, 0, num);
					os.flush();
					sumSent += num;
				}
			}
			catch (Throwable ex) {
				if (globals.debugHttpServer)
					ln("sender got exception: " + ExceptionUtils.getExMsg(ex));
			}
			if (globals.debugHttpServer)
				ln("Sender wrote " + sumSent + " bytes total.");
		}
	}

	public void ln(@Nullable String s)
	{
		req.ln(s);
	}
}
