/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest.errorPage;

import java.net.ConnectException;
import java.net.MalformedURLException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.UnknownHostException;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.util.HttpException;
import org.jabhts.annotation.attribute.content.AHref;
import org.jabhts.annotation.attribute.content.ASrc;
import org.jabhts.annotation.attribute.content.AValue;
import org.jabhts.annotation.attribute.sub.SHasClass;
import org.jabhts.annotation.element.EText;
import org.jabhts.annotation.flowControl.EIf;
import org.jabhts.annotation.global.TCompile;
import org.jabhts.runtime.html.HtmlPrintStringWriter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.html.URLUtils;

@TCompile
public class ErrorPage
{
	@NotNull final GlimmerGlobals globals;
	@NotNull final HttpException error;
	@Nullable final URL requestURL;

	final boolean webkit;

	public ErrorPage(@NotNull GlimmerGlobals globals,
	                 @NotNull HttpException error,
	                 @Nullable String userAgent,
	                 @Nullable URL requestURL) throws MalformedURLException
	{
		this.globals = globals;
		this.error = error;
		this.requestURL = requestURL;
		webkit = (userAgent != null) && userAgent.contains("WebKit");
	}

	public ResultData getResultData() throws Exception
	{
		HtmlPrintStringWriter pw = new HtmlPrintStringWriter();
		new ErrorPageTp(this).emit(pw);
		pw.flush();
		Headers headers = new Headers(error.headers.status, error.headers.statusMessage,
				"text/html; charset=UTF-8", null, null);
		return new ResultData(headers, pw.toString());
	}

	@SHasClass(tag = "body", classes = "WebKit")
	boolean isWebKit()
	{
		return webkit;
	}

	@ASrc
	String srcLogo()
	{
		return globals.adminDispatcher.getPublicUrlBase() + "/logo/large.png";
	}

	@EText(tag = "title")
	String pagetitle()
	{
		String s = "Failed to load page: " + getTitle();
		if (requestURL != null && requestURL.getHost() != null && !requestURL.getHost().isEmpty())
			s = requestURL.getHost() + ": " + s;
		return s;
	}

	@EText
	String getTitle()
	{
		Throwable cause = error.getCause();
		if (cause instanceof ConnectException)
			return "The server denies connections.";
		if (cause instanceof SocketTimeoutException)
			return "The server does not respond at all.";
		if (cause instanceof UnknownHostException)
			return "The host does not exist.";
		return "Request failed.";
	}

	@EText
	String getMessage()
	{
		Throwable cause = error.getCause();
		if (cause instanceof ConnectException)
			return "The server might be too busy or might be disabled. Try again a bit later.";
		if (cause instanceof SocketTimeoutException)
			return "Try again a bit later.";
		if (cause instanceof UnknownHostException)
			return "Check the URL or try googling for it.";
		return error.getMessage();
	}

	@EIf({"SuggestionContainer", "OtherTools"})
	boolean hasUrl()
	{
		return requestURL != null && requestURL.getHost() != null && requestURL.getHost().length() > 0;
	}

	@AValue
	String valueSuggestionInput()
	{
		assert requestURL != null;
		return requestURL.getHost();
	}

	@AHref
	String hrefGoogleCacheLink()
	{
		assert requestURL != null;
		return "http://www.google.com/search?q=cache:" + URLUtils.encodeURL(requestURL.toExternalForm());
	}

	@AHref
	String hrefCoralizeLink()
	{
		assert requestURL != null;
		StringBuilder sb = new StringBuilder(100);
		sb.append("http://");
		sb.append(requestURL.getHost());
		if (requestURL.getPort() > 80)
			sb.append('.').append(requestURL.getPort());
		sb.append(".nyud.net:8080");
		String p = requestURL.getPath();
		sb.append(p != null && p.length() > 0 ? p : "/");
		return sb.toString();
	}

	@AHref
	String hrefWaybackLink()
	{
		assert requestURL != null;
		return "http://web.archive.org/web/*/" + requestURL;
	}

	@AHref
	String hrefWhoIsLink()
	{
		assert requestURL != null;
		return "http://whois.domaintools.com/" + requestURL.getHost();
	}

	@AHref
	String hrefTracerouteLink()
	{
		assert requestURL != null;
		return "http://private.dnsstuff.com/tools/tracert.ch?ip=" + requestURL.getHost();
	}

	@AHref
	String hrefIsItJustMe()
	{
		assert requestURL != null;
		return "http://www.downforeveryoneorjustme.com/" + requestURL.getHost();
	}
}

