/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.callback;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.jetbrains.annotations.NotNull;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;

public abstract class CallbackObjectBase extends ScriptableObject implements Logger
{
	public final GlimmerGlobals globals;
	public final Scriptable scope;
	@NotNull public final Scriptable prototype;
	@NotNull public CallbackRhinoBinder binder;

	protected Context ctx;

	private boolean bound;
	private List<CallbackObjectBase> subObjects = Collections.emptyList();

	private DebugSubObject debugSubObject;

	protected CallbackObjectBase(@NotNull CallbackObjectBase parent,
	                             @NotNull CallbackRhinoBinder binder)
	{
		this.globals = parent.globals;
		this.scope = parent.scope;
		this.prototype = parent.prototype;
		this.binder = binder;
		setPrototype(prototype);
		if (parent.subObjects.isEmpty())
			parent.subObjects = new ArrayList<>();
		parent.subObjects.add(this);
	}

	protected CallbackObjectBase(@NotNull Scriptable scope,
	                             @NotNull Scriptable prototype,
	                             @NotNull CallbackRhinoBinder binder,
	                             @NotNull GlimmerGlobals globals)
	{
		this.scope = scope;
		this.prototype = prototype;
		this.binder = binder;
		this.globals = globals;
	}

	public void bind()
	{
		binder.bind(this);
		bound = true;
		for (CallbackObjectBase cob : subObjects)
			cob.bind();
	}

	protected void prepareObjectBase(@NotNull Context ctx) throws Exception
	{
		if (!bound)
			throw new Exception("Not bound!");
		this.ctx = ctx;
		for (CallbackObjectBase cob : subObjects)
			cob.prepareObjectBase(ctx);
	}

	public void notifyLeftContext()
	{
		for (CallbackObjectBase cob : subObjects)
			cob.notifyLeftContext();
		ctx = null;
	}

	public void checkInsideScope()
	{
		if (ctx == null)
			throw Context.reportRuntimeError("Called outside scope, does not have context.");
	}

	@Override
	public String getClassName()
	{
		return getClass().getSimpleName();
	}

	public void put(String name, Scriptable start, Object value)
	{
		//System.err.println("put(" + name + ")");
		if (!binder.boundNames.contains(name))
			throw Context.reportRuntimeError("This callback object has no '" + name + "' property. Check spelling!");
		super.put(name, start, value);
	}

	public void put(int index, Scriptable start, Object value)
	{
		throw Context.reportRuntimeError("This callback object is not an array!");
	}

	@RhinoCallback
	public DebugSubObject jsGet_debug() throws Exception
	{
		if (debugSubObject == null) {
			debugSubObject = new DebugSubObject(this);
			debugSubObject.bind();
			debugSubObject.prepareDebugSubObjectBase(ctx);
		}
		return debugSubObject;
	}
}
