/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rule;

import java.util.regex.Pattern;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;
import static java.util.regex.Pattern.CASE_INSENSITIVE;

public abstract class AbstractRule implements Comparable<AbstractRule>
{
	/**
	 * Is this a preview version of the rule.
	 */
	public final boolean isPreview;

	/**
	 * The name of the filter which contains this rule.
	 */
	@NotNull public final String filterName;

	/**
	 * The id' xml attribute in the &lt;filter&gt; element.
	 */
	public final int filterId;

	/**
	 * The rule-id' xml attribute in the &lt;rule&gt; element.
	 */
	public final int ruleId;

	/**
	 * User sort priority
	 */
	public final int userPriority;

	/**
	 * Rule type
	 */
	public final RuleType ruleType;

	/**
	 * least specific first (lowest ordinal()), and most specific last
	 */
	protected enum RuleType
	{
		Blocker, Whitelist, Request, Modification, Keyword
	}

	/**
	 * Only for iPhone/iPod clients?
	 */
	public final boolean mobileOnly;

	//=========================================================================

	/**
	 * The host/domain, if any.
	 */
	@Nullable public final String host;

	/**
	 * Host/domain as lowercase.
	 */
	@Nullable public final String hostLC;

	/**
	 * How the host must match the 'path' field.
	 */
	@NotNull public final HostMatchType hostMatchType;

	/**
	 * The compiled regexp if the host is a regexp.
	 */
	@Nullable public final Pattern hostRegexp;

	//=========================================================================

	/**
	 * The (sub)path, if any.
	 */
	@Nullable public final String path;

	/**
	 * The (sub)path as lowercase, if any.
	 */
	@Nullable public final String pathLC;

	/**
	 * How the path must match the 'path' field.
	 */
	@NotNull public final PathMatchType pathMatchType;

	/**
	 * The compiled regexp if the path is a regexp.
	 */
	@Nullable public final Pattern pathRegexp;

	//=========================================================================

	/**
	 * The query, if any.
	 */
	@Nullable public final String query;

	/**
	 * The query as lowercase, if any.
	 */
	@Nullable public final String queryLC;

	/**
	 * How the query must match the 'query' field.
	 */
	@NotNull public final QueryMatchType queryMatchType;

	/**
	 * The compiled regexp if the query is a regexp.
	 */
	@Nullable public final Pattern queryRegexp;

	//=========================================================================

	protected AbstractRule(boolean isPreview,
	                       int filterId,
	                       @NotNull String filterName,
	                       @NotNull XmlArgs ruleXA,
	                       @NotNull RuleType ruleType) throws Exception
	{
		this.isPreview = isPreview;
		this.filterId = filterId;
		this.filterName = filterName;
		this.ruleId = ruleXA.getOptInt("@rule-id", -1);
		this.userPriority = ruleXA.getOptInt("@priority", 2);
		this.ruleType = ruleType;
		this.mobileOnly = ruleXA.getOptBoolean("@mobile");
		//-----------------------------------------------------
		String s = ruleXA.getOptString("@host");
		if (s == null || s.length() == 0) {
			host = null;
			hostLC = null;
			hostMatchType = HostMatchType.Ignored;
		} else if (s.startsWith("*.")) {
			host = s.substring(2);
			hostLC = host.toLowerCase();
			hostMatchType = HostMatchType.Domain;
		} else {
			host = s;
			hostLC = host.toLowerCase();
			hostMatchType = HostMatchType.fromXmlValue(ruleXA.getOptString("@host-type"));
		}
		hostRegexp = (hostMatchType == HostMatchType.Regexp)
				? Pattern.compile(host, CASE_INSENSITIVE) : null;
		//-----------------------------------------------------
		s = ruleXA.getOptString("@path");
		if (s == null || s.length() == 0) {
			path = null;
			pathLC = null;
			pathMatchType = PathMatchType.Ignored;
		} else {
			path = s;
			pathLC = path.toLowerCase();
			pathMatchType = PathMatchType.fromXmlName(ruleXA.getOptString("@path-type"));
		}
		switch (pathMatchType) {
			case Regexp:
			case NotRegexp:
				pathRegexp = Pattern.compile(path, CASE_INSENSITIVE);
				break;
			default:
				pathRegexp = null;
		}
		//-----------------------------------------------------
		s = ruleXA.getOptString("@query");
		if (s == null || s.length() == 0) {
			query = null;
			queryLC = null;
			queryMatchType = QueryMatchType.Ignored;
		} else {
			query = s;
			queryLC = query.toLowerCase();
			queryMatchType = QueryMatchType.fromXmlValue(ruleXA.getOptString("@query-type"));
		}
		switch (queryMatchType) {
			case Regexp:
			case NotRegexp:
				queryRegexp = Pattern.compile(query, CASE_INSENSITIVE);
				break;
			default:
				queryRegexp = null;
		}
	}

	/**
	 * Do NOT rely on the contents of this string for anything. It is only for debugging.
	 */
	public abstract String toString();

	/**
	 * Do NOT rely on the contents of this string for anything. It is only for debugging.
	 */
	protected void toStringMatching(StringBuilder sb)
	{
		sb.append(userPriority);
		if (mobileOnly)
			sb.append(",mobile");
		sb.append(";");
		sb.append(hostMatchType.debugDescription(host));
		sb.append(";");
		sb.append(pathMatchType.debugDescription(path));
		sb.append(";");
		sb.append(queryMatchType.debugDescription(query));
	}

	protected void appendFilterName(StringBuilder sb)
	{
		sb.append("_{").append(filterName).append('}');
	}

	/**
	 * More specific filters sort before more general filters.
	 */
	public int compareTo(@NotNull AbstractRule o)
	{
		int x = o.userPriority - userPriority; // large numbers comes before small numbers
		if (x != 0)
			return x;
		//
		x = diff(o.pathLC, o.pathMatchType.ordinal(), pathLC, pathMatchType.ordinal());
		if (x != 0)
			return x;
		//
		x = diff(o.hostLC, o.hostMatchType.ordinal(), hostLC, hostMatchType.ordinal());
		if (x != 0)
			return x;
		//
		x = diff(o.queryLC, o.queryMatchType.ordinal(), queryLC, queryMatchType.ordinal());
		if (x != 0)
			return x;
		//
		x = (o.mobileOnly ? 1 : 0) - (mobileOnly ? 1 : 0); // mobile is more specific
		if (x != 0)
			return x;
		//
		return o.ruleType.ordinal() - ruleType.ordinal();
	}

	private static int diff(@Nullable String s1, int t1, @Nullable String s2, int t2)
	{
		if (t1 != t2)
			return t1 - t2;
		if (s1 == null || s2 == null) // if the types are equal, then s1 and s2 are both null or both are non-null
			return 0;
		int diff = s1.length() - s2.length();
		if (diff != 0)
			return diff;
		else
			return s1.compareToIgnoreCase(s2);
	}
}
