/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.util;

import java.io.PrintWriter;
import java.io.StringWriter;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

public class DocumentPrinter
{
	public final Document doc;

	private PrintWriter out;
	private int indent;
	private static final int linelength = 90;

	public DocumentPrinter(Document doc)
	{
		this.doc = doc;
	}

	public String printToString() throws Exception
	{
		StringWriter sw = new StringWriter(500);
		PrintWriter pw = new PrintWriter(sw);
		print(pw);
		pw.flush();
		return sw.toString();
	}

	public void print(PrintWriter pw) throws Exception
	{
		out = pw;
		indent = 0;
		out.println("<?xml version='1.0' encoding='UTF-8' ?>");
		printNode(doc.getDocumentElement());
		out = null;
	}

	protected void printNode(Node node) throws Exception
	{
		if (node == null)
			return;
		int type = node.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:
				printDocumentNode(node);
				break;

			case Node.ELEMENT_NODE:
				printElementNode(node);
				break;

			case Node.ENTITY_REFERENCE_NODE:
				printEntityNode(node);
				break;

			case Node.CDATA_SECTION_NODE:
				printCDataNode(node);
				break;

			case Node.TEXT_NODE:
				printTextNode(node);
				break;

			case Node.PROCESSING_INSTRUCTION_NODE:
				printProcessingInstructionNode(node);
				break;

			case Node.COMMENT_NODE:
				printCommentNode(node);
				break;

			case Node.ATTRIBUTE_NODE:
				printUnexpectedNodeType(node, "ATTRIBUTE_NODE");
				break;

			case Node.DOCUMENT_FRAGMENT_NODE:
				printUnexpectedNodeType(node, "DOCUMENT_FRAGMENT_NODE");
				break;

			case Node.DOCUMENT_TYPE_NODE:
				printUnexpectedNodeType(node, "DOCUMENT_TYPE_NODE");
				break;

			case Node.ENTITY_NODE:
				printUnexpectedNodeType(node, "ENTITY_NODE");
				break;

			case Node.NOTATION_NODE:
				printUnexpectedNodeType(node, "NOTATION_NODE");
				break;

			default:
				printUnexpectedNodeType(node, "(unknown, id = " + type + ")");
		}
	}

	protected void printDocumentNode(Node node) throws Exception
	{
		Node e = ((Document)node).getDocumentElement();
		printNode(e);
	}

	protected void printElementNode(Node node) throws Exception
	{
		printIndent();
		out.print('<');
		out.print(node.getNodeName());
		printNodeAttributes(node, indent + 2 + node.getNodeName().length());
		Node child = node.getFirstChild();
		if (child == null) {
			out.println("/>");
			return;
		}
		boolean hasOnlySimpleChilds = checkIsOnlySimpleChilds(child);
		out.print('>');
		if (!hasOnlySimpleChilds)
			out.println();
		indent++;
		while (child != null) {
			if (!hasOnlySimpleChilds) {
				printNode(child);
			} else if (child.getNodeType() == Node.TEXT_NODE) {
				printNormalizedString(child.getNodeValue());
			} else {
				printEntityNode(child);
			}
			child = child.getNextSibling();
		}
		indent--;
		if (!hasOnlySimpleChilds)
			printIndent();
		out.print("</");
		out.print(node.getNodeName());
		out.println('>');
	}

	protected boolean checkIsOnlySimpleChilds(Node child) throws Exception
	{
		while (child != null) {
			int type = child.getNodeType();
			if (type != Node.TEXT_NODE && type != Node.ENTITY_REFERENCE_NODE)
				return false;
			child = child.getNextSibling();
		}
		return true;
	}

	protected void printTextNode(Node node) throws Exception
	{
		printIndent();
		printNormalizedString(node.getNodeValue());
		out.println();
	}

	protected void printEntityNode(Node node) throws Exception
	{
		out.print('&');
		out.print(node.getNodeName());
		out.print(';');
	}

	protected void printCDataNode(Node node) throws Exception
	{
		printIndent();
		out.print("<![CDATA[");
		out.print(node.getNodeValue());
		out.println("]]>");
	}

	protected void printCommentNode(Node node) throws Exception
	{
		printIndent();
		out.print("<!--");
		String s = node.getNodeValue();
		int len = s.length();
		if (len > 0) {
			char prev = s.charAt(0);
			out.print(prev);
			for (int i = 1; i < len; i++) {
				char ch = s.charAt(i);
				if (ch == '-' && prev == '-')
					out.print(' '); // -- not allowed in comments
				out.print(ch);
				prev = ch;
			}
		}
		out.println("-->");
	}

	protected void printUnexpectedNodeType(Node node, String type) throws Exception
	{
		printIndent();
		out.print("<!-- unexpected node type = " + type + ": '");
		out.print(node.getNodeValue());
		out.println("' -->");
	}

	protected void printProcessingInstructionNode(Node node) throws Exception
	{
		printIndent();
		out.print("<?");
		out.print(node.getNodeName());
		String data = node.getNodeValue();
		if (data != null && data.length() > 0) {
			out.print(' ');
			out.print(data);
		}
		out.println("?>");
	}

	void printNodeAttributes(Node n, int attrIndent) throws Exception
	{
		NamedNodeMap attrs = n.getAttributes();
		if (attrs == null)
			return;
		int numAttrs = attrs.getLength();
		int curLineLength = attrIndent;
		for (int i = 0; i < numAttrs; i++) {
			Attr attr = (Attr)attrs.item(i);
			String attrName = attr.getNodeName();
			String attrValue = attr.getNodeValue();
			int attrLen = attrName.length() + attrValue.length() + 3;
			if (i > 0 && curLineLength + attrLen > linelength) {
				out.print('\n');
				printIndent(attrIndent);
				curLineLength = attrIndent;
			} else {
				out.print(' ');
			}
			curLineLength += attrLen;
			printNormalizedString(attrName);
			out.print('=');
			out.print('"');
			printNormalizedString(attrValue);
			out.print('"');
		}
	}

	/**
	 * Normalizes the given string.
	 */
	protected void printNormalizedString(String s) throws Exception
	{
		if (s == null)
			return;
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '<':
					out.print("&lt;");
					break;
				case '>':
					out.print("&gt;");
					break;
				case '&':
					out.print("&amp;");
					break;
				case '"':
					out.print("&quot;");
					break;
				case '\r': /* fall-though */
				case '\n':
					out.print(ch);
					break;
				default:
					if ((ch >= 32 && ch < 0x7F)
							|| (ch >= 192 && ch <= 0xFFEF && ch != 222 && ch != 215 && ch != 254 && ch != 247)) {
						out.print(ch);
						continue;
					}
					out.print("&#");
					out.print(Integer.toString(ch));
					out.print(';');
			}
		}
	}

	protected void printIndent() throws Exception
	{
		printIndent(indent);
	}

	protected void printIndent(int theIndent) throws Exception
	{
		int ii = theIndent;
		while (ii > 0) {
			int subLen = Math.min(ii, indentCache.length);
			out.write(indentCache, 0, subLen);
			ii -= subLen;
		}
	}

	private static final char[] indentCache = new char[100];

	static {
		for (int i = 0; i < indentCache.length; i++)
			indentCache[i] = ' ';
	}
}
