/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.util;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.jetbrains.annotations.NotNull;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class XmlParserHelper
{
	/**
	 * Create a w3c Document. Synchronized as factory isn't thread safe.
	 */
	@NotNull
	public static synchronized Document createDocument() throws ParserConfigurationException
	{
		return createDocumentBuilder().newDocument();
	}

	@NotNull
	public static synchronized Document parse(InputStream is) throws Exception
	{
		Document doc = createDocumentBuilder().parse(is);
		trimWhitespaces(doc.getDocumentElement());
		return doc;
	}

	public static void saveDocument(Document doc, File f) throws Exception
	{
		String s = new DocumentPrinter(doc).printToString();
		File tmpF = new File(f.getPath() + "-tmp");
		try {
			try (FileOutputStream fos = new FileOutputStream(tmpF)) {
				fos.write(s.getBytes("UTF-8"));
				fos.flush();
				// For filesystems which journals metadata only:
				// Make sure data goes to platter so we're not left with empty files
				// if the system crashes right after renameTo()
				// https://bugs.edge.launchpad.net/ubuntu/+source/linux/+bug/317781/comments/54
				FileDescriptor fd = fos.getFD();
				fd.sync(); // ensures data written to drive platter, not just OS buffers.
			}
		}
		catch (IOException ex) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete " + tmpF);
			throw ex;
		}
		if (!tmpF.renameTo(f)) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete " + tmpF);
			throw new IOException("Can't rename tmp file: " + tmpF.getPath());
		}
	}

	public static void trimWhitespaces(Element parentE)
	{
		if (parentE == null)
			return;
		Node n = parentE.getFirstChild();
		while (n != null) {
			Node next = n.getNextSibling();
			switch (n.getNodeType()) {
				case Node.ELEMENT_NODE:
					trimWhitespaces((Element)n);
					break;
				case Node.TEXT_NODE:
					if (n.getNodeValue().trim().length() == 0)
						parentE.removeChild(n);
			}
			n = next;
		}
	}

	@NotNull
	public static Document parse(File f) throws Exception
	{
		try (FileInputStream fis = new FileInputStream(f)) {
			return parse(new BufferedInputStream(fis, 2048));
		}
	}

	private static DocumentBuilder createDocumentBuilder() throws ParserConfigurationException
	{
		if (factory == null) {
			factory = DocumentBuilderFactory.newInstance();
			factory.setExpandEntityReferences(true);
			factory.setIgnoringComments(false);
			factory.setIgnoringElementContentWhitespace(false);
			factory.setNamespaceAware(true);
			factory.setValidating(false);
			try {
				factory.setXIncludeAware(false);
			}
			catch (Exception ex) {
				System.err.println("WARNING: you have an OLD xml parser installed,");
				System.err.println("         probably in /Library/Java/Extensions");
				System.err.println("DocumentBuilderFactory: " + factory.getClass().getName());
				System.err.println(ExceptionUtils.getExMsg(ex));
			}
			factory.setAttribute("http://apache.org/xml/features/dom/defer-node-expansion", false);
			factory.setAttribute("http://apache.org/xml/features/validation/schema", false);
			factory.setAttribute("http://apache.org/xml/features/validation/schema-full-checking", false);
			//simpleFactory.setAttribute("http://apache.org/xml/features/validate-annotations", true);
			//factory.setAttribute("http://apache.org/xml/features/validation/warn-on-undeclared-elemdef", true);
			factory.setAttribute("http://apache.org/xml/features/warn-on-duplicate-entitydef", true);
			factory.setAttribute("http://apache.org/xml/features/validation/warn-on-duplicate-attdef", true);
			//factory.setAttribute("http://apache.org/xml/features/standard-uri-conformant", true);
		}
		return factory.newDocumentBuilder();
	}

	private static DocumentBuilderFactory factory;

	/**
	 * Zero-ton.
	 */
	private XmlParserHelper()
	{
	}

	public static void main(String[] args) throws ParserConfigurationException
	{
		createDocumentBuilder();
	}
}
