/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <spawn.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <unistd.h>
#include <jni.h>
#include <dlfcn.h>
#include <CoreFoundation/CoreFoundation.h>
#include <Foundation/Foundation.h>

static int debugFlag = 0;
static int makeDebugReport = 0;
static const char *mainClassPath = "org.glimmerblocker.proxy.GlimmerBlocker";

#define PROXYAPPDIR "/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerProxy.app"
#define JARDIR PROXYAPPDIR "/Contents/Resources/Java"
#define CLASSPATH JARDIR "/glimmerblocker-server.jar" \
/**/					 ":" JARDIR "/annotations.jar" \
/**/					 ":" JARDIR "/dnsjava.jar" \
/**/					 ":" JARDIR "/smalljs.jar"

#define PPANE "/Library/PreferencePanes/GlimmerBlocker.prefPane"

#define JREBUNDLE PPANE "/Contents/GlimmerBlockerProxy.app/Contents/Resources/jre"
#define JREHOME JREBUNDLE "/Contents/Home/jre"

typedef jint (JNICALL *CreateJavaVM_t)(JavaVM **pvm, void **env, void *args);
typedef void (JNICALL *SetPreferredJVM_t)(const char *prefJVM);

static JavaVM *jvm = NULL;
static NSString *jreLaunchFailure = @"GB-JRE-Launch-Failure";
static JavaVMInitArgs vm_args;
static const int maxVmArgs = 100;
static NSMutableDictionary *seenArgs = NULL;

static void addJavaOption(const char* s)
{
	if (vm_args.nOptions >= maxVmArgs) {
		NSLog(@"Too many VM args: %d", vm_args.nOptions);
		exit(1);
	}
	if (!seenArgs)
		seenArgs = [NSMutableDictionary dictionary];
	const char *keyEnd = s;
	while (*keyEnd && *keyEnd != ':' && (*keyEnd < '0' || *keyEnd > '9'))
		keyEnd++;
	NSString *key = [[NSString alloc] initWithBytes:s length:(keyEnd - s) encoding:NSUTF8StringEncoding];
	NSString *prev = seenArgs[key];
	if (prev) {
		NSLog(@"Has vmarg '%@', ignores '%s'", prev, s);
		return;
	}
	NSString *v = @(s);
	NSLog(@"Adds vm arg: '%@'", v);
	seenArgs[key] = v;
	vm_args.options[vm_args.nOptions++].optionString = (char*)s;
}

static void initVmArgs()
{
	if (debugFlag)
		NSLog(@"initVmArgs.1");
	memset(&vm_args, 0, sizeof(vm_args));
	vm_args.version	= JNI_VERSION_1_6;
	vm_args.options	= calloc(maxVmArgs + 2, sizeof(JavaVMOption));
	vm_args.nOptions = 0;
	vm_args.ignoreUnrecognized = JNI_TRUE;
	if (!vm_args.options)
		[NSException raise:jreLaunchFailure format:@"calloc failed"];
	//
	NSArray *userArgs = [NSArray arrayWithContentsOfFile:@"/Library/GlimmerBlocker/JavaArgs.plist"];
	if (userArgs) {
		NSLog(@"Did load JavaArgs.plist");
		for (int i = 0; i < [userArgs count]; i++) {
			NSString *s = userArgs[i];
			addJavaOption([s UTF8String]);
		}
	} else {
		NSLog(@"Has no JavaArgs.plist");
	}
	addJavaOption("-Xms25m");
	addJavaOption("-Xmx50m");
	addJavaOption("-XX:+UseCompressedOops");
	addJavaOption("-Djava.awt.headless=true");
	addJavaOption("-Dfile.encoding=UTF-8");
	addJavaOption("-Duser.region=US");
	addJavaOption("-Duser.language=en"); // JVM classes might throw NPE at startup if this is not defined.
	struct stat statbuf;
	if (getenv("GB_JAVA_DEBUGGER") || !stat("/Library/GlimmerBlocker/enable-java-debugging.txt", &statbuf)) {
		NSLog(@"Enables java debugging.  Pid = %d", getpid());
		addJavaOption("-Xint");
		addJavaOption("-Xshare:off");
		addJavaOption("-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=5005");
	}
	if (!stat("/Library/GlimmerBlocker/enable-jmxremote.txt", &statbuf)) {
		NSLog(@"Enables jmxremote for e.g. jvisualvm.  Pid = %d", getpid());
		NSLog(@" use 'Add JMXConnection' to 127.0.0.1:9494");
		addJavaOption("-Dcom.sun.management.jmxremote.port=9494");
		addJavaOption("-Dcom.sun.management.jmxremote.ssl=false");
		addJavaOption("-Dcom.sun.management.jmxremote.authenticate=false");
	}
	addJavaOption("-Djava.class.path=" CLASSPATH); // using -classpath xxx:yyy doesn't work.
	if (!stat("/Library/GlimmerBlocker/preferIPv4Stack.txt", &statbuf)) {
		addJavaOption("-Djava.net.preferIPv4Stack=true");
		NSLog(@"Launches GB using java.net.preferIPv4Stack");
	} else if (!stat("/Library/GlimmerBlocker/preferIPv6Stack.txt", &statbuf)) {
		addJavaOption("-Djava.net.preferIPv6Addresses=true");
		NSLog(@"Launches GB using java.net.preferIPv6Addresses");
	}
	//------------------------------------
	if (debugFlag) {
		NSLog(@"Java VM options:");
		for (int i = 0; i < vm_args.nOptions; i++)
			NSLog(@"   %s", vm_args.options[i].optionString);
	}
	if (debugFlag)
		NSLog(@"initVmArgs.2");
}


#define CHECK_EXCEPTION(MSG) \
	if ((*env)->ExceptionOccurred(env)) { \
		NSLog(@"Java: Got exception from " MSG "."); \
		[NSException raise:jreLaunchFailure format:@"Got exception from: " MSG]; \
	}

static void startEmbeddedJre()
{
	JNIEnv *env = NULL;
	initVmArgs();
	@autoreleasepool {
		@try {
			// load the libjli.dylib of the embedded JRE (or JDK) bundle.
			//
			// New version.  Seems to work OK on Mac OS 10.9 and 10.10 without prompting for old java install.
			// Can't use CFBundleLoadExecutableAndReturnError() as there is no executable and
			// codesign version 2 (10.9.5) doesn't allow a symlink for the executable.
			// If the dylib is copied into the folder, it can't find libjava.
			if (debugFlag)
				NSLog(@"dlopens libjava.1");
			void *lib = dlopen(JREHOME "/lib/jli/libjli.dylib", RTLD_NOW);
			if (debugFlag || !lib)
				NSLog(@"did dlopen, handle = %p", lib);
			CreateJavaVM_t createJVMFxnPtr = dlsym(lib, "JNI_CreateJavaVM");
			if (debugFlag || !createJVMFxnPtr)
				NSLog(@"did dlsym, func = %p", lib);
			jint createJVMStatus = createJVMFxnPtr(&jvm, (void **)&env, &vm_args);
			NSLog(@"got createJVMStatus = %d", createJVMStatus);
/*
			// This code MUST use the CF* functions as Apple has hacked dlopen() to prompt to install the old Java 6 VM.
			CFURLRef url = CFURLCreateFromFileSystemRepresentation(kCFAllocatorDefault, (const UInt8 *)JREBUNDLE, strlen(JREBUNDLE), false);
			if (!url) {
				[NSException raise:jreLaunchFailure format:@"could not make file url to jre bundle: %s", JREBUNDLE];
			}
			if (debugFlag)
				NSLog(@"jre bundle url: %@", url);
			CFBundleRef javaVMBundle = CFBundleCreate(kCFAllocatorDefault, url);
			if (!javaVMBundle) {
				[NSException raise:jreLaunchFailure format:@"CFBundleCreate failed loading jre bundle: %s", JREBUNDLE];
			}
			if (debugFlag)
				NSLog(@"jre bundle ref created: %ld", (long)javaVMBundle);
			//
			CFErrorRef cferr = NULL;
			Boolean jreBundleLoaded = CFBundleLoadExecutableAndReturnError(javaVMBundle, &cferr);
			if (cferr != nil || !jreBundleLoaded) {
				NSError *err = (__bridge NSError*)cferr;
				[NSException raise:jreLaunchFailure format:@"could not load the JRE/JDK: %@", err];
			}
			if (debugFlag)
				NSLog(@"jre bundle loaded: %ld", (long)javaVMBundle);
			CreateJavaVM_t createJVMFxnPtr = CFBundleGetFunctionPointerForName(javaVMBundle, CFSTR("JNI_CreateJavaVM"));
			if (!createJVMFxnPtr) {
				[NSException raise:jreLaunchFailure format:@"got NULL JNI_CreateJavaVM from java CF dylib"];
			}
			if (debugFlag)
				NSLog(@"Got pointer to JNI_CreateJavaVM, wants vm_args.version = %d with %d args", vm_args.version, vm_args.nOptions);
			// instantiate the JVM
			jint createJVMStatus = createJVMFxnPtr(&jvm, (void **)&env, &vm_args);
*/
			if (createJVMStatus != JNI_OK) {
				[NSException raise:jreLaunchFailure format:@"failed to JNI_CreateJavaVM (%d)", createJVMStatus];
			}
			if (YES || debugFlag)
				NSLog(@"did call JNI_CreateJavaVM()");
			//
			char *mainClassPathSlash = strdup(mainClassPath);
			if (!mainClassPathSlash)
				[NSException raise:jreLaunchFailure format:@"strdup failed"];
			for (char *p = mainClassPathSlash; *p; p++) {
				if (*p == '.')
					*p = '/';
			}
			jclass mainClass = (*env)->FindClass(env, mainClassPathSlash);
			if (!mainClass) {
				NSLog(@"Java: Can't load main class.");
				CHECK_EXCEPTION("Loading main class");
			}
			if (debugFlag)
				NSLog(@"Got main class");
			//
			jmethodID mainID = (*env)->GetStaticMethodID(env, mainClass, "main", "([Ljava/lang/String;)V");
			if (!mainID) {
				NSLog(@"Java: Can't find 'public void main(...)' in main class.");
				CHECK_EXCEPTION("Finding main() method");
			}
			if (debugFlag)
				NSLog(@"Got main() method");
			//
			jobjectArray mainArgs = (*env)->NewObjectArray(env, debugFlag ? 1 : 0, (*env)->FindClass(env, "java/lang/String"), NULL);
			if (mainArgs == NULL) {
				NSLog(@"Java: Can't create String[] array.");
				CHECK_EXCEPTION("Creating String[] array");
			}
			if (debugFlag)
				NSLog(@"made String[] array");
			if (debugFlag) {
				jstring s = (*env)->NewStringUTF(env, "-stderr");
				if (!s) {
					NSLog(@"Java: Can't create String constant.");
					CHECK_EXCEPTION("Creating String constant");
				}
				(*env)->SetObjectArrayElement(env, mainArgs, 0, s);
				CHECK_EXCEPTION("Setting -stderr argument");
				if (debugFlag)
					NSLog(@"did set String[0] to -stderr");
			}
			if (debugFlag)
				NSLog(@"Java: Calls main()");
			(*env)->CallStaticVoidMethod(env, mainClass, mainID, mainArgs);
			CHECK_EXCEPTION("Executing main method");
			if (debugFlag)
				NSLog(@"Java: main() returned.");
		} @catch (NSException *e) {
			NSLog(@"%@: %@", e, [e callStackSymbols]);
		}
		if (jvm) {
			(*jvm)->DetachCurrentThread(jvm);
			(*jvm)->DestroyJavaVM(jvm);
		}
	}
}

static void sourceCallBack(void *info) // dummy for CFRunLoop so it doesn't exit.
{
}

static int sys(const char *x)
{
	NSLog(@"Executes: %s", x);
	int r = system(x);
	if (r)
		NSLog(@"Got returncode = %d from %s", r, x);
	NSLog(@"");
	return r;
}

void checkCodesign(const char *id, const char *relPath)
{
#define SIGNATURE "2307A8B6203D3C312D83BD0F958796CB20EDB5EE"
	char cmd[1024];
	sprintf(cmd, "codesign -vv '-R=identifier %s and certificate leaf = H\"%s\"' %s", id, SIGNATURE, relPath);
	NSLog(@"Executes: %s", cmd);
	int r = system(cmd);
	if (r) {
		NSLog(@"");
		NSLog(@"");
		NSLog(@"#############################################################");
		NSLog(@"####  Codesign verification failed with returncode = %3d ####", r);
		NSLog(@"#############################################################");
		NSLog(@"");
		NSLog(@"");
		sprintf(cmd, "codesign -vddd -r- %s", relPath);
		NSLog(@"Executes: %s", cmd);
		system(cmd);
	}
	NSLog(@"");
}

int main(int argc, const char **argv)
{
	unsetenv("IFS");
	setenv("SHELL", "/bin/sh", 1);
	setenv("PATH", "/bin:/usr/bin:/usr/sbin", 1);
	setenv("JAVA_HOME", JREHOME, 1);
	unsetenv("CDPATH");
	struct stat statbuf;
	if (!stat("/Library/GlimmerBlocker/enable-javalauncher-debugging.txt", &statbuf))
		debugFlag = 1;
	int r = chdir("/tmp");
	if (r)
		NSLog(@"GlimmerBlockerProxy: failed chdir('/tmp'): %d", r);
	int displayedInTextEdit = 0;
	int exitAfterJavaMainReturns = 0;
	for (int i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "--make-report")) {
			makeDebugReport = debugFlag = 1;
		} else if (!strcmp(argv[i], "--textedit")) {
			displayedInTextEdit = 1;
		} else if (!strcmp(argv[i], "-d")) {
			debugFlag = 1;
		} else if (!strcmp(argv[i], "--main-class")) {
			if (++i >= argc) {
				NSLog(@"Missing argument to --main-class");
				exit(1);
			}
			mainClassPath = argv[i];
			exitAfterJavaMainReturns = 1;
		} else if (!strncmp(argv[i], "--gb-version", 12)) {
			// ignore
		} else {
			NSLog(@"Ignored command line arg: %s", argv[i]);
		}
	}
	struct rlimit rlp;
	struct rlimit rlpOld;
	getrlimit(RLIMIT_NOFILE, &rlpOld);
	getrlimit(RLIMIT_NOFILE, &rlp);
	if (rlp.rlim_cur < rlp.rlim_max) {
		rlp.rlim_cur = MAX(rlp.rlim_max, 10240);
		setrlimit(RLIMIT_NOFILE, &rlp);
	}
	if (makeDebugReport) {
		if (displayedInTextEdit) {
			NSLog(@"#");
			NSLog(@"#");
			NSLog(@"# Please copy/paste the content of this window into a mail to");
			NSLog(@"#	 feedback@glimmerblocker.org");
			NSLog(@"#");
			NSLog(@"#");
		} else {
			NSLog(@"GlimmerBlockerProxy: debug log enabled.");
		}
		NSLog(@"rlimit.files: %lld -> %lld, hard = %lld",
				(long long)rlpOld.rlim_cur, (long long)rlp.rlim_cur, (long long)rlp.rlim_max);
		sys("uname -a");
		sys("printenv");
		sys("id");
		sys("cat /Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/Info.plist"
				"| perl -pe 's!\\s*</?key>\\s*!!g;'"
				"| perl -ne 's!</?string>!!g; print if /BuildMachineOSBuild|CFBundleVersion|DTPlatformBuild|DTSDKBuild|DTSDKName|DTXcode|DTXcodeBuild/;'");
		sys("ls -ld /Library/Logs/GlimmerBlocker");
		sys("ls -l /Library/Logs/GlimmerBlocker");
		sys("ls -ld /Library/GlimmerBlocker");
		sys("ls -l /Library/GlimmerBlocker");
		sys("codesign --verify  --verbose /Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerProxy.app");
		sys("dscl . -read /Users/_glimmerblocker");
		sys("du -sh "JREHOME);
		//sys(JREHOME "/bin/java -version");
		sys("perl -e 'use IO qw(File); print \"perl is ok\\n\";'"); // http://www.theregister.co.uk/2009/02/16/apple_update_perl_breakage/
		sys("system_profiler SPSoftwareDataType | egrep -i 'version|kernel'");
		sys("system_profiler SPHardwareDataType | perl -ne 'print unless /Serial Number|Hardware UUID/i;' # omits serial + hardwareid.");
		checkCodesign("org.glimmerblocker.prefsPane", PPANE);
		checkCodesign("org.glimmerblocker.proxyApp", PPANE "/Contents/GlimmerBlockerProxy.app");
		checkCodesign("org.glimmerblocker.networkConfigurationListener", PPANE "/Contents/NetworkConfigurationListener.app");
		checkCodesign("org.glimmerblocker.authHelper", PPANE "/Contents/AuthHelper.app");
		checkCodesign("org.glimmerblocker.installer", PPANE "/Contents/GlimmerBlockerInstaller.app");
		sys("ulimit -a");
	}
	startEmbeddedJre();
	sleep(2);
	if (makeDebugReport) {
		NSLog(@"#");
		NSLog(@"# Debug-mode launch of GlimmerBlocker proxy now done.");
		NSLog(@"#");
		NSLog(@"# Please copy/paste the content of this window into a mail to");
		NSLog(@"#	 feedback@glimmerblocker.org");
		if (!displayedInTextEdit)
			NSLog(@"# and then close this window and quit Terminal.");
		NSLog(@"#");
		exit(0);
	}
	if (exitAfterJavaMainReturns)
		exit(0);
	//
	// Create a a sourceContext to be used by our source that makes
	// sure the CFRunLoop doesn't exit right away
	CFRunLoopSourceContext sourceContext;
	sourceContext.version = 0;
	sourceContext.info = NULL;
	sourceContext.retain = NULL;
	sourceContext.release = NULL;
	sourceContext.copyDescription = NULL;
	sourceContext.equal = NULL;
	sourceContext.hash = NULL;
	sourceContext.schedule = NULL;
	sourceContext.cancel = NULL;
	sourceContext.perform = &sourceCallBack;
	//
	CFRunLoopSourceRef sourceRef = CFRunLoopSourceCreate(NULL, 0, &sourceContext);
	CFRunLoopAddSource(CFRunLoopGetCurrent(), sourceRef, kCFRunLoopCommonModes);
	CFRunLoopRun();
	CFRelease(sourceRef);
	return 0;
}
